/*
   This file is part of the MoCA video library.
   Name: CameraOperationTest.cpp
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/

// C++ includes
#include <string>
#include <iostream>
using namespace std;

// MoCA includes
#include <FileList.h>
#include <CameraModel.h>
#include <CameraOperation.h>
#include <MotionVector.h>
#include <MedianFilter.h>



void usage()
{
	cerr << "Usage: CameraOperation [options] filename|path" << endl;
	cerr << CAMERA_OPERATION_VERSION << endl;
	cerr << endl;
	cerr << "options:" << endl;
	cerr << " -recursive      analyses all *.avi files in path" << endl;
	cerr << endl;
	cerr << "Example:" <<  endl;
	cerr << "CameraOperation test.avi" << endl;
	cerr << "CameraOperation -recursive /opt/movies/" << endl;
	exit(1);
} // usage


void processFile (string &fn)
{
	{
		cerr << "start camera model..." << ends;
		CameraModelData data;
		CameraModel cm (fn);
		cm.process (data);
		cerr << endl << "start camera model...done " << endl;
		cout << data << endl;

		// write average motion vector length
		VideoDecoder reader (fn);
		ImageSize size = reader.getImageSize();
		SignalSize s (data.data.size());
		Signal<int32> avg (s);
		avg.clear(0);

		int32 av=0;
		for (int32 i=0; i<data.data.size(); i++) {
			CameraModelItem item = data.data[i];
			
			ieee64 averageLength = 0;
			if (item.isValid()) {
				
				ieee64 oldX=0;
				ieee64 oldY=0;
				ieee64 newX, newY;
				item.getVectors(oldX, oldY, newX, newY);
				newX-=oldX; newY-=oldY; 
				ieee64 tmp = newX*newX + newY*newY;
				if (tmp>0) tmp = sqrt (tmp);
				averageLength = tmp;
				
				oldX=size[DIM_X]-1;
				item.getVectors(oldX, oldY, newX, newY);
				newX-=oldX; newY-=oldY; 
				tmp = newX*newX + newY*newY;
				if (tmp>0) tmp = sqrt (tmp);
				averageLength += tmp;
				
				oldY=size[DIM_Y]-1;
				item.getVectors(oldX, oldY, newX, newY);
				newX-=oldX; newY-=oldY; 
				tmp = newX*newX + newY*newY;
				if (tmp>0) tmp = sqrt (tmp);
				averageLength += tmp;
				
				oldX=0;
				item.getVectors(oldX, oldY, newX, newY);
				newX-=oldX; newY-=oldY; 
				tmp = newX*newX + newY*newY;
				if (tmp>0) tmp = sqrt (tmp);
				averageLength += tmp;
				averageLength = averageLength / 4 + 0.5;
			}

			avg (i) = (int32)averageLength;
			av+=avg (i);
			//cout << i << "\t" << averageLength << endl;
		}
		av/= data.data.size();
		cerr << "average value: " << av << endl;
		int32 av2=0;
		int32 cnt=0;
		for (int32 i=0; i<data.data.size(); i++) {
			if (avg (i) > 2*av) {
				av2+=avg (i);
				cnt++;
			} // if
		}
		if (cnt>0) av2/=cnt++;
		cerr << "average of high values: " << av2 << endl;



		// store data
		string fn2, ext;
		FileHandling::getFileExtension(fn,ext, fn2);
		fn2+=".txt";
		cerr << "store data in " << fn2 << endl;
		ofstream os (fn2.c_str(), std::ios::out | std::ios::binary);
		os << avg << endl;
		os.close();
		
	}

	{
		cerr << "start camera operation..." << ends;
		CameraOperationData data;
		CameraOperation co (fn);
		co.process (data);
		cerr << endl << "start camera operation...done " << endl;
		cout << data << endl;
	}


	{
		cerr << "start motion vector..." << ends;
		MotionVectorStatus data;
		MotionVector mv (fn);
		mv.process (data);
		cerr << endl << "start motion vector...done " << endl;
		cout << data << endl;
	}
	
	return;
}


#include<IImageStream.h>
#include<LuminanceFilter.h>

int main(int argc, char* argv[])
{

	string filename = "";

    if (argc<2) usage();
	string s;
	bool recursive=false;

	for (int i=1; i<argc; i++) {
		s=argv[i];
		for (uint32 k=0; k<s.size(); k++)
		  s[k]=toupper(s[k]);

		if (s=="-RECURSIVE") recursive=true;
		else {
			if (i != argc-1) usage();
		} // else
	} // for


	if (!recursive) {
		processFile (s);
	} else {
		
		// read all AVI files in directory
		{
			cerr << endl << "analyse AVI files..." << endl;
			vector<FileInfo> aviFiles;
			FileList files(s, true);    
			files.process("*.avi", aviFiles);
			
			// open mpeg files
			vector<FileInfo>::iterator it;
			for (it = aviFiles.begin(); it != aviFiles.end(); it++) {
				
				cout << endl  << "processing file: "<< (*it).name << endl;
				processFile ((*it).name);
			} // for
			cout << endl << aviFiles.size() << " files found." << endl;
		}


		// read all MPEG files in directory
		{
			cerr << endl << "analyse MPG files..." << endl;
			vector<FileInfo> aviFiles;
			FileList files(s, true);    
			files.process("*.mpg", aviFiles);
			
			// open mpeg files
			vector<FileInfo>::iterator it;
			for (it = aviFiles.begin(); it != aviFiles.end(); it++) {
				
				cout << endl  << "processing file: "<< (*it).name << endl;
				processFile ((*it).name);
			} // for
			cout << endl << aviFiles.size() << " files found." << endl;
		}

	} // else

	return 0;
} // main


