/*
    BlockMotionFeature.cpp     calculates the motion of a block in a video
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <BlockMotionFeature.h>



/*!
  Calculates summed absolute error of two macroblocks

  \param src   The source video
  \param frame Number of the current frame
  \param curX
  \param curY  Position of the macroblock in the current frame
  \param nextX
  \param nextY  Position of the macroblock in the next frame
*/


ieee64
BlockMotionFeature::match(const Video<uint8>& src, indexType frame,
      indexType curX, indexType curY,
      indexType nextX, indexType nextY)
{
    ieee64 distance = 0.0;

    for (indexType y = 0; y < blockSize; y++) {
        const uint8* curP  = src.getSamplePtr(curX, curY+y, frame);
        const uint8* nextP = src.getSamplePtr(nextX, nextY+y, frame+1);
        for (indexType x = 0; x < blockSize; x++) {
            distance += abs(*curP - *nextP);
            curP  += src.getSampleOffsetX(curX+x);
            nextP += src.getSampleOffsetX(nextX+x);
        }
    }

    return distance;
}



ieee64
BlockMotionFeature::match(Image<uint8>& src1, Image<uint8>& src2, 
      indexType curX, indexType curY,
      indexType nextX, indexType nextY)
{
    ieee64 distance = 0.0;

    for (indexType y = 0; y < blockSize; y++) {
        uint8* curP  = src1.getSamplePtr(curX, curY+y);
        uint8* nextP = src2.getSamplePtr(nextX, nextY+y);
        for (indexType x = 0; x < blockSize; x++) {
            distance += abs(*curP - *nextP);
            curP  += src1.getSampleOffsetX(curX+x);
            nextP += src2.getSampleOffsetX(nextX+x);
        }
    }

    return distance;
}


/*!
  Simple full search algorithm
  \TODO Even for border blocks the full search range is used
  \TODO Position (0,0) is calculated twice

  \param src The source video
  \param dest Where to place motion vectors and mean absolute error
 */

void
BlockMotionFeature::fullSearch(const Video<uint8>& src, Video<int8>& dest)
{
    VideoSize destSize = dest.getSize();
    ieee64 norm = 1.0 / (blockSize * blockSize);

    // for each frame
    for (sizeType z = 0; z < destSize[DIM_Z]; z++) {
        cout << "slice " << z << endl;

        indexType yPos = 0;
        for (sizeType y = 0; y < destSize[DIM_Y]; y++) {
            indexType xPos = 0;
            for (sizeType x = 0; x < destSize[DIM_X]; x++) {

                // try "no motion case" first
                int16 vectorX = 0, vectorY = 0;

                ieee64 minDistance = match(src, z, xPos, yPos, xPos, yPos);

                // search within search range
                for (int8 dy = -searchRange; dy <= searchRange; dy++) {
                    for (int8 dx = -searchRange; dx <= searchRange; dx++) {
                        ieee64 distance = match(src, z, xPos, yPos, xPos + dx, yPos + dy);
                        if (distance < minDistance) {
                            minDistance = distance;
                            vectorX = dx;
                            vectorY = dy;
                        } // if
                    } // for dx
                } // for dy

                dest(x,y,z,0) = vectorX;
                dest(x,y,z,1) = vectorY;

		ieee64 tmp = minDistance * norm;
		if (tmp<-120) tmp=-120.0;
		if (tmp> 120) tmp= 120.0;
                dest(x,y,z,2) = (int8)tmp;

                xPos += blockSize;
            } // for x
            yPos += blockSize;
        } // for y

    } // for z
}


void
BlockMotionFeature::fullSearch(Image<uint8>& src1, Image<uint8>& src2, Image<int8>& dest)
{
    ImageSize destSize = dest.getSize();
    ieee64 norm = 1.0 / (blockSize * blockSize);
	
	indexType yPos = 0;
	for (sizeType y = 0; y < destSize[DIM_Y]; y++) {
		indexType xPos = 0;
		for (sizeType x = 0; x < destSize[DIM_X]; x++) {
			
			// try "no motion case" first
			int8 vectorX = 0, vectorY = 0;
			
			ieee64 minDistance = match(src1, src2, xPos, yPos, xPos, yPos);
			
			// search within search range
			for (indexType dy = -searchRange; dy <= searchRange; dy++) {
				for (indexType dx = -searchRange; dx <= searchRange; dx++) {
					ieee64 distance = match(src1, src2, xPos, yPos, xPos + dx, yPos + dy);
					if (distance < minDistance) {
						minDistance = distance;
						vectorX = dx;
						vectorY = dy;
					} // if
				} // for dx
			} // for dy
			
			dest(x,y,0) = vectorX;
			dest(x,y,1) = vectorY;

			ieee64 tmp = minDistance * norm;
			if (tmp<-120) tmp=-120.0;
			if (tmp> 120) tmp= 120.0;
			dest(x,y,2) = (int8)tmp;
			
			xPos += blockSize;
		} // for x
		yPos += blockSize;
	} // for y
	
	return;
}


void
BlockMotionFeature::fullSearchScaled(Image<uint8>& src1, Image<uint8>& src2, Image<int8>& defaultMotionVectors, Image<int8>& motionVectors)
{
    ImageSize motionVectorSize = motionVectors.getSize();
    ImageSize imgSize = src1.getSize();
	int offset = searchRange;
	
	for (sizeType y = 0; y < motionVectorSize[DIM_Y]; y++) {
		for (sizeType x = 0; x < motionVectorSize[DIM_X]; x++) {

			indexType imgPosX1 = x*blockSize + offset;
			indexType imgPosY1 = y*blockSize + offset;
			if (imgPosX1+blockSize >= imgSize[DIM_X]) {
				imgPosX1 = imgSize[DIM_X]-blockSize-1;
			} // if
			if (imgPosY1+blockSize >= imgSize[DIM_Y]) {
				imgPosY1 = imgSize[DIM_Y]-blockSize-1;
			} // if

			indexType imgPosX2 = imgPosX1 + defaultMotionVectors(x,y,0);
			indexType imgPosY2 = imgPosY1 + defaultMotionVectors(x,y,1);
			
			// try "no motion case" first
			ieee64 vectorX = 0;
			ieee64 vectorY = 0;			
			ieee64 minDistance = match(src1, src2, imgPosX1, imgPosY1, imgPosX1, imgPosY1);
			
			// search within search range
			int px1=imgPosX2-searchRange; 
			int px2=imgPosX2+searchRange;
			int py1=imgPosY2-searchRange;
			int py2=imgPosY2+searchRange;
			if (px1<0) {
				px1=0;
			} // if
			if (px2+blockSize>imgSize[DIM_X]) {
				px2=imgSize[DIM_X]-blockSize;
			} // if
			if (py1<0) {
				py1=0;
			} // if
			if (py2+blockSize>imgSize[DIM_Y]) {
				py2=imgSize[DIM_Y]-blockSize;
			} // if

			//for (indexType dy = -searchRange; dy <= searchRange; dy++) {
			//	for (indexType dx = -searchRange; dx <= searchRange; dx++) {
			for (indexType dy = py1; dy <= py2; dy++) {
				for (indexType dx = px1; dx <= px2; dx++) {
					ieee64 distance = match(src1, src2, imgPosX1, imgPosY1, dx, dy);
					if (distance < minDistance) {
						minDistance = distance;
						vectorX = dx-imgPosX1;
						vectorY = dy-imgPosY1;
					} // if
				} // for dx
			} // for dy
			
			if (vectorX<-125) vectorX=-125;
			if (vectorY<-125) vectorY=-125;
			if (vectorX> 125) vectorX= 125;
			if (vectorY> 125) vectorY= 125;
			if (minDistance>125) minDistance=125;
			if (minDistance<  0) minDistance=  0;

			motionVectors(x,y,0) = (int8)vectorX;
			motionVectors(x,y,1) = (int8)vectorY;			
			motionVectors(x,y,2) = (int8)minDistance;			
		} // for x
	} // for y
	
	return;
}




/*!
  The function calculates block motion vectors for the whole video
  sequence based on a full search algorithm. Macroblocks from frame i
  are searched in frame i+1.  The destination video consists of three
  bands: x components of the vectors, y components of the vectors, and
  the mean absolute distance for this block.
*/

void
BlockMotionFeature::process (const Video<uint8>& src, Video<int8>& dest)
{
    VideoSize srcSize = src.getSize();
    VideoSize destSize;

    destSize[DIM_X]    = srcSize[DIM_X]/blockSize;
    destSize[DIM_Y]    = srcSize[DIM_Y]/blockSize;
    destSize[DIM_Z]    = srcSize[DIM_Z] - 1;

    // motion vector (x,y) and error
    destSize[DIM_BAND] = 3;
    dest.resize(destSize);

    fullSearch(src, dest);
}


void
BlockMotionFeature::process (Image<uint8>& src1, Image<uint8>& src2, Image<int8>& motionVectors)
{
	// get size of source images
    ImageSize srcSize = src1.getSize();

	// get size of motion vector field
    ImageSize motionVectorSize;

    motionVectorSize[DIM_X]    = srcSize[DIM_X]/blockSize;
    motionVectorSize[DIM_Y]    = srcSize[DIM_Y]/blockSize;
    motionVectorSize[DIM_BAND] = 3;
    motionVectorSize[DIM_Z] = 1;
    motionVectors.resize(motionVectorSize);
	motionVectors.clear(0);

	// scale image
	Image<uint8> tmp1;
	Image<uint8> tmp2;
	Image<int8> tmpMotionVectors;

	int scaleSize=64;
	while (scaleSize > searchRange) scaleSize/=2;

	int oldSearchRange=searchRange;
	searchRange=3;

	while (scaleSize > 1) {
		
		// scale image and search motion vectors
		ScaleFilter<uint8> scale;
		scale.reduce (src1, tmp1, scaleSize, scaleSize);
		scale.reduce (src2, tmp2, scaleSize, scaleSize);
		
		// set size of motion field			
		motionVectorSize[DIM_X]    = (srcSize[DIM_X]/blockSize)/scaleSize;
		motionVectorSize[DIM_Y]    = (srcSize[DIM_Y]/blockSize)/scaleSize;
		tmpMotionVectors.resize(motionVectorSize);
		
		// calculate motion vectors
		fullSearchScaled (tmp1,tmp2,motionVectors,tmpMotionVectors);
		
		// copy and duplicate motion vectors
		for (int y=0; y<motionVectorSize[DIM_Y]; y++) {
			for (int x=0; x<motionVectorSize[DIM_X]; x++) {
				for (int jy=0; jy<scaleSize; jy++) {
					for (int jx=0; jx<scaleSize; jx++) {
						motionVectors(scaleSize*x+jx, scaleSize*y+jy,0) = 2*tmpMotionVectors (x,y,0);
						motionVectors(scaleSize*x+jx, scaleSize*y+jy,1) = 2*tmpMotionVectors (x,y,1);
					} // for				
				} // for				
			} // for
		} // for
		
		scaleSize/=2;
	} // while

	searchRange=5;
	tmpMotionVectors=motionVectors;
	fullSearchScaled (src1,src2,tmpMotionVectors, motionVectors);	
	searchRange=oldSearchRange;
	return;
}


void
BlockMotionFeature::processFullSearch (Image<uint8>& src1, Image<uint8>& src2, ieee64 &mx, ieee64 &my)
{
	// get size of source images
    ImageSize srcSize = src1.getSize();

	// search only one block (the full image)
	int32 searchArea = 2*searchRange+1;
	int32 imgX = srcSize[DIM_X] - searchRange;
	int32 imgY = srcSize[DIM_Y] - searchRange;
	Signal<int32> dist (SignalSize(256));
	ieee64 optDiff = 10e10;
	int32 optX=0;
	int32 optY=0;
	for (int32 sy=-searchRange; sy<=searchRange; sy++) {
		for (int32 sx=-searchRange; sx<=searchRange; sx++) {

			// init histogram
			dist.clear(0);
			int cnt=0;

			// get difference
			if (srcSize[DIM_BAND]==1) {
				for (int32 y=searchRange; y<imgY; y++) {
					for (int32 x=searchRange; x<imgX; x++) {
						int32 d0=abs(src1(x,y) - src2(sx+x, sy+y));
						dist(d0)++;
						cnt++;
					} // for
				} // for
			} else {
				for (int32 y=searchRange; y<imgY; y++) {
					for (int32 x=searchRange; x<imgX; x++) {
						int32 d0=abs(src1(x,y,0) - src2(sx+x, sy+y,0));
						int32 d1=abs(src1(x,y,1) - src2(sx+x, sy+y,1));
						int32 d2=abs(src1(x,y,2) - src2(sx+x, sy+y,2));
						d0=max(max(d0,d1),d2);
						dist(d0)++;
						cnt++;
					} // for
				} // for
			} // else


			// calculate difference betweeen both frames
			ieee64 maxCnt = 0.9*cnt;
			ieee64 diff=0;
			ieee64 c=0;
			for (int32 i=0; i<256; i++) {
				diff+=i*dist(i);
				c+=dist(i);
				if (c>maxCnt) break;
			} // for
			if (c>0) diff/=c;
			if (optDiff>diff) {
				optDiff=diff;
				optX=sx;
				optY=sy;
			} // if

		} // for
	} // for
	mx=optX;
	my=optY;

	return;
}

void
BlockMotionFeature::drawMotionVectors (const Image<int8>& motionVectors, Image<uint8>& srcDest)
{
	int32 offset = blockSize/2;
	ImageSize size = motionVectors.getSize();
	DrawLine line (Color::BLACK);
	for (int y=0; y<size[DIM_Y]; y++) {
		for (int x=0; x<size[DIM_X]; x++) {
			int32 offX = x*blockSize +offset;
			int32 offY = y*blockSize +offset;
			ImagePosition p0 (offX, offY);
			ImagePosition p1 (offX + motionVectors(x,y,0), offY + motionVectors(x,y,1));
			line.process(p0,p1,srcDest);
		} // for
	} // for
	return;
}
