/*
    CameraBackgroundFeature.cpp     calculate a background (mosaic) image of a shot
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



// MoCA includes
#include <CameraBackgroundFeature.h>
const string debugPath = "D:/debug/";


CameraBackgroundFeature::
CameraBackgroundFeature(const VideoSize &srcVidSize, 
			int32 bgSizeX, int32 bgSizeY, int32 bgShiftX, int32 bgShiftY, 
			bool calculateObjects, bool precise, bool estimateObjectLocation, 
				DefaultObjectPosition objPos)
{
	this->calculateObjects = calculateObjects;
	this->precise = precise;
	this->estimateObjectLocation = estimateObjectLocation;

	defaultObjPos = objPos;
	defaultPx0 = -1;
	defaultPx1 = -1;
	defaultPy0 = -1;
	defaultPy1 = -1;
	if (defaultObjPos != POSITION_UNKNOWN) {
		int32 w = srcVidSize[DIM_X]; 
		int32 h = srcVidSize[DIM_Y]; 

		switch (defaultObjPos) {
		case POSITION_CENTER:
			defaultPx0 = w/4;
			defaultPx1 = (3*w)/4;
			defaultPy0 = h/4;
			defaultPy1 = (3*h)/4;
			break;
		case POSITION_LEFT:
			defaultPx0 = 0;
			defaultPx1 = w/4;
			defaultPy0 = 0;
			defaultPy1 = h;
			break;
		case POSITION_RIGHT:
			defaultPx0 = (3*w)/4;
			defaultPx1 = w;
			defaultPy0 = 0;
			defaultPy1 = h;
			break;
		case POSITION_TOP:
			defaultPx0 = 0;
			defaultPx1 = w;
			defaultPy0 = 0;
			defaultPy1 = h/4;
			break;
		case POSITION_BOTTOM:
			defaultPx0 = 0;
			defaultPx1 = w;
			defaultPy0 = (3*h)/4;
			defaultPy1 = h;
			break;
		default:
			cerr << "error in CameraBackgroundFeature: unknown default image position!" << endl;
			break;
		} // switch
	} // if


	this->bgSizeX  = bgSizeX;
	this->bgSizeY  = bgSizeY;
	this->bgShiftX = bgShiftX;
	this->bgShiftY = bgShiftY;

	this->srcVidSize           = srcVidSize;
	this->srcVidSize[DIM_BAND] = 1;

	srcImgSize[DIM_X]      = srcVidSize[DIM_X];
	srcImgSize[DIM_Y]      = srcVidSize[DIM_Y];
	srcImgSize[DIM_Z]      = 1;
	srcImgSize[DIM_BAND]   = 1;

} // CameraBackgroundFeature


void 
CameraBackgroundFeature::findBgVidBorders (const Video<uint8> &bgVid, int32 &minX, int32 &maxX, int32 &minY, int32 &maxY)
{
	VideoSize size = bgVid.getSize();

	// initialize minX,...	
	minX = -20;
	minY = -20;
	maxX = size[DIM_X]-1+20;
	maxY = size[DIM_Y]-1+20;

	bool borderValid=false;
	while ((minY<size[DIM_Y]-20-1) && (!borderValid)) {
		minY+=20;
		for (int32 x=0; x<size[DIM_X]; x++) {
			if (bgVid(x,minY,0)!=255) {
				borderValid=true;
				break;
			} // if
		} // for
	} // while

	borderValid=false;
	while ((minX<size[DIM_X]-20-1) && (!borderValid)) {
		minX+=20;
		for (int32 y=0; y<size[DIM_Y]; y++) {
			if (bgVid(minX,y,0)!=255) {
				borderValid=true;
				break;
			} // if
		} // for
	} // while

	borderValid=false;
	while ((maxY>20) && (!borderValid)) {
		maxY-=20;
		for (int32 x=0; x<size[DIM_X]; x++) {
			if (bgVid(x,maxY,0)!=255) {
				borderValid=true;
				break;
			} // if
		} // for
	} // while

	borderValid=false;
	while ((maxX>20) && (!borderValid)) {
		maxX-=20;
		for (int32 y=0; y<size[DIM_Y]; y++) {
			if (bgVid(maxX,y,0)!=255) {
				borderValid=true;
				break;
			} // if
		} // for
	} // while




	// analyse all frames of bgVid
	for (int32 z=0; z<size[DIM_Z]; z++) {
		
		borderValid=false;
		while ((minY>0) && (!borderValid)) {
			borderValid=true;
			for (int32 x=0; x<size[DIM_X]; x++) {
				if (bgVid(x,minY,z)!=255) {
					borderValid=false;
					minY--;
					while ((minY>0) && (bgVid(x,minY,z)!=255)) {
						minY--;
					} // while
					break;
				} // if
			} // for
		} // while
		
		borderValid=false;
		while ((maxY<size[DIM_Y]-1) && (!borderValid)) {
			borderValid=true;
			for (int32 x=0; x<size[DIM_X]; x++) {
				if (bgVid(x,maxY,z)!=255) {
					borderValid=false;
					maxY++;
					while ((maxY<size[DIM_Y]-1) && (bgVid(x,maxY,z)!=255)) {
						maxY++;
					} // while
					break;
				} // if
			} // for
		} // while
		
		borderValid=false;
		while ((minX>0) && (!borderValid)) {
			borderValid=true;
			for (int32 y=0; y<size[DIM_Y]; y++) {
				if (bgVid(minX,y,z)!=255) {
					borderValid=false;
					minX--;
					while ((minX>0) && (bgVid(minX,y,z)!=255)) {
						minX--;
					} // while
					break;
				} // if
			} // for
		} // while
		
		borderValid=false;
		while ((maxX<size[DIM_X]-1) && (!borderValid)) {
			borderValid=true;
			for (int32 y=0; y<size[DIM_Y]; y++) {
				if (bgVid(maxX,y,z)!=255) {
					borderValid=false;
					maxX++;
					while ((maxX<size[DIM_X]-1) && (bgVid(maxX,y,z)!=255)) {
						maxX++;
					} // while
					break;
				} // if
			} // for
		} // while
	} // for


	// the borders match exactly
	minX++;
	maxX--;
	minY++;
	maxY--;

	return;
} // findBgVidBorders


void 
CameraBackgroundFeature::findBgImgBorders (Image<uint8> &bgImg, int32 &minX, int32 &maxX, int32 &minY, int32 &maxY)
{
	ImageSize size = bgImg.getSize();

	// initialize minX,...	
	minX = -20;
	minY = -20;
	maxX = size[DIM_X]-1+20;
	maxY = size[DIM_Y]-1+20;

	bool borderValid=false;
	while ((minY<size[DIM_Y]-20-1) && (!borderValid)) {
		minY+=20;
		for (int32 x=0; x<size[DIM_X]; x++) {
			if (bgImg(x,minY)!=255) {
				borderValid=true;
				break;
			} // if
		} // for
	} // while

	borderValid=false;
	while ((minX<size[DIM_X]-20-1) && (!borderValid)) {
		minX+=20;
		for (int32 y=0; y<size[DIM_Y]; y++) {
			if (bgImg(minX,y)!=255) {
				borderValid=true;
				break;
			} // if
		} // for
	} // while

	borderValid=false;
	while ((maxY>20) && (!borderValid)) {
		maxY-=20;
		for (int32 x=0; x<size[DIM_X]; x++) {
			if (bgImg(x,maxY)!=255) {
				borderValid=true;
				break;
			} // if
		} // for
	} // while

	borderValid=false;
	while ((maxX>20) && (!borderValid)) {
		maxX-=20;
		for (int32 y=0; y<size[DIM_Y]; y++) {
			if (bgImg(maxX,y)!=255) {
				borderValid=true;
				break;
			} // if
		} // for
	} // while




	// analyse bgImg
	borderValid=false;
	while ((minY>0) && (!borderValid)) {
		borderValid=true;
		for (int32 x=0; x<size[DIM_X]; x++) {
			if (bgImg(x,minY)!=255) {
				borderValid=false;
				minY--;
				while ((minY>0) && (bgImg(x,minY)!=255)) {
					minY--;
				} // while
				break;
			} // if
		} // for
	} // while
	
	borderValid=false;
	while ((maxY<size[DIM_Y]-1) && (!borderValid)) {
		borderValid=true;
		for (int32 x=0; x<size[DIM_X]; x++) {
			if (bgImg(x,maxY)!=255) {
				borderValid=false;
				maxY++;
				while ((maxY<size[DIM_Y]-1) && (bgImg(x,maxY)!=255)) {
					maxY++;
				} // while
				break;
			} // if
		} // for
	} // while
	
	borderValid=false;
	while ((minX>0) && (!borderValid)) {
		borderValid=true;
		for (int32 y=0; y<size[DIM_Y]; y++) {
			if (bgImg(minX,y)!=255) {
				borderValid=false;
				minX--;
				while ((minX>0) && (bgImg(minX,y)!=255)) {
					minX--;
				} // while
				break;
			} // if
		} // for
	} // while
	
	borderValid=false;
	while ((maxX<size[DIM_X]-1) && (!borderValid)) {
		borderValid=true;
		for (int32 y=0; y<size[DIM_Y]; y++) {
			if (bgImg(maxX,y)!=255) {
				borderValid=false;
				maxX++;
				while ((maxX<size[DIM_X]-1) && (bgImg(maxX,y)!=255)) {
					maxX++;
				} // while
				break;
			} // if
		} // for
	} // while


	// the borders match exactly
	minX++;
	maxX--;
	minY++;
	maxY--;

	return;
} // findBgImgBorders



void 
CameraBackgroundFeature::bgMedian(int32 lastFrameToAnalyse, Video<uint8> &bgVid, Image<uint8> &backgrd)
{

	
	VideoSize bgVidSize = bgVid.getSize(); 	
	ImageSize bgImgSize (bgVidSize[DIM_X], bgVidSize[DIM_Y], 1);
	backgrd.resize(bgImgSize);
	backgrd.clear(255);


	// get crop parameters of background pixel of bgVid(x,y,0)
	int32 minX, minY, maxX, maxY;
	findBgVidBorders (bgVid, minX, maxX, minY, maxY);


	// calculate the median
	vector<uint8> pixel;
	for (int32 y=minY; y<=maxY; y++) {
		for (int32 x=minX; x<=maxX; x++) {
			pixel.clear();
			for (int32 z=0; z<lastFrameToAnalyse; z++) {
				uint8 p = bgVid(x,y,z,0);
				if (p!=255) {
					pixel.push_back(p);
				} // if
			} // for
			
			// get median 
			int32 n = pixel.size();
			if (n>0) {
				sort(pixel.begin(), pixel.end());
				backgrd(x,y) = pixel[n/2];
			} // if
			
		} // for
	} // for

	return;
} // bgMedian


void 
CameraBackgroundFeature::
getCorners (const Image<uint8> &img, vector<CornerPosition>& cornersFloat, vector<CornerPositionInt>& cornersInt) 
{
	int32 nCorners = 0;
	int32 cornerTh = 100;
	int32 it=0;
	int32 w = img.getSize()[DIM_X];
	int32 h = img.getSize()[DIM_Y];
	CornerDetection cornerDetect (cornerTh, true);

	while (nCorners<50 || nCorners>300) {
		it++;
		if (it>6) break;

		cornersFloat.clear();
		cornerDetect.setThreshold(cornerTh);
		cornerDetect.process (img, cornersFloat);
		cornerDetect.filterCorners (cornersFloat,7,7,w-1-8,h-1-8);
		if (defaultObjPos != POSITION_UNKNOWN) {
			cornerDetect.removeCornersRect (cornersFloat, defaultPx0,defaultPy0,defaultPx1,defaultPy1);
		} // if

		nCorners = cornersFloat.size();
		if (nCorners <100) {
			cornerTh -=10;
		} else {
			cornerTh +=10;
		} // if
	} // while

	cornerDetect.cornersToInt (cornersFloat, cornersInt);
	return;
}




bool  
CameraBackgroundFeature::
getTransformation (const Video<uint8> &vid, Video<uint8> &tmpBgVid, vector<TransformationData> &vidTrans)
{
	bool error = false;
	VideoSize bgVidSize (bgSizeX,bgSizeY,vid.getSize()[DIM_Z],1);

	
	// initialize 
	PerspectiveTransformation mosaictransform;
	Image<uint8> last, curr;
	Image<uint8> tmp;
	tmpBgVid.resize (bgVidSize); 
	ImageSize bgImgSize (bgVidSize[DIM_X], bgVidSize[DIM_Y], 1);
	int32 w = vid.getSize()[DIM_X];
	int32 h = vid.getSize()[DIM_Y];
	VideoSize vidSize = vid.getSize();
	vidTrans.clear();


	// (1) calc transformation from last to curr 
	// (2) calc the transformed bg
	try {
		
		// read first image
		curr = vid.getImage(0);
		
		// initialize position of first frame in background image
		TransformationData trans, transI;
		trans.tx = bgShiftX;
		trans.ty = bgShiftY;
		
		TransformationData::processInverse(trans, transI);
		mosaictransform.normalizeParameters(transI);
		mosaictransform.TransformBitmap_LinInterpol (curr, tmp, bgImgSize, transI);
		tmpBgVid.setImage(0,tmp);
		
		
		// store transformation
		vidTrans.push_back(transI);
		
		
		// detect corners
		vector<CornerPosition> lastcorners,currcorners;
		vector<CornerPositionInt> lastcorners_int,currcorners_int;
		getCorners (curr, currcorners, currcorners_int);
	
		
		// process all frames
		for (int32 frameNo=1; frameNo<vidSize[DIM_Z]; frameNo++) {
			
			cerr << "\ranalyse background: " << frameNo << "/" << (vidSize[DIM_Z]-1) << "      " << ends ;
			
			last = curr;
			lastcorners = currcorners;
			lastcorners_int = currcorners_int;
			
			// get next frame
			curr = vid.getImage(frameNo);
			
			// calculate camera parameters of next two frames 
			CameraModelItem item;
			CameraModel cm (false, false, defaultObjPos);
			cm.getCorners (curr, currcorners, currcorners_int);
			cm.getCameraModelItem (currcorners, currcorners_int, lastcorners, lastcorners_int, curr, last, item);


			TransformationData transTmp = item.transData;
			mosaictransform.normalizeParameters(transTmp);
			
			
			// calc inverse transformation
			TransformationData::processInverse(transTmp, transI);
			mosaictransform.normalizeParameters(transI);
			
			// apply the new transformation to trans
			TransformationData::process(transI, trans);
			mosaictransform.normalizeParameters(trans);
			
			
			// transform curr to background with precise parameters
			TransformationData::processInverse(trans, transI);
			mosaictransform.normalizeParameters(transI);
			mosaictransform.TransformBitmap_LinInterpol(curr, tmp, bgImgSize, transI);
			tmpBgVid.setImage(frameNo,tmp);
			
			
			// store transformation
			vidTrans.push_back(transI);
			
		} // for
		
	} catch (...) {
		error = true;
		cerr << "CameraBackgroundFeature::getTrandformation: unknown excption, no data calculated!" << endl;
		//exit(1);
	} // catch
	cerr << "\ranalyse background...done        " << endl;
	

	return error;
} // getTrandformation 


#if 0
bool  
CameraBackgroundFeature::
getPreciseTransformation (const Video<uint8> &vid, Video<uint8> &maskVid1, 
						  Video<uint8> &tmpBgVid, Image<uint8> &bgImg, vector<TransformationData> &vidTrans)
{
	bool error = false;
	VideoSize bgVidSize (bgSizeX,bgSizeY,vid.getSize()[DIM_Z],1);
	PerspectiveTransformation mosaictransform;

	SignalSize minErrorSize (bgVidSize[DIM_Z]);
	Signal<ieee64> minError(minErrorSize);
	for (int32 frameNo=0; frameNo<bgVidSize[DIM_Z]; frameNo++) {
		minError(frameNo) = 0.0;
	} // for

	// transform maskVid1 to transMaskVid
	Video<uint8> transMaskVid (vid.getSize());
	{
		ImageSize tmpSize (vid.getSize()[DIM_X], vid.getSize()[DIM_Y], 1); 
		vector<TransformationData>::iterator it;
		int32 frameNo = 0;
		Image<uint8> tmp;
		for (it=vidTrans.begin(); it!=vidTrans.end(); it++) {

			// transform frame to background
			TransformationData trans = (*it);
			TransformationData transI; 
			TransformationData::processInverse(trans, transI);
			mosaictransform.TransformBitmap_LinInterpol(maskVid1.getImage(frameNo), tmp, tmpSize, transI);
			transMaskVid.setImage(frameNo, tmp);	

#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
			cout << "debug" << endl;
			cout << "vidTrans: " << frameNo << endl;
			cout << trans << endl;
#endif


#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
			cout << "debug" << endl;
			WriteImage (transMaskVid.getImage(frameNo), FileHandling::getNumberedFilename (debugPath + "transMaskVid", frameNo, ".pgm"));
#endif

			frameNo++;
		} // for
	}
	
	// initialize 
	Image<uint8> last, curr;
	Image<uint8> tmp;
	tmpBgVid.resize (bgVidSize); 
	ImageSize bgImgSize (bgVidSize[DIM_X], bgVidSize[DIM_Y], 1);
	int32 w = vid.getSize()[DIM_X];
	int32 h = vid.getSize()[DIM_Y];
	VideoSize vidSize = vid.getSize();
	vidTrans.clear();


	// (1) calc transformation from last to curr 
	// (2) calc the transformed bg
	try {
		
		// read first image
		curr = vid.getImage(0);
		
		// initialize position of first frame in background image
		TransformationData trans, transI;
		trans.tx = bgShiftX;
		trans.ty = bgShiftY;
		
		TransformationData::processInverse(trans, transI);
		mosaictransform.normalizeParameters(transI);
		mosaictransform.TransformBitmap_LinInterpol (curr, tmp, bgImgSize, transI);
		Image<uint8> storedBgImg = tmp;
		Image<uint8> storedMaskVid = maskVid1.getImage(0);
		
		//tmpBgVid.setImage(0,tmp);
		
		// initialize tmpBgVid and maskVid1 with background image
		tmpBgVid.setImage(0,bgImg);
		maskVid1.setImage(0, bgImg);


#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		cout << "debug" << endl;
		WriteImage (maskVid1.getImage(0), FileHandling::getNumberedFilename (debugPath+"precise-maskVid", 0, ".pgm"));
#endif
		
		// store transformation
		vidTrans.push_back(transI);
		
		
		// detect corners
		vector<CornerPosition> lastcorners,currcorners;
		vector<CornerPositionInt> lastcorners_int,currcorners_int;
		getCorners (curr, currcorners, currcorners_int);


#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		cout << "debug" << endl;
		{
			Image<uint8> disp = curr;
			DrawCross cross (10, Color::WHITE);
			for (int32 i=0;i<currcorners_int.size();i++)
			{
				int32 x = currcorners_int[i].x;
				int32 y = currcorners_int[i].y;
				
				cross.process (ImagePosition (x,y), disp);
			}
			cerr << "debug" << endl;
			WriteImage (disp, FileHandling::getNumberedFilename (debugPath+"corners", 0, ".pgm"));
		}

		{
			cerr << "debug" << endl;
			WriteImage (transMaskVid.getImage(0), FileHandling::getNumberedFilename (debugPath+"mask", 0, ".pgm"));
		}
#endif		

		//crndetect.filterCornersInt (currcorners_int, transMaskVid.getImage(0));
		cerr << "TODO: CameraBackgroundFeature::getPreciseTransformation" << endl;


#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		cout << "debug" << endl;
		{
			Image<uint8> disp = curr;
			DrawCross cross (10, Color::WHITE);
			for (int32 i=0;i<currcorners_int.size();i++)
			{
				int32 x = currcorners_int[i].x;
				int32 y = currcorners_int[i].y;
				
				cross.process (ImagePosition (x,y), disp);
			}
			cerr << "debug" << endl;
			WriteImage (disp, FileHandling::getNumberedFilename (debugPath+"corners-removed", 0, ".pgm"));
		}
#endif		

		// process all frames
		for (int32 frameNo=1; frameNo<vidSize[DIM_Z]; frameNo++) {
			
			cerr << "\rcreate background panorama: " << frameNo << "/" << (vidSize[DIM_Z]-1) << "      " << ends ;

			if (frameNo ==  vidSize[DIM_Z]/2) {
				tmpBgVid.setImage(0,storedBgImg);

				// copy tmp to maskVid(frameNo), if the mask is not background
				for (int32 y=0; y<bgImgSize[DIM_Y]; y++) {
					for (int32 x=0; x<bgImgSize[DIM_X]; x++) {
						if (storedMaskVid(x,y,0) == 255) {
							maskVid1(x,y,0,0) = 255;
						} else {
							maskVid1(x,y,0,0) = storedBgImg (x,y,0);
						} // else
					} // for
				} // for

#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
				cout << "debug" << endl;
				WriteImage (maskVid1.getImage(0), FileHandling::getNumberedFilename (debugPath+"precise-maskVid-replaced", 0, ".pgm"));
				WriteImage (tmpBgVid.getImage(0), FileHandling::getNumberedFilename (debugPath+"precise-tmpBgVid-replaced", 0, ".pgm"));
#endif
			} // if

			
			last = curr;
			lastcorners = currcorners;
			lastcorners_int = currcorners_int;
			
			// get next frame
			curr = vid.getImage(frameNo);
			
			// calculate the motion model for 2 frames
			CameraModelItem item;
			CameraModel cm (false, false, defaultObjPos);
			cm.calcFramesWithMask (item, curr, last, 
							       currcorners, currcorners_int, 
							       transMaskVid.getImage(frameNo), 
							       lastcorners, lastcorners_int);


			TransformationData transTmp = item.transData;
			/*
			transTmp.a[0][0] = item.a00;
			transTmp.a[0][1] = item.a01;
			transTmp.a[1][0] = item.a10;
			transTmp.a[1][1] = item.a11;
			transTmp.tx = item.tx;
			transTmp.ty = item.ty;
			transTmp.px = item.px;
			transTmp.py = item.py;
			*/
			mosaictransform.normalizeParameters(transTmp);
			
			
			// calc inverse transformation
			TransformationData::processInverse(transTmp, transI);
			mosaictransform.normalizeParameters(transI);
			
			// apply the new transformation to trans
			TransformationData::process(transI, trans);
			mosaictransform.normalizeParameters(trans);
			


			// calculate exact transformation
			//cout << trans << endl;
			CameraTransformationPrecise precise;

#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
			cout << "debug" << endl;
			WriteImage (bgImg, FileHandling::getNumberedFilename (debugPath+"precise---bgImg", frameNo, ".pgm"));
			WriteImage (curr, FileHandling::getNumberedFilename (debugPath+"precise---curr", frameNo, ".pgm"));
			WriteImage (transMaskVid.getImage(frameNo), FileHandling::getNumberedFilename (debugPath+"precise---mask", frameNo, ".pgm"));
#endif
			precise.process (bgImg, curr, transMaskVid.getImage(frameNo), trans, minError(frameNo));
			//cout << "minError: " << minError(frameNo) << endl;
			//cout << trans << endl;
			mosaictransform.normalizeParameters(trans);

			
			// transform curr to background with precise parameters
			TransformationData::processInverse(trans, transI);
			mosaictransform.normalizeParameters(transI);
			mosaictransform.TransformBitmap_LinInterpol(curr, tmp, bgImgSize, transI);
			tmpBgVid.setImage(frameNo,tmp);
			

#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
			cout << "debug" << endl;
			WriteImage (tmp, FileHandling::getNumberedFilename (debugPath+"precise-bg", frameNo, ".pgm"));
#endif		

			// copy tmp to maskVid(frameNo), if the mask is not background
			for (int32 y=0; y<bgImgSize[DIM_Y]; y++) {
				for (int32 x=0; x<bgImgSize[DIM_X]; x++) {
					if (maskVid1(x,y,frameNo,0) != 255) {
						maskVid1(x,y,frameNo,0) = tmp (x,y,0);
					} // if
				} // for
			} // for

			// update bgImg (median)
			bgMedian(frameNo+1, maskVid1, bgImg);

#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
			cout << "debug" << endl;
			WriteImage (bgImg, "D:/precise-bg.png");
#endif		

	


			// store transformation
			vidTrans.push_back(transI);
			
		} // for
		
	} catch (...) {
		error = true;
		cerr << "CameraBackgroundFeature::getTrandformation: unknown excption, no data calculated!" << endl;
		//exit(1);
	} // catch
	cerr << "\rcreate background panorama...done        " << endl;


	// copy bgVid(0) to bgImg, if (bgVid == 255)
	for (int32 y=0; y<bgVidSize[DIM_Y]; y++) {
		for (int32 x=0; x<bgVidSize[DIM_X]; x++) {
			if (bgImg(x,y) == 255) {
				bgImg(x,y) = tmpBgVid(x,y,0);
			} // if
		} // for
	} // for


#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
			cout << "debug" << endl;
			string debugFn = debugPath+"precise-bgVid.yuv";
			cout << tmpBgVid.getSize()[DIM_X] << "  " << tmpBgVid.getSize()[DIM_Y] << endl;
			WriteYVideo (tmpBgVid, debugFn);
			WriteYVideo (tmpBgVid, debugPath+"rel-bg-img.yuv");
#endif

	return error;
} // getPreciseTransformation 
#endif



bool 
CameraBackgroundFeature::
cropBackground (const Video<uint8> &tmpBgVid, Video<uint8> &bgVid, vector<TransformationData> &vidTrans)
{
	bool error = false;
	try {

		// (3) crop transformed bg
		// get crop parameters of background pixel of bgVid(x,y,0)
		int32 newMinX, newMinY, newMaxX, newMaxY;
		findBgVidBorders (tmpBgVid, newMinX, newMaxX, newMinY, newMaxY);
		newMinX-=4;
		newMinY-=4;
		newMaxX+=4;
		newMaxY+=4;
		
		int32 newWidth = newMaxX-newMinX+1;
		newWidth = 4*((1+newWidth)/4);
		int32 newHeight = newMaxY-newMinY+1;
		newHeight= 4*((1+newHeight)/4);
		
		VideoSize cropBgSize (newWidth, newHeight, tmpBgVid.getSize()[DIM_Z]);
		bgVid.resize(cropBgSize);
		for (int32 z=0; z<cropBgSize[DIM_Z]; z++) {
			for (int32 y=0; y<cropBgSize[DIM_Y]; y++) {
				for (int32 x=0; x<cropBgSize[DIM_X]; x++) {
					bgVid(x,y,z) = tmpBgVid (x+newMinX,y+newMinY,z);
				} // for
			} // for
		} // for
		bgSizeX = newWidth;
		bgSizeY = newHeight;
		bgShiftX = bgShiftX - newMinX;
		bgShiftY = bgShiftY - newMinY;
		
		
		// apply new shift position to transformation 
		TransformationData newShiftTrans;
		newShiftTrans.tx = newMinX;
		newShiftTrans.ty = newMinY;
		
		PerspectiveTransformation mosaictransform;
		vector<TransformationData>::iterator it;
		for (it=vidTrans.begin(); it!=vidTrans.end(); it++) {
			
			//cout << (*it) << endl;
			
			// apply newShiftTrans to (*it)
			TransformationData::process(newShiftTrans, (*it));
			
			//cout << (*it) << endl;
		} // for

	} catch (...) {
		error = true;
		cerr << "CameraBackgroundFeature::cropBackground: unknown excption, no data calculated!" << endl;
		//exit(1);
	} // catch

	return error;
} // cropBackground



bool 
CameraBackgroundFeature::
getBackgroundImage (const Video<uint8> &bgVid, Video<uint8> &maskVid, Image<uint8> &bgImg, bool estimateObjectLocation)
{
	
	bool error = false;
	VideoSize size = bgVid.getSize();

	SignalSize boxDataSize (size[DIM_Z], 4); // minX, maxX, minY, maxY
	Signal<int32> boxData (boxDataSize);
	boxData.clear(0);
	
	
	if (estimateObjectLocation) {
		
		// calc diff in transformed bg (tmpBgVid)
		// calc difference of 2 following frames,
		// if diff > th, than object (=no background), set value top 255 (undefined)
		try {
			maskVid=bgVid;
			
			// get motion pixel
			vector<TransformationPointInt32> bgMotionPixelListCurr;
			vector<TransformationPointInt32> bgMotionPixelListNext;
			bgMotionPixelListCurr.clear();
			bgMotionPixelListNext.clear();
			
			// store rectangle (motion area)
			SignalSize bgMotionDataSize (size[DIM_Z],5); // avgX, avgY, varX, varY, numberOfPixel 
			Signal<int32> bgMotionData (bgMotionDataSize); 
			for (int32 i=0; i<bgMotionDataSize[DIM_X]; i++) {
				bgMotionData (i,0) = 0;
				bgMotionData (i,1) = 0;
				bgMotionData (i,2) = 0;
				bgMotionData (i,3) = 0;
				bgMotionData (i,4) = 0;
			} // for
			
			// build difference image and locate position of the object 
			// (we assume that only one object exists)
			
			const int thDiff = 32;
			for (int32 z=0; z<size[DIM_Z]; z++) {
				
				bgMotionPixelListCurr = bgMotionPixelListNext;
				bgMotionPixelListNext.clear();
				
				// compare with next frame
				if (z<size[DIM_Z]-1) { // last iteration (z==cropBgSize[DIM_Z]-1) calculates avg and var
					for (int32 y=0; y<size[DIM_Y]; y++) {
						for (int32 x=0; x<size[DIM_X]; x++) {
							uint8 p1 = bgVid(x,y,z+1);
							uint8 p2 = bgVid(x,y,z);
							if ((p1!=255) && (p2!=255)) {
								if (abs(((int)p1)-p2) > thDiff) {
									// if the difference is very large, the pixel 
									// should not be a background pixel
									
									bgMotionData (z,0) += x;
									bgMotionData (z,1) += y;
									bgMotionData (z,4)++;
									
									bgMotionData (z+1,0) += x;
									bgMotionData (z+1,1) += y;
									bgMotionData (z+1,4)++;
									
									bgMotionPixelListCurr.push_back(TransformationPointInt32 (x,y));
									bgMotionPixelListNext.push_back(TransformationPointInt32 (x,y));
									
									
									maskVid(x,  y,  z)   = 255;
									maskVid(x,  y,  z+1) = 255;
								} // if
							} // if
						} // for
					} // for
				} // if
				
				// calc avg 
				if (bgMotionData (z,4) > 0) {
					bgMotionData (z,0) /= bgMotionData (z,4);  // avgX
					bgMotionData (z,1) /= bgMotionData (z,4);  // avgY
				} // if
				
				// calc variance
				vector<TransformationPointInt32>::iterator it;
				int32 avgX = bgMotionData (z,0);
				int32 avgY = bgMotionData (z,1);
				int32 sigX = 0;
				int32 sigY = 0;
				for (it=bgMotionPixelListCurr.begin(); it!=bgMotionPixelListCurr.end(); it++) {
					TransformationPointInt32 pt = (*it);
					sigX += (pt.x-avgX) * (pt.x-avgX);
					sigY += (pt.y-avgY) * (pt.y-avgY);
				} // for
				if (bgMotionPixelListCurr.size()>0) {
					sigX /= bgMotionPixelListCurr.size();  // variance
					sigY /= bgMotionPixelListCurr.size();  // variance
				} // if
				if (sigX>0) {
					sigX = (int32) (sqrt((double)sigX) + 0.5);
				} // if
				if (sigY>0) {
					sigY = (int32) (sqrt((double)sigY) + 0.5);
				} // if
				
				
				// find all motion pixel within this area
				int32 pMinX = (int32)(avgX - 2.0*sigX);
				int32 pMaxX = (int32)(avgX + 2.0*sigX);
				int32 pMinY = (int32)(avgY - 2.0*sigY);
				int32 pMaxY = (int32)(avgY + 2.0*sigY);
				
				// set limits of pmax,...
				if (pMinX<0) pMinX=0;
				if (pMinY<0) pMinY=0;
				if (pMaxX>size[DIM_X]-1) pMaxX=size[DIM_X]-1;
				if (pMaxY>size[DIM_Y]-1) pMaxY=size[DIM_Y]-1;
				while ((pMaxX-pMinX+1) > 0.75*size[DIM_X]) {
					pMinX++;
					pMaxX--;
				} // while
				while ((pMaxY-pMinY+1) > 0.75*size[DIM_Y]) {
					pMinY++;
					pMaxY--;
				} // while
				
				
				  /*
				  cerr << z << ": " 
				  << pMinX << " " 
				  << pMaxX<< " " 
				  << pMinY<< " " 
				  << pMaxY<< " " 
				  << endl;	
				*/
				
				
				// second iteration
				avgX = 0;
				avgY = 0;
				int32 n = 0;
				for (it=bgMotionPixelListCurr.begin(); it!=bgMotionPixelListCurr.end(); it++) {
					TransformationPointInt32 pt = (*it);
					if ((pt.x>=pMinX) && (pt.x<=pMaxX) && (pt.y>=pMinY) && (pt.y<=pMaxY)) {
						avgX+=pt.x;
						avgY+=pt.y;
						n++;
					} // if
				} // for
				if (n>0) {
					avgX/=n;
					avgY/=n;
				} // if
				
				sigX = 0;
				sigY = 0;
				for (it=bgMotionPixelListCurr.begin(); it!=bgMotionPixelListCurr.end(); it++) {
					TransformationPointInt32 pt = (*it);
					if ((pt.x>=pMinX) && (pt.x<=pMaxX) && (pt.y>=pMinY) && (pt.y<=pMaxY)) {
						sigX+=(pt.x-avgX) * (pt.x-avgX);
						sigY+=(pt.y-avgY) * (pt.y-avgY);
					} // if
				} // for
				if (n>0) {
					sigX/=n;
					sigY/=n;
				} // if
				if (sigX>0) {
					sigX = (int32) (sqrt((double)sigX) + 0.5);
				} // if
				if (sigY>0) {
					sigY = (int32) (sqrt((double)sigY) + 0.5);
				} // if
				
				
				pMinX = (int32)(avgX - 1.75*sigX);
				pMaxX = (int32)(avgX + 1.75*sigX);
				pMinY = (int32)(avgY - 1.75*sigY);
				pMaxY = (int32)(avgY + 1.75*sigY);
				
				// set limits of pmax,...
				if (pMinX<0) pMinX=0;
				if (pMinY<0) pMinY=0;
				if (pMaxX>size[DIM_X]-1) pMaxX=size[DIM_X]-1;
				if (pMaxY>size[DIM_Y]-1) pMaxY=size[DIM_Y]-1;
				while ((pMaxX-pMinX+1) > 0.70*size[DIM_X]) {
					pMinX++;
					pMaxX--;
				} // while
				while ((pMaxY-pMinY+1) > 0.70*size[DIM_Y]) {
					pMinY++;
					pMaxY--;
				} // while
				
				
				  /*
				  cerr << z << ": " 
				  << pMinX << " " 
				  << pMaxX<< " " 
				  << pMinY<< " " 
				  << pMaxY<< " " 
				  << endl;
				*/
				
				
				// third iteration
				avgX = 0;
				avgY = 0;
				n = 0;
				for (it=bgMotionPixelListCurr.begin(); it!=bgMotionPixelListCurr.end(); it++) {
					TransformationPointInt32 pt = (*it);
					if ((pt.x>=pMinX) && (pt.x<=pMaxX) && (pt.y>=pMinY) && (pt.y<=pMaxY)) {
						avgX+=pt.x;
						avgY+=pt.y;
						n++;
					} // if
				} // for
				if (n>0) {
					avgX/=n;
					avgY/=n;
				} // if
				
				sigX = 0;
				sigY = 0;
				for (it=bgMotionPixelListCurr.begin(); it!=bgMotionPixelListCurr.end(); it++) {
					TransformationPointInt32 pt = (*it);
					if ((pt.x>=pMinX) && (pt.x<=pMaxX) && (pt.y>=pMinY) && (pt.y<=pMaxY)) {
						sigX+=(pt.x-avgX) * (pt.x-avgX);
						sigY+=(pt.y-avgY) * (pt.y-avgY);
					} // if
				} // for
				if (n>0) {
					sigX/=n;
					sigY/=n;
				} // if
				if (sigX>0) {
					sigX = (int32) (sqrt((double)sigX) + 0.5);
				} // if
				if (sigY>0) {
					sigY = (int32) (sqrt((double)sigY) + 0.5);
				} // if
				
				
				pMinX = (int32)(avgX - 1.75*sigX);
				pMaxX = (int32)(avgX + 1.75*sigX);
				pMinY = (int32)(avgY - 1.75*sigY);
				pMaxY = (int32)(avgY + 1.75*sigY);
				
				
				if (pMinX<0) pMinX=0;
				if (pMinY<0) pMinY=0;
				if (pMaxX>size[DIM_X]-1) pMaxX=size[DIM_X]-1;
				if (pMaxY>size[DIM_Y]-1) pMaxY=size[DIM_Y]-1;
				while ((pMaxX-pMinX+1) > size[DIM_X]/2)	{
					pMinX++;
					pMaxX--;
				} // while
				while ((pMaxY-pMinY+1) > size[DIM_Y]/2)	{
					pMinY++;
					pMaxY--;
				} // while
				
				  /*
				  cerr << z << ": " 
				  << pMinX << " " 
				  << pMaxX<< " " 
				  << pMinY<< " " 
				  << pMaxY<< " " 
				  << endl;
				*/
				boxData (z,0) = pMinX;
				boxData (z,1) = pMaxX;
				boxData (z,2) = pMinY;
				boxData (z,3) = pMaxY;
				
			} // for
		
		
			// smooth position of rectangles
			Signal<int32> tmp = boxData;
			for (int32 z=1; z<size[DIM_Z]-1; z++) {
				boxData(z,0) = (tmp (z-1,0) + 2*tmp (z,0) + tmp (z+1,0))/4;
				boxData(z,1) = (tmp (z-1,1) + 2*tmp (z,1) + tmp (z+1,1))/4;
				boxData(z,2) = (tmp (z-1,2) + 2*tmp (z,2) + tmp (z+1,2))/4;
				boxData(z,3) = (tmp (z-1,3) + 2*tmp (z,3) + tmp (z+1,3))/4;
			} // for
			
#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
			cout << "debug" << endl;
			for (int32 z=0; z<size[DIM_Z]; z++) {
				Image<uint8> img = bgVid.getImage(z);
				DrawRectangle rect (Color::BLACK);
				rect.process(ImagePosition(boxData(z,0),boxData(z,2)), ImagePosition(boxData(z,1),boxData(z,3)),img);
				WriteImage (img, FileHandling::getNumberedFilename ("D:/debug/object_region_no_bg_", z, ".png"));
			} // for
#endif
			
		} catch (...) {
			error = true;
			cerr << "CameraBackgroundFeature::getBackgroundImage: unknown excption, no data calculated!" << endl;
			return error;
		} // catch
	} // estimateObjectLocation
	




	try {
		// copy orig bg vid-frame to maskVid and set area to background
		// increase the size of the rectangle by factor 2
		maskVid = bgVid;
		ImageSize tmpBgImgSize (size[DIM_X],size[DIM_Y],1);
		Image<uint8> tmpBgModified (tmpBgImgSize);
		tmpBgModified.clear(0);
		for (int32 z=0; z<size[DIM_Z]; z++) {
			int32 pMinX = boxData (z,0);
			int32 pMaxX = boxData (z,1);
			int32 pMinY = boxData (z,2);
			int32 pMaxY = boxData (z,3);
			
			// increase the size of the box
			int32 cx = (pMinX+pMaxX)/2;
			int32 cy = (pMinY+pMaxY)/2;
			int32 dx = cx-pMinX;
			int32 dy = cy-pMinY;
			pMinX -= dx;
			pMaxX += dx;
			pMinY -= dy;
			pMaxY += dy;
			if (pMinX<0) pMinX=0;
			if (pMinY<0) pMinY=0;
			if (pMaxX>size[DIM_X]-1) pMaxX=size[DIM_X]-1;
			if (pMaxY>size[DIM_Y]-1) pMaxY=size[DIM_Y]-1;
			
			// copy values
			for (int32 y=pMinY; y<=pMaxY; y++) {
				for (int32 x=pMinX; x<=pMaxX; x++) {
					maskVid (x,y,z,0) = 255;   // set to background
					tmpBgModified (x,y,0) = 255; // remember modified regions
				} // for
			} // for
		} // for


		// apply default information about object position
		if (defaultObjPos != POSITION_UNKNOWN) {
			// copy values
			for (int32 z=0; z<size[DIM_Z]; z++) {
				for (int32 y=defaultPy0; y<=defaultPy1; y++) {
					for (int32 x=defaultPx0; x<=defaultPx1; x++) {
						maskVid (x,y,z,0) = 255;   // set to background
						tmpBgModified (x,y,0) = 255; // remember modified regions
					} // for
				} // for
			} // for
		} // if


		
#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		{
			cout << "debug" << endl;
			WriteImage (tmpBgModified,"D:/debug/bg_modified_a.png");
			
			cout << "debug" << endl;
			for (int32 z=0; z<size[DIM_Z]; z++) {
				WriteImage (maskVid.getImage(z),FileHandling::getNumberedFilename("D:/debug/vid_no_bg_a_",z,".png"));
			} // for
		}
#endif
		
		
		// select a pixel for each undefined one 
		SignalSize distSize (size[DIM_Z]);
		Signal<int32> maxDist (distSize);
		for (int32 y=0; y<size[DIM_Y]; y++) {
			for (int32 x=0; x<size[DIM_X]; x++) {
				if (tmpBgModified (x,y,0) == 255) {
					
					// we have to select several fitting bg pixels
					maxDist.clear(0);
					for (int32 z=0; z<size[DIM_Z]; z++) {
						int32 pMinX = boxData (z,0);
						int32 pMaxX = boxData (z,1);
						int32 pMinY = boxData (z,2);
						int32 pMaxY = boxData (z,3);
						int32 cx = (pMinX+pMaxX)/2;
						int32 cy = (pMinY+pMaxY)/2;
						ieee64 dist = (cx-x)*(cx-x) + (cy-y)*(cy-y);
						maxDist(z) = (int32)dist;
					} // for
					
					// select threshold
					Signal<int32> tmp = maxDist;
					int32 medianTh;
					MedianFilter<int32> mf (1);
					mf.process (tmp,medianTh);
					
					// select larger values
					for (int32 z=0; z<size[DIM_Z]; z++) {
						if (maxDist (z) > medianTh) {
							
							// set pixel in maskVid
							maskVid (x,y,z,0) = bgVid (x,y,z,0);   
						} // if
					} // for
					
				} // if
			} // for
		} // for
		
		
		
		
		
#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		cout << "debug" << endl;
		for (int32 z=0; z<size[DIM_Z]; z++) {
			WriteImage (maskVid.getImage(z),FileHandling::getNumberedFilename("D:/debug/vid_no_bg_b_",z,".png"));
		} // for
		//exit(1);
#endif
		
		
		
		// calculate background (median)
		bgMedian(size[DIM_Z], maskVid, bgImg);
		
		// copy bgVid(0) to bgImg, if (bgVid == 255)
		for (int32 y=0; y<size[DIM_Y]; y++) {
			for (int32 x=0; x<size[DIM_X]; x++) {
				if (bgImg(x,y) == 255) {
					bgImg(x,y) = bgVid(x,y,0);
				} // if
			} // for
		} // for
		
#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		{
			cout << "debug" << endl;
			string s = "D:/debug/vid_mask-bg.png";
			WriteImage (bgImg, s);
		}
#endif
		
	} catch (...) {
		error = true;
		cerr << "CameraBackgroundFeature::getBackgroundImage: unknown excption, no data calculated!" << endl;
		//exit(1);
	} // catch
	
	
	return error;
} // getBackgroundImage




// (1) calc transformation between 2 following frames (last and curr)
// (2) calc the transformed background
// (3) crop transformed background
// (4) if precise: crop precise transformation and background
// (5) if calculateobjects: calc objects (diff between transformed background and frame)
//
// parameters:
//    vid          : source video
//    motionMaskVid: mask of the video with objects (object is white)
//    bgImg        : background image
void 
CameraBackgroundFeature::
process (const Video<uint8> &vid, Video<uint8> &motionMaskVid, Image<uint8> &bgImg, vector<TransformationData> &vidTrans)
{
	
	Video<uint8> bgVid;
	vidTrans.clear();

	bool error = false;
	{
		// calc transformation and transform frames to background (tmpBgVid)
		Video<uint8> tmpBgVid; 
		error = getTransformation (vid, tmpBgVid, vidTrans);
		if (error) return;
		
		// crop background
		error = cropBackground (tmpBgVid, bgVid, vidTrans);
		if (error) return;
	}


#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		cerr << endl << endl << "DEBUG" << endl;
		{
			// store background video
			for (int32 i=0; i<bgVid.getSize()[DIM_Z]; i++) {
				string debugFn = FileHandling::getNumberedFilename ("D:/debug/1_bg_vid__",i,".png"); 
				cout << "store: " << debugFn << endl;
				WriteImage (bgVid.getImage(i),debugFn);
			}// for
		}
		cerr  << "DEBUG-END" << endl << endl;
#endif

	
	{
		// calc background image
		error = getBackgroundImage(bgVid, motionMaskVid, bgImg, estimateObjectLocation);
		if (error) return;
	}


#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
	cerr << "DEBUG" << endl;
	{
		// store video
		for (int32 i=0; i<motionMaskVid.getSize()[DIM_Z]; i++) {
			string debugFn = FileHandling::getNumberedFilename ("D:/debug/4_motionMaskVid_",i,".png"); 
			cout << "store: " << debugFn << endl;
			WriteImage (motionMaskVid.getImage(i),debugFn);
		}// for
	}
	{
		string debugFn = "D:/debug/5_bgImg.png"; 
		cout << "store: " << debugFn << endl;
		WriteImage (bgImg, debugFn);
	}
#endif
	
	

#if 0
	// get precise calculation
	// (1) get Transformation by comparing bgVid and bgImg
	if (precise) {
		// calc transformation and transform frames to background (tmpBgVid)
		//Video<uint8> tmpBgVid; 
		//error = getPreciseTransformation (vid, motionMaskVid, tmpBgVid, bgImg, vidTrans);
		error = getPreciseTransformation (vid, motionMaskVid, bgVid, bgImg, vidTrans);
		if (error) return;
		

#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		cout << "debug" << endl;
		{
			WriteImage (bgImg, debugPath+"bgImg-precise.pgm");
			string debugFn = debugPath+"bgVid-precise.yuv";
			cout << bgVid.getSize()[DIM_X] << "  " << bgVid.getSize()[DIM_Y] << endl;
			WriteYVideo (bgVid, debugFn);
		}
#endif
	} // if
#endif

	// calculate objects (difference of bgVid and bgImg)
	if (calculateObjects) {
		calcObjects (vid, bgVid, bgImg, vidTrans, motionMaskVid);
	} // if

	return;
}

void optimizeBorder (const Image<uint8> &lumi, Image<uint8> &mask, int maxStep, uint8 maxDiff)
{

	ImageSize srcImgSize = lumi.getSize();

	// calc exact borders (lower border)
	for (int32 x=0; x<srcImgSize[DIM_X]; x++) {
		for (int32 y=0; y<srcImgSize[DIM_Y]; y++) {
			if ((mask(x,y) == 0) && (mask(x,y-1) == 255)) {
				
				int32 px = x;
				int32 py = y;
				int dLumi = abs(lumi(px,py) - lumi(px,py-1));
				int32 maxPy = py;
				int32 maxLumi= dLumi;
				while ((py>0) && (dLumi<maxDiff) && (py-y<maxStep)) {
					py--;
					dLumi = abs(lumi(px,py) - lumi(px,py-1));
					if (maxLumi<=dLumi) {
						maxLumi=dLumi;
						maxPy = py;
					} // if
				} // while
				
				for (int32 iy=py; iy<y;iy++) {
					mask(x,iy)=0;
				} // for
				
				//if (dLumi < 5) {
				//	mask(x,y-1) = 0;
				//} // if
			} // if
		} // for
	} // for
	// calc exact borders (upper border)
	for (int32 x=0; x<srcImgSize[DIM_X]; x++) {
		for (int32 y=0; y<srcImgSize[DIM_Y]; y++) {
			if ((mask(x,y) == 0) && (mask(x,y+1) == 255)) {
				
				int32 px = x;
				int32 py = y;
				int dLumi = abs(lumi(px,py) - lumi(px,py+1));
				int32 maxPy = py;
				int32 maxLumi= dLumi;
				while ((py<srcImgSize[DIM_Y]-1) && (dLumi<maxDiff) && (y-py<maxStep)) {
					py++;
					dLumi = abs(lumi(px,py) - lumi(px,py+1));
					if (maxLumi<=dLumi) {
						maxLumi=dLumi;
						maxPy = py;
					} // if
				} // while
				
				for (int32 iy=y+1; iy<=py;iy++) {
					mask(x,iy)=0;
				} // for
			} // if
		} // for
	} // for
	// calc exact borders (right border)
	for (int32 y=0; y<srcImgSize[DIM_Y]; y++) {
		for (int32 x=0; x<srcImgSize[DIM_X]; x++) {
			if ((mask(x,y) == 0) && (mask(x-1,y) == 255)) {
				
				int32 px = x;
				int32 py = y;
				int dLumi = abs(lumi(px,py) - lumi(px-1,py));
				int32 maxPx = px;
				int32 maxLumi= dLumi;
				while ((px>0) && (dLumi<maxDiff) && (px-x<maxStep)) {
					px--;
					dLumi = abs(lumi(px,py) - lumi(px-1,py));
					if (maxLumi<=dLumi) {
						maxLumi=dLumi;
						maxPx = px;
					} // if
				} // while
				
				for (int32 ix=px; ix<x;ix++) {
					mask(ix,y)=0;
				} // for
			} // if
		} // for
	} // for
	// calc exact borders (left border)
	for (int32 y=0; y<srcImgSize[DIM_Y]; y++) {
		for (int32 x=0; x<srcImgSize[DIM_X]; x++) {
			if ((mask(x,y) == 0) && (mask(x+1,y) == 255)) {
				
				int32 px = x;
				int32 py = y;
				int dLumi = abs(lumi(px,py) - lumi(px+1,py));
				int32 maxPx = px;
				int32 maxLumi= dLumi;
				while ((px<srcImgSize[DIM_X]-1) && (dLumi<maxDiff) && (x-px<maxStep)) {
					px++;
					dLumi = abs(lumi(px,py) - lumi(px+1,py));
					if (maxLumi<=dLumi) {
						maxLumi=dLumi;
						maxPx = px;
					} // if
				} // while
				
				for (int32 ix=x+1; ix<=px;ix++) {
					mask(ix,y)=0;
				} // for
			} // if
		} // for
	} // for

	return;
}



void 
CameraBackgroundFeature::
calcObjects (const Video<uint8> &vid, const Video<uint8> &bgVid, const Image<uint8> &bgImg, const vector<TransformationData> &vidTrans, Video<uint8> &objMaskVid)
{
	PerspectiveTransformation mosaictransform;
	Image<uint8> disp;
	Image<uint8> dispTrans;

	objMaskVid.resize(srcVidSize);

	int32 frame=0;
	vector<TransformationData>::const_iterator it;
	for (it=vidTrans.begin(); it!=vidTrans.end(); it++) {

		cerr << "\rget objects: " << frame << "/" << (srcVidSize[DIM_Z]-1) << ends;

		// calculate difference image
		DifferenceFilter diffFilter;
		diffFilter.process (bgVid.getImage(frame), bgImg, disp);

#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		cout << "debug" << endl;
		//WriteImage (disp, FileHandling::getNumberedFilename (debugPath + "diff_", frame, ".pgm"));
		WriteImage (bgVid.getImage(frame), FileHandling::getNumberedFilename (debugPath + "bgVid_", frame, ".pgm"));
		WriteImage (bgImg, debugPath + "bgImg__.png");
#endif
	
		// if background, do not calc diff
		for (int32 y=0; y<srcImgSize[DIM_Y]; y++) {
			for (int32 x=0; x<srcImgSize[DIM_X]; x++) {
				if ((bgVid(x,y,frame,0) == 255) || (bgImg(x,y,0) == 255)) {
					disp (x,y,0) = 128;
				} // if
			} // for
		} // for

#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		//cout << "debug" << endl;
		//WriteImage (disp, FileHandling::getNumberedFilename (debugPath + "diff-with-bg_", frame, ".pgm"));
#endif


		// transform frame to background
		TransformationData trans = (*it);
		TransformationData transI; 
		TransformationData::processInverse(trans, transI);
		mosaictransform.TransformBitmap_LinInterpol(disp, dispTrans, srcImgSize, transI);
		//objMaskVid.setImage(frame, dispTrans);	

#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		//cout << "debug" << endl;
		//WriteImage (dispTrans, FileHandling::getNumberedFilename (debugPath + "diff/diff1_", frame, ".pgm"));
#endif

		// set border to 128
		for (int32 x=0; x<srcImgSize[DIM_X]; x++) {
			dispTrans(x,0) = 128;
			dispTrans(x,1) = 128;
			dispTrans(x,srcImgSize[DIM_Y]-2) = 128;
			dispTrans(x,srcImgSize[DIM_Y]-1) = 128;
		} // for
		for (int32 y=2; y<srcImgSize[DIM_Y]-2; y++) {
			dispTrans(0,y) = 128;
			dispTrans(1,y) = 128;
			dispTrans(srcImgSize[DIM_X]-2,y) = 128;
			dispTrans(srcImgSize[DIM_X]-1,y) = 128;
		} // for

#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		//cout << "debug" << endl;
		//WriteImage (dispTrans, FileHandling::getNumberedFilename (debugPath + "diff/diff2_", frame, ".pgm"));
#endif


		// markov
		Image<uint8> mask;
		MarkovFilter markov;
		markov.process(vid.getImage(frame), dispTrans, mask);

#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		cout << "debug" << endl;
		WriteImage (vid.getImage(frame), FileHandling::getNumberedFilename (debugPath + "vid_", frame, ".pgm"));
#endif

		// set border to 0
		for (int32 x=0; x<srcImgSize[DIM_X]; x++) {
			mask(x,0) = 0;
			mask(x,srcImgSize[DIM_Y]-1) = 0;
		} // for
		for (int32 y=1; y<srcImgSize[DIM_Y]-1; y++) {
			mask(0,y) = 0;
			mask(srcImgSize[DIM_X]-1,y) = 0;
		} // for


#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		/*
		cout << "debug" << endl;
		WriteImage (mask, FileHandling::getNumberedFilename (debugPath + "mask/mask_", frame, ".pgm"));

		disp = vid.getImage(frame);
		for (int32 y=0; y<srcImgSize[DIM_Y]; y++) {
			for (int32 x=0; x<srcImgSize[DIM_X]; x++) {
				int p = disp(x,y);
				if (mask(x,y) == 255) {
					p += 112;
					if (p<180) p=180;
					if (p>255) p=255;
				} else {
					p -= 64;
					if (p>172) p=172;
					if (p<0) p=0;
				} // else
				disp(x,y) = (uint8)p;
			} // for
		} // for
		WriteImage (disp, FileHandling::getNumberedFilename (debugPath + "object/object_", frame, ".pgm"));
		*/
#endif


		// calc exact borders (first iteration)
		int maxStep=15;
		uint8 maxDiff = 4;
		optimizeBorder(vid.getImage(frame), mask, maxStep, maxDiff);

#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		//cout << "debug" << endl;
		//WriteImage (mask, FileHandling::getNumberedFilename (debugPath + "mask/mask-borders1_", frame, ".pgm"));
#endif

		// calc exact borders (second iteration)
		maxStep=10;
		maxDiff = 3;
		optimizeBorder(vid.getImage(frame), mask, maxStep, maxDiff);

#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		//cout << "debug" << endl;
		//WriteImage (mask, FileHandling::getNumberedFilename (debugPath + "mask/mask-borders2_", frame, ".pgm"));
#endif

		// smooth border of object
		{
			int masksize=3;
			ClosingFilter<uint8> closeF (masksize);
			closeF.process(mask);

#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		//cout << "debug" << endl;
		//WriteImage (mask, FileHandling::getNumberedFilename (debugPath + "mask/mask-smooth1_", frame, ".pgm"));
#endif

			OpeningFilter<uint8> openF (masksize);
			openF.process(mask);

#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		//cout << "debug" << endl;
		//WriteImage (mask, FileHandling::getNumberedFilename (debugPath + "mask/mask-smooth2_", frame, ".pgm"));
#endif
		}


#ifdef DEBUG_CAMERA_BACKGROUND_FEATURE
		/*
		cout << "debug" << endl;
		WriteImage (mask, FileHandling::getNumberedFilename (debugPath + "mask/mask-smoothed_", frame, ".pgm"));

		disp = vid.getImage(frame);
		for (int32 y=0; y<srcImgSize[DIM_Y]; y++) {
			for (int32 x=0; x<srcImgSize[DIM_X]; x++) {
				int p = disp(x,y);
				if (mask(x,y) == 255) {
					p += 112;
					if (p<180) p=180;
					if (p>255) p=255;
				} else {
					p -= 64;
					if (p>172) p=172;
					if (p<0) p=0;
				} // else
				disp(x,y) = (uint8)p;
			} // for
		} // for
		WriteImage (disp, FileHandling::getNumberedFilename (debugPath + "object/object-smoothed_", frame, ".pgm"));
		*/
#endif

		// store mask
		objMaskVid.setImage(frame, mask);
		frame++;
	} // for
	//cerr << endl;
	
	return;
} // calcObjects


// store images and object mask video
void 
CameraBackgroundFeature::
storeData (const Video<uint8> &motionMaskVid, const Image<uint8> &bgImg, string &fnPrefix)
{
	// store bg image
	string fn = fnPrefix + "_bg.png";
	WriteImage (bgImg, fn);

	// store bg object mask
	int32 n = motionMaskVid.getSize()[DIM_Z];
	for (int32 i=0; i<n; i++) {
		fn = fnPrefix + "_mask_";
		fn = FileHandling::getNumberedFilename(fn, i, ".png");
		WriteImage (motionMaskVid.getImage(i), fn);
	}

	return;
}


// add the objects in the object mask to the background image
void 
CameraBackgroundFeature::
addObject (const Image<uint8> &img, const Image<uint8> &mask, Image<uint8> &bgImg, TransformationData &trans, string &fnPrefix)
{
	// transform img and mask
	ImageSize destSize = bgImg.getSize();
	Image<uint8> imgTrans (destSize);
	Image<uint8> maskTrans (destSize);
	PerspectiveTransformation perspTrans;
	perspTrans.processImg (trans, img, imgTrans);
	perspTrans.processImg (trans, mask, maskTrans);

	// store original and transformed images
	string fn;
	//fn = fnPrefix + "img.png";   WriteImage (img, fn);
	//fn = fnPrefix + "mask.png";  WriteImage (mask, fn);
	//fn = fnPrefix + "imgT.png";  WriteImage (imgTrans, fn);
	//fn = fnPrefix + "maskT.png"; WriteImage (maskTrans, fn);


	// copy object
	ImageSize size = img.getSize();
	Image<uint8> object = img;
	for (int32 y=0; y<size[DIM_Y]; y++) {
		for (int32 x=0; x<size[DIM_X]; x++) {
			if (mask(x,y) != 255) {
				object (x,y) = 255;
			} // if
		} // for
	} // for

	// transform object
	Image<uint8> objectTrans (destSize); 
	perspTrans.processImg (trans, object, objectTrans);
	fn = "object_" + fnPrefix + ".png";   WriteImage (object, fn);
	fn = "objectT_" + fnPrefix + ".png";  WriteImage (objectTrans, fn);

	return;
}


