/*
    CcrFeature.cpp     calculate the edge change ratio (CCR)
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// MoCA includes
#include <CcrFeature.h>



std::ostream& operator<< (ostream& os, const CcrData &data)
{
	os << "version        : " << data.version          << endl;
	os << "filename       : " << data.filename         << endl;
	os << "ccr size       : " << data.size             << endl;
	os << "minimum value  : " << data.minCcrValue      << endl;
	os << "maximum value  : " << data.maxCcrValue      << endl;
	os << "average value  : " << data.avgCcrValue      << endl;
	os << "ccr file       : " << data.ccrFilename      << endl;
	os << "pin file       : " << data.pinFilename      << endl;
	os << "pout file      : " << data.poutFilename     << endl;
	os << "moca path      : " << data.mocaPath         << endl;
	os << endl;

	return os;
} // operator<<

OTagStream& operator<< (OTagStream& os, const CcrData& data)
{
    TagFormat::writeNodeValue(os, "version",     data.version);
    TagFormat::writeNodeValue(os, "filename",    data.filename);
    TagFormat::writeNodeValue(os, "size",     data.size);
    TagFormat::writeNodeValue(os, "minCcrValue", data.minCcrValue);
    TagFormat::writeNodeValue(os, "maxCcrValue", data.maxCcrValue);
    TagFormat::writeNodeValue(os, "avgCcrValue", data.avgCcrValue);
    TagFormat::writeNodeValue(os, "ccrFilename", data.ccrFilename);
    TagFormat::writeNodeValue(os, "pinFilename", data.pinFilename);
    TagFormat::writeNodeValue(os, "poutFilename",data.poutFilename);
	os << endl;

	// write ccr file
	{
		OTagStream ots (data.mocaPath+data.ccrFilename);
		TagFormat::writeNodeValue(ots, data.ccr);
		ots.close();
	}

	// write ccr file
	{
		OTagStream ots (data.mocaPath+data.pinFilename);
		TagFormat::writeNodeValue(ots, data.pin);
		ots.close();
	}

	// write ccr file
	{
		OTagStream ots (data.mocaPath+data.poutFilename);
		TagFormat::writeNodeValue(ots, data.pout);
		ots.close();
	}

	return os;
} // operator<<

ITagStream& operator>> (ITagStream& is, CcrData& data)
{
    char c;
    bool failFlag = false;

	if (is >> c) {
		is.putback(c);

		TagFormat::readNodeValue(is, "version",     data.version);
		TagFormat::readNodeValue(is, "filename",    data.filename);
		int32 tmp = 0;
		TagFormat::readNodeValue(is, "size", tmp);
		data.size = tmp;

		TagFormat::readNodeValue(is, "minCcrValue", data.minCcrValue);
		TagFormat::readNodeValue(is, "maxCcrValue", data.maxCcrValue);
		TagFormat::readNodeValue(is, "avgCcrValue", data.avgCcrValue);
		TagFormat::readNodeValue(is, "ccrFilename", data.ccrFilename);
		TagFormat::readNodeValue(is, "pinFilename", data.pinFilename);
		TagFormat::readNodeValue(is, "poutFilename", data.poutFilename);
		if (data.ccrFilename == "") {
			failFlag = true;
		} // if
		if (is.rdstate() & ios::failbit) {
			failFlag = true;
		} // if
	} // if

	data.ccr.clear();
	data.pin.clear();
	data.pout.clear();
	SignalSize size (data.size);
    if (failFlag) {
		throw MocaException ("CcrFeature: Illegal file format!");
	} else {
        is.clear();

		// read hist file
		try {
			{
				ITagStream its (data.mocaPath+data.ccrFilename);
				TagFormat::readNodeValue(its, data.ccr, size);
				its.close();
			}
			{
				ITagStream its (data.mocaPath+data.pinFilename);
				TagFormat::readNodeValue(its, data.pin, size);
				its.close();
			}
			{
				ITagStream its (data.mocaPath+data.poutFilename);
				TagFormat::readNodeValue(its, data.pout, size);
				its.close();
			}
		} catch (...) {
			failFlag = true;
		} // try

		if (data.ccr.getSize()[DIM_X] <= 0) {
			failFlag = true;
		} // if
		if (data.pin.getSize()[DIM_X] <= 0) {
			failFlag = true;
		} // if
		if (data.pout.getSize()[DIM_X] <= 0) {
			failFlag = true;
		} // if
	    if (failFlag) {
			string s = "CcrFeature: Illegal data file format (" + data.mocaPath+data.ccrFilename + "," + 
				        data.mocaPath+data.pinFilename + " or " + data.mocaPath+data.poutFilename + ")!";
			throw MocaException (s);
		} // if
    } // else

    return is;
} // operator>> 



const string CcrFeature::DEFAULT_CCR_FEATURE_FILE_NAME = "ccr.idx";



void
CcrFeature::process (CcrData& data) 
{

	// read data
	ReadMocaData<CcrData> reader (pathFilename, CCR_FEATURE_FILENAME);
	bool dataValid = reader.process (CCR_FEATURE_VERSION, data); 

	if (!dataValid) {

		// create new index
		cerr << "CcrFeature: Creating index file ..." << endl;
		readValues(data);


		// save index file
		WriteMocaData<CcrData> writer (pathFilename, CCR_FEATURE_FILENAME);
		writer.process(data);
		cerr << "CcrFeature: Creating index file ...done" << endl;
	} // if

	return;
}



// calculate CCR difference values
void
CcrFeature::readValues (CcrData& data) 
{

	// open video
	VideoDecoder reader (pathFilename, CODEC_Y);
	size = reader.getSize();


    // set data
	data.size = size[DIM_Z];
    data.ccr.resize(data.size);
    data.pin.resize(data.size);
    data.pout.resize(data.size);

	data.avgCcrValue = 0;
	data.maxCcrValue = -1;
	data.minCcrValue = 10e10;


	// get filename without path, create new directory if necesarry
	data.version  = CCR_FEATURE_VERSION;
	CreateMocaPath pathReader (pathFilename, false);
	data.filename = pathReader.file;
	data.mocaPath = pathReader.destinationPath;
	data.ccrFilename = "ccr_data.idx";
	data.pinFilename = "ccr_in_data.idx";
	data.poutFilename = "ccr_out_data.idx";


	// clear data
    data.ccr.clear(0);
    data.pin.clear(0);
    data.pout.clear(0);


	// get camera parameters
	CameraModelData cameraParameters;
	{
		CameraModel reader(pathFilename);
		reader.process(cameraParameters);
	}


SignalSize dbgSize (size[DIM_Z]);
Signal<ieee64> dbg1(dbgSize);
Signal<ieee64> dbg2(dbgSize);
Signal<ieee64> dbg3(dbgSize);
Signal<ieee64> dbg4(dbgSize);
Signal<ieee64> dbg5(dbgSize);
Signal<ieee64> dbg6(dbgSize);
Signal<ieee64> dbg7(dbgSize);
Signal<ieee64> dbg8(dbgSize);

    // process camera model data
    for (uint32 i= 0; i < size[DIM_Z]-1; i++) {
		cerr << "\rframe: " << i << "            " << ends;

		int32 candidateMatches = cameraParameters.data[i].numberCornerMatchesAvailable;
		int32 realMatches = cameraParameters.data[i].numberCornerMatched;
		int32 currCorners = cameraParameters.data[i].currCorners;
		int32 lastCorners = cameraParameters.data[i].lastCorners;
		int32 maxVectorLength = cameraParameters.data[i].maxVectorLength;

		ieee64 v1, v2, v3, v4, v5, v6, v7, v8, v9, v10;
		v1 = (lastCorners - realMatches);
		v2 = (currCorners - realMatches);
		if (realMatches>0) {
			v1 /= realMatches;
			v2 /= realMatches;
		} // if
		v3 =  max (v1, v2);
		v4 = 100.0;  
		if (candidateMatches>0) {
			v4/= candidateMatches;
		} // if
		v5 = 100.0;  
		if (realMatches>0) {
			v5/= realMatches;
		} else {
			v5/= 0.1;
		} // else
		v6 = labs (currCorners - lastCorners);
		v7 = max(currCorners, lastCorners);  
		if (realMatches>0) {
			v7/= realMatches;
		} else {
			v7/= 0.1;
		} // else
		v8 = candidateMatches;  
		if (realMatches>0) {
			v8/= realMatches * realMatches;
		} else {
			v8/= 0.1;
		} // else

		v9 = (currCorners - realMatches);
		if (currCorners>0) {
			v9 /= currCorners;
		} // if
		v10 = (lastCorners - realMatches);
		if (lastCorners>0) {
			v10 /= lastCorners;
		} // if



		data.pout(i) = v1;
		data.pin(i)  = v2;
		data.ccr(i)  = v10;

		
dbg1(i) = v1;
dbg2(i) = v2;
dbg3(i) = v3;
dbg4(i) = v4;
dbg5(i) = v5;
dbg6(i) = v6;
dbg7(i) = v7;
dbg8(i) = v7;



		
		if (data.minCcrValue > data.ccr(i)) {
			data.minCcrValue = data.ccr(i);
		} // if
		if (data.maxCcrValue < data.ccr(i)) {
			data.maxCcrValue = data.ccr(i);
		} // if
		data.avgCcrValue += data.ccr(i);

    } // for
	data.avgCcrValue /= data.size;

#if 0
// debug
{
	string s = mocaPath + "dbg1.txt";
	OTagStream ots (s);
	TagFormat::writeNodeValue(ots, dbg1);
	ots.close();
}
{
	string s = mocaPath + "dbg2.txt";
	OTagStream ots (s);
	TagFormat::writeNodeValue(ots, dbg2);
	ots.close();
}
{
	string s = mocaPath + "dbg3.txt";
	OTagStream ots (s);
	TagFormat::writeNodeValue(ots, dbg3);
	ots.close();
}
{
	string s = mocaPath + "dbg4.txt";
	OTagStream ots (s);
	TagFormat::writeNodeValue(ots, dbg4);
	ots.close();
}
{
	string s = mocaPath + "dbg5.txt";
	OTagStream ots (s);
	TagFormat::writeNodeValue(ots, dbg5);
	ots.close();
}
{
	string s = mocaPath + "dbg6.txt";
	OTagStream ots (s);
	TagFormat::writeNodeValue(ots, dbg6);
	ots.close();
}
{
	string s = mocaPath + "dbg7.txt";
	OTagStream ots (s);
	TagFormat::writeNodeValue(ots, dbg7);
	ots.close();
	cerr << "debug files stored: " << s << endl;
}
cerr << "debug-end" << endl;
exit(1); 
#endif

	return;
} // process





