/*
    ContourEdgeList.cpp     extract edge pixels of a contour
    Copyright (C) 2004 Oliver Schuster

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <ContourEdgeList.h>

// convert to black (=object) and white (=background) image
void
ContourEdgeList::
convertImage (const Image<uint8> &img, Image<uint8> &dest)
{

	// convert image to grayscale and binarize image
	Image<uint8> tmp;
	GrayscaleFilter<uint8> gray;
	gray.process(img, tmp);
	BinarizeFilter<uint8> bf (128);
	bf.process(tmp);
	ImageSize size = tmp.getSize();


	// analyse image border to get the background color of the image (this is 0 or 255),
	int32 c0=0; // counts background pixels
	int32 c1=0; // counts foreground pixels
	int32 widthP = size[DIM_X]/12 + 2;
	int32 heightP = size[DIM_Y]/12 + 2;
	for (int32 y=0; y<size[DIM_Y]; y++) {
		for (int32 x=0; x<widthP; x++) {
			if (tmp(x,y) == background) c0++; else c1++;
			if (tmp(size[DIM_X]-1-x,y) == background) c0++; else c1++;
		} // for
	} // for
	for (int32 y=0; y<heightP; y++) {
		for (int32 x=0; x<size[DIM_X]; x++) {
			if (tmp(x,y) == background) c0++; else c1++;
			if (tmp(x, size[DIM_Y]-1-y) == background) c0++; else c1++;
		} // for
	} // for
	
	// background and object-color should be switched
	if (c0<c1) {
		InverseFilter invF;
		invF.process(tmp);
	} // if

#if 0
	// debug
	WriteImage (tmp, "D:/images/obj01.png");
#endif

	// add a small border
	size[DIM_X]+=shiftOffset+shiftOffset;
	size[DIM_Y]+=shiftOffset+shiftOffset;
	dest.resize(size);
	dest.clear(255);
	ImagePosition p (shiftOffset,shiftOffset);
	dest.copy (tmp, p);

#if 0
	// debug
	WriteImage (dest, "D:/images/obj02.png");
#endif


	return;
} // convertImage


ieee64
ContourEdgeList::
getLength (vector<ImageEdgePosition> &iep)
{
	ieee64 length = 0.0;
	int32 n = iep.size();
	ieee64 dx = ((iep[0])[DIM_X] - (iep[n-1])[DIM_X]);
	ieee64 dy = ((iep[0])[DIM_Y] - (iep[n-1])[DIM_Y]);
	dx = dx*dx + dy*dy;
	if (dx>0) length+= sqrt(dx);

	for (int32 i=0; i<n-1; i++) {
		dx = ((iep[i])[DIM_X] - (iep[i+1])[DIM_X]);
		dy = ((iep[i])[DIM_Y] - (iep[i+1])[DIM_Y]);
		dx = dx*dx + dy*dy;
		if (dx>0) length+= sqrt(dx);
	} // for

	return length;
}

void
ContourEdgeList::
process (const Image<uint8> &src, vector<ImageEdgePosition> &dest)
{
    dest.clear();
#if 0
	// debug
	WriteImage (src, "D:/images/obj00.png");
#endif


	// convert to black (=object) and white (=background) image
	Image<uint8> img;
	convertImage(src, img);

    bool pixelFound = markOuterEdges (img);

	if (pixelFound) {
		ImageSize size = img.getSize ();
		
		indexType startx = 0;
		indexType starty = 0;
		
		// assertion: there is at least one background pixel between border and object
		for (indexType x = 1; x < size[DIM_X] - 1 && startx == 0; x++) {
			for (indexType y = 1; y < size[DIM_Y] - 1 && startx == 0; y++) {
				if (img(x, y) != background) {
					startx = x;
					starty = y;
				} // if
			}
		} // for
		
		if (startx == 0 && starty == 0) {
			dest.clear();
			return;
		} // if
		
		ImageEdgePosition actPos(startx, starty, WEST);
		ImageEdgePosition startPos = actPos;
		
		do {
			ImageEdgePosition newPos;
			
			indexType x = actPos[DIM_X];
			indexType y = actPos[DIM_Y];
			
			switch (actPos.getEdgePosition()) {
			case NORTH:
				if (edgeImage(x+1, y-1).west) {
					newPos = ImageEdgePosition(x+1, y-1, WEST);
				} else if (edgeImage(x+1, y).north) {
					newPos = ImageEdgePosition(x+1, y, NORTH);
				} else if (edgeImage(x, y).east) {
					newPos = ImageEdgePosition(x, y, EAST);
				} else {
					// error
					dest.clear();
					return;
				} // if
				break;
			case SOUTH:
				if (edgeImage(x-1, y+1).east) {
					newPos = ImageEdgePosition(x-1, y+1, EAST);
				} else if (edgeImage(x-1, y).south) {
					newPos = ImageEdgePosition(x-1, y, SOUTH);
				} else if (edgeImage(x, y).west) {
					newPos = ImageEdgePosition(x, y, WEST);
				} else {
					// error
					dest.clear();
					return;
				} // if
				break;
			case EAST:
				if (edgeImage(x+1, y+1).north) {
					newPos = ImageEdgePosition(x+1, y+1, NORTH);
				} else if (edgeImage(x, y+1).east) {
					newPos = ImageEdgePosition(x, y+1, EAST);
				} else if (edgeImage(x, y).south) {
					newPos = ImageEdgePosition(x, y, SOUTH);
				} else {
					// error
					dest.clear();
					return;
				} // if
				break;
			case WEST:
				if (edgeImage(x-1, y-1).south) {
					newPos = ImageEdgePosition(x-1, y-1, SOUTH);
				} else if (edgeImage(x, y-1).west) {
					newPos = ImageEdgePosition(x, y-1, WEST);
				} else if (edgeImage(x, y).north) {
					newPos = ImageEdgePosition(x, y, NORTH);
				} else {
					// error
					dest.clear();
					return;
				} // if
				break;
			default:
				// error
				dest.clear();
				return;
				break;
			} // switch
			
			dest.push_back(actPos);
			actPos = newPos;
			
		} while (actPos != startPos);
	} // if

	// remove border (subract shiftOffset)
	for (int32 i=0; i<dest.size(); i++) {
		dest[i][DIM_X] -= shiftOffset;
		dest[i][DIM_Y] -= shiftOffset;
	} // for

} // process


bool
ContourEdgeList::
markOuterEdges (const Image<uint8>& src)
{
    ImageSize size = src.getSize ();
	bool pixelFound = false;

    // get max value of T
	indexType x;
	indexType y;
    edgeImage.resize ((ImageSize(size[DIM_X], size[DIM_Y])));
    for (y = 0; y < size[DIM_Y]; y++) {
        for (x = 0; x < size[DIM_X]; x++) {
            edgeImage(x, y).north = false;
            edgeImage(x, y).south = false;
            edgeImage(x, y).east  = false;
            edgeImage(x, y).west  = false;
        }
    }

    // assertion: there is at least one background pixel between border and object
    for (y = 1; y < size[DIM_Y] - 1; y++) {
        for (x = 1; x < size[DIM_X] - 1; x++) {
            if (src(x, y) != background) {
				pixelFound = true;
                // test all directions for neighboring background pixels
                edgeImage(x, y).north = (src(x, y - 1) == background);
                edgeImage(x, y).south = (src(x, y + 1) == background);
                edgeImage(x, y).east  = (src(x + 1, y) == background);
                edgeImage(x, y).west  = (src(x - 1, y) == background);
            }
        }
    }
	return pixelFound;
} // markOuterEdges


