/*
   This file is part of the MoCA video library.
   Name: ContourParametrization.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


/* -*- C++ -*-

   file.h  -- short description

   Author:    Stephan RIchter
              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim, Germany
              email: kopf@pi4.informatik.uni-mannheim.de

*/



#ifndef CONTOUR_PARAMETRISATION_H
#define CONTOUR_PARAMETRISATION_H

// C++ includes
#include <math.h>
#include <vector>

// MoCA includes
#include <MocaTypes.h>
#include <Image.h>
#include <DrawLine.h>
#include <DrawDot.h>
#include <ImagePosition.h>
#include <ImageEdgePosition.h>


/*!
  \brief parametizes a vector of contour pixels

  A vector of ImageEdgePositions (input) is transformed into a vector
  of ImagePositions which describe a conto7ur of an object.
 */

class ContourParametrization
{
public:
	//! the constructor
    ContourParametrization(uint16 numberContourPixel);

    //! destructor
    ~ContourParametrization();
    
    //! The actual process
    void process(vector<ImageEdgePosition> &src, vector<ImagePosition>& dest);

	void drawContourDots (const vector<ImagePosition>& data, Image<uint8> &img, uint8 bgColor=255)
	{
		// get required image size
		int32 maxX = 1;
		int32 maxY = 1;
		for (int32 i=0; i<data.size(); i++) {
			if (maxX < (data[i])[DIM_X])  maxX = (data[i])[DIM_X];
			if (maxY < (data[i])[DIM_Y])  maxY = (data[i])[DIM_Y];
		} // for
		maxX++; maxY++;

		// set image size
		ImageSize size = img.getSize();
		if (size[DIM_X] <= maxX) size[DIM_X] = maxX+1;
		if (size[DIM_Y] <= maxY) size[DIM_Y] = maxY+1;
		if (size[DIM_Z] < 1) size[DIM_Z] = 1;
		if (size[DIM_BAND] < 1) size[DIM_BAND] = 1;
		img.resize(size);

		// set bg color
		img.clear(bgColor);
		Color fgColor (255-bgColor);

		// set dots
		DrawDot dot;
		for (int32 i=0; i<data.size(); i++) {
			dot.process(data[i], fgColor, img);
		} // for
	}


	void drawContourLines (const vector<ImagePosition>& data, Image<uint8> &img, uint8 bgColor=255)
	{
		// get required image size
		int32 maxX = 1;
		int32 maxY = 1;
		for (int32 i=0; i<data.size(); i++) {
			if (maxX < (data[i])[DIM_X])  maxX = (data[i])[DIM_X];
			if (maxY < (data[i])[DIM_Y])  maxY = (data[i])[DIM_Y];
		} // for
		maxX++; maxY++;

		// set image size
		ImageSize size = img.getSize();
		if (size[DIM_X] <= maxX) size[DIM_X] = maxX+1;
		if (size[DIM_Y] <= maxY) size[DIM_Y] = maxY+1;
		if (size[DIM_Z] < 1) size[DIM_Z] = 1;
		if (size[DIM_BAND] < 1) size[DIM_BAND] = 1;
		img.resize(size);

		// set bg color
		img.clear(bgColor);
		Color fgColor (255-bgColor);

		// set dots
		DrawLine line;
		for (int32 i=0; i<data.size(); i++) {
			int32 j = (i+1) % (data.size());
				
			line.process(data[i],data[j], fgColor, img);
		} // for
	}



	const static uint16 DEFAULT_CONTOUR_SIZE;
protected:

private:

	//! calculate distance between 2 pixel
	ieee64 getDistance(ImageEdgePosition &p1, ImageEdgePosition &p2);
    //! number of contour pixels
    uint16 numberContourPixel;

}; // class ContourParametrization


#endif


