/*
   This file is part of the MoCA video library.
   Name: CornerDetection.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*
   Harris corner detector
   -----------------------------
   (c) 2002 Dirk Farin
   Dept. Computer Science IV
   University of Mannheim
 */

#ifndef CORNER_DETECTION_H
#define CORNER_DETECTION_H

//#include "libvideogfx/graphics/datatypes/image.hh"
//#include "libvideogfx/graphics/datatypes/primitives.hh"
//#include "libvideogfx/containers/dynarray.hh"
//using namespace videogfx;

// C++ includes
#include<vector>
using namespace std;

// MoCA includes
#include<Image.h>
#include<MocaTypes.h>
#include <OImageStream.h>
#include <DrawCross.h>

struct CornerPosition
{
    CornerPosition () { }
    CornerPosition (const ieee64& _x,const ieee64& _y) { x = _x; y = _y; }
	
    ieee64 x,y;
};

struct CornerPositionInt
{
    CornerPositionInt () { }
    CornerPositionInt (const int32& _x,const int32& _y) { x = _x; y = _y; }
	
    int32 x,y;
};

class CornerDetection
{
public:
	// cornerTh: low threshold detects more corners
	// useSignificantCorners: remove neighboring corners
	CornerDetection(int32 cornerTh = 100, bool useSignificantCorners=false)
	  : d_steps_input(1),
		d_steps_correlation(5),
		d_flat_threshold(cornerTh),
		useSignificantCorners(useSignificantCorners)
	{}
	
	void process (const Image<uint8>& img, vector<CornerPosition>& corners);
	void process (const Image<uint8>& img, vector<CornerPositionInt>& corners)
	{
		vector<CornerPosition> tmpCorners;
		process (img, tmpCorners);
		cornersToInt(tmpCorners, corners);
		return;
	} // process
	
	void setThreshold (int32 th)
	{
		d_flat_threshold = th;
		return;
	}

	void drawCorners (Image<uint8> &img, const vector<CornerPositionInt>& corners);
	void drawCorners (Image<uint8> &img, const vector<CornerPosition>& corners)
	{
		vector<CornerPositionInt> tmpCorners;
		cornersToInt(corners, tmpCorners);
		drawCorners (img, tmpCorners);
		return;
	}
	
	
	// remove corners outside the rectangle
	void filterCorners(vector<CornerPosition>& corners,int left,int top,int right,int bottom);
	void filterCornersInt(vector<CornerPositionInt>& corners,int left,int top,int right,int bottom);

	// remove corner, if (mask(x,y) == 255)
	void filterCornersInt(vector<CornerPositionInt>& corners, const Image<uint8> &mask);


	// remove corners within rectangle
	void removeCornersRect(vector<CornerPosition>& corners,int left,int top,int right,int bottom);
	void removeCornersRectInt(vector<CornerPositionInt>& corners,int left,int top,int right,int bottom);


	void cornersToInt(const vector<CornerPosition> &corners, vector<CornerPositionInt> &dest);


private:
  int32 d_steps_input;
  int32 d_steps_correlation;
  int32 d_flat_threshold;

  Image<uint8> smooth_input;
  Image<int16> xx,yy,xy;

  bool useSignificantCorners; // reduce the number of corners significantly
};





template<class T> 
static void CornerBiFilter(Image<T>& bm, int n)
{
	if (n==0)
		return;
	
	int width,height;
	ImageSize size = bm.getSize();
	width  = size[DIM_X];
	height = size[DIM_Y];
	
	T* buf1 = new T[max(width,height)];
	T* buf2 = new T[max(width,height)];
	T* bmPtr;
	
	for (int y=0;y<height;y++) {
		T* s=NULL;
		T* d=NULL;
		
		for (int i=0;i<n;i++) {
			if (i==0) {
				s = bm.getSamplePtr(0,y);
				d = buf1;
			} else {
				if ((i & 1) == 1) {
					s = buf1;
					d = buf2;
				} else {
					s = buf2;
					d = buf1;
				} // else
			} // else
			
			for (int x=1;x<width-1;x++)
				d[x] = (s[x-1]+s[x]+s[x]+s[x+1])>>2;
			
			d[0]=d[1];
			d[width-1]=d[width-2];
		} // for
		
		bmPtr = bm.getSamplePtr(0,y);
		for (int x=0;x<width;x++) {
			//bm(x,y) = d[x];
			*bmPtr = d[x];
			bmPtr++;
		} // for
			
    } // for

	for (int x=0;x<width;x++) {

		const offsetType* bmOffsetPtr = &(bm.getSampleOffsetY (0));
		bmPtr = bm.getSamplePtr(x,0);
		for (int y=0;y<height;y++) {
			//buf1[y] = bm (x,y);
			buf1[y] = *bmPtr;
	        bmPtr += *bmOffsetPtr;
		    bmOffsetPtr++;
		} // for
		
		T* s=NULL;
		T* d=NULL;
		
		for (int i=0;i<n;i++) {
			if ((i & 1) == 0) {
				s = buf1;
				d = buf2;
			} else {
				s = buf2;
				d = buf1;
			} // else
			
			for (int y=1;y<height;y++)
				d[y] = (s[y-1]+s[y]+s[y]+s[y+1])>>2;
			
			d[0]=d[1];
			d[height-1]=d[height-2];
		}
		
		bmOffsetPtr = &(bm.getSampleOffsetY (1));
		bmPtr = bm.getSamplePtr(x,1);
		for (int y=1;y<height-1;y++) {
			//bm(x,y) = d[y];
			*bmPtr = d[y];
	        bmPtr += *bmOffsetPtr;
		    bmOffsetPtr++;
		} // for
    } // for

	delete[] buf1;
	delete[] buf2;
	return;
}



#endif

