/*
   This file is part of the MoCA video library.
   Name: CssMatching.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#ifndef CSS_MATCHING_H
#define CSS_MATCHING_H

#include <vector>
#include <cmath>

#include <CssMaximaFeature.h>
#include <ContourParametrization.h>

/*!

 */
class CssMatching
{
public:
    //! the default constructor
    CssMatching(int cP, int thP, int thH, int thW, int thS, bool checkConcave, 
		        bool xyRatio, bool rotateCss, bool matchGlobalShapeDescriptors=false)  
		: contourPixels       (cP), 
		  thPosition          (ieee64(thP)/100), 
		  thHeight            (ieee64(thH)/100), 
		  thWidth             (ieee64(thW)/100),
		  thSignificantHeight (ieee64(thS)/100), 
		  optimizeConcave     (checkConcave),
		  xyRatio             (xyRatio),		  
		  rotateCss           (rotateCss),
	      matchGlobalShapeDescriptors (matchGlobalShapeDescriptors) {};

    //! the destructor
    ~CssMatching() {};

    //! main method to call
    int32 process(CssMaximumData &one, CssMaximumData &two)
    {
		int32 dist = 0;

		if (xyRatio) {
			ieee64 f=1.0;
			if (one.xyRatio <= 0.0) one.xyRatio = 1.0;
			if (two.xyRatio <= 0.0) two.xyRatio = 1.0;
			if (one.xyRatio>two.xyRatio) {
				f=one.xyRatio / two.xyRatio;
			} else {
				f=two.xyRatio / one.xyRatio;
			} // else
			if (f>1.5) {
				dist=INT_MAX;
			} // if
		} // if

		if (dist!= INT_MAX) {
			int32 first  = distance(one, two);
			int32 second = distance(two, one);
			
			// TODO: why do we return the MAXIMUM?????
			dist = (first > second) ? first : second;
		} // if

        return dist;
    }

    //! main method to call (extended CSS)
    int32 processExtended (CssMaximumData &oneStd, CssMaximumData &oneExt,
					       CssMaximumData &twoStd, CssMaximumData &twoExt)
	{
		int32 dist = 0;
		if (xyRatio) {
			ieee64 f=1.0;
			if (oneStd.xyRatio <= 0.0) oneStd.xyRatio = 1.0;
			if (twoStd.xyRatio <= 0.0) twoStd.xyRatio = 1.0;
			if (oneStd.xyRatio>twoStd.xyRatio) {
				f=oneStd.xyRatio / twoStd.xyRatio;
			} else {
				f=twoStd.xyRatio / oneStd.xyRatio;
			} // else
			if (f>1.5) {
				dist=INT_MAX;
			} // if
		} // if

		if (dist!= INT_MAX) {
			int32 first  = distanceExt(oneStd, oneExt, twoStd, twoExt);
			int32 second = distanceExt(twoStd, twoExt, oneStd, oneExt);
			
			// TODO: why do we return the MAXIMUM?????
			dist = (first > second) ? first : second;
		} // if

        return dist;

	}


protected:

private:


    //! calculate distance between two CssMaxima
    int distance   (CssMaximumData&, CssMaximumData&);
    int distanceExt(const CssMaximumData &, const CssMaximumData& , const CssMaximumData&, const CssMaximumData&);

    //!
    vector<int> calculateShiftList(CssMaximumData &cssi, CssMaximumData &cssp);


    //!
    int calculateDiff(CssMaximumData &cssi, CssMaximumData &cssp);


    //! number of pixels on a contour
    int contourPixels;

    //! threshold (percent)
    ieee64 thPosition;
    ieee64 thHeight;
    ieee64 thWidth;
    ieee64 thSignificantHeight;
	bool optimizeConcave;
	bool xyRatio;
	bool rotateCss;
	bool matchGlobalShapeDescriptors;

}; // CssMatching

#endif
