/*
    EcrFeature.cpp     calculate the edge change ratio (ECR)
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// MoCA includes
#include <EcrFeature.h>



std::ostream& operator<< (ostream& os, const EcrData &data)
{
	os << "version        : " << data.version          << endl;
	os << "filename       : " << data.filename         << endl;
	os << "ecr size       : " << data.size             << endl;
	os << "minimum value  : " << data.minEcrValue      << endl;
	os << "maximum value  : " << data.maxEcrValue      << endl;
	os << "average value  : " << data.avgEcrValue      << endl;
	os << "ecr file       : " << data.ecrFilename      << endl;
	os << "pin file       : " << data.pinFilename      << endl;
	os << "pout file      : " << data.poutFilename     << endl;
	os << "Moca path      : " << data.mocaPath         << endl;
	os << endl;

	return os;
} // operator<<

OTagStream& operator<< (OTagStream& os, const EcrData& data)
{
    TagFormat::writeNodeValue(os, "version",     data.version);
    TagFormat::writeNodeValue(os, "filename",    data.filename);
    TagFormat::writeNodeValue(os, "size",     data.size);
    TagFormat::writeNodeValue(os, "minEcrValue", data.minEcrValue);
    TagFormat::writeNodeValue(os, "maxEcrValue", data.maxEcrValue);
    TagFormat::writeNodeValue(os, "avgEcrValue", data.avgEcrValue);
    TagFormat::writeNodeValue(os, "ecrFilename", data.ecrFilename);
    TagFormat::writeNodeValue(os, "pinFilename", data.pinFilename);
    TagFormat::writeNodeValue(os, "poutFilename",data.poutFilename);
	os << endl;

	// write ecr file
	{
		OTagStream ots (data.mocaPath+data.ecrFilename);
		TagFormat::writeNodeValue(ots, data.ecr);
		ots.close();
	}

	// write ecr file
	{
		OTagStream ots (data.mocaPath+data.pinFilename);
		TagFormat::writeNodeValue(ots, data.pin);
		ots.close();
	}

	// write ecr file
	{
		OTagStream ots (data.mocaPath+data.poutFilename);
		TagFormat::writeNodeValue(ots, data.pout);
		ots.close();
	}

	return os;
} // operator<<

ITagStream& operator>> (ITagStream& is, EcrData& data)
{
    char c;
    bool failFlag = false;

	if (is >> c) {
		is.putback(c);

		TagFormat::readNodeValue(is, "version",     data.version);
		TagFormat::readNodeValue(is, "filename",    data.filename);
		int32 tmp = 0;
		TagFormat::readNodeValue(is, "size", tmp);
		data.size = tmp;

		TagFormat::readNodeValue(is, "minEcrValue", data.minEcrValue);
		TagFormat::readNodeValue(is, "maxEcrValue", data.maxEcrValue);
		TagFormat::readNodeValue(is, "avgEcrValue", data.avgEcrValue);
		TagFormat::readNodeValue(is, "ecrFilename", data.ecrFilename);
		TagFormat::readNodeValue(is, "pinFilename", data.pinFilename);
		TagFormat::readNodeValue(is, "poutFilename", data.poutFilename);
		if (data.ecrFilename == "") {
			failFlag = true;
		} // if
		if (is.rdstate() & ios::failbit) {
			failFlag = true;
		} // if
	} // if

	data.ecr.clear();
	data.pin.clear();
	data.pout.clear();
    if (failFlag) {
		throw MocaException ("EcrFeature: Illegal file format!");
	} else {
        is.clear();

		// read hist file
		try {
			{
				ITagStream its (data.mocaPath+data.ecrFilename);
				TagFormat::readNodeValue(its, data.ecr);
				its.close();
			}
			{
				ITagStream its (data.mocaPath+data.pinFilename);
				TagFormat::readNodeValue(its, data.pin);
				its.close();
			}
			{
				ITagStream its (data.mocaPath+data.poutFilename);
				TagFormat::readNodeValue(its, data.pout);
				its.close();
			}
		} catch (...) {
			failFlag = true;
		} // try

		if (data.ecr.size() <= 0) {
			failFlag = true;
		} // if
		if (data.pin.size() <= 0) {
			failFlag = true;
		} // if
		if (data.pout.size() <= 0) {
			failFlag = true;
		} // if
	    if (failFlag) {
			string s = "EcrFeature: Illegal data file format (" + data.mocaPath+data.ecrFilename + "," + 
				        data.mocaPath+data.pinFilename + " or " + data.mocaPath+data.poutFilename + ")!";
			throw MocaException (s);
		} // if
    } // else

    return is;
} // operator>> 



const string EcrFeature::DEFAULT_ECR_FEATURE_FILE_NAME = "ecr.idx";


EcrFeature::EcrFeature ()
{
	edgeThreshold  =  128;
	dilateRadius   =    3;
	radiusmult     =    2;
	sumreg         = 0.05; 
	sumradius      =    4;
	maxRadius      =   10;
	transradMax    =   20;
	transThreshold =  0.2;

} // EcrFeature



EcrFeature::EcrFeature (uint8 edgeThreshold, 
						uint8 dilateRadius,
						uint8 radiusmult,
						ieee64 sumreg,
						uint8 sumradius,
						uint8 maxRadius,
						uint8 transradMax,
						ieee64 transThreshold)
{
	this->edgeThreshold  = edgeThreshold;
	this->dilateRadius   = dilateRadius;
	this->radiusmult     = radiusmult;
	this->sumreg         = sumreg;
	this->sumradius      = sumradius;
	this->maxRadius      = maxRadius;
	this->transradMax    = transradMax;
	this->transThreshold = transThreshold;	
} // EcrFeature




void
EcrFeature::process (EcrData& data) 
{

	// read data
	pathFilename = data.pathFilename;
	ReadMocaData<EcrData> reader (pathFilename, ECR_FEATURE_FILENAME);
	bool dataValid = reader.process (ECR_FEATURE_VERSION, data); 

	if (!dataValid) {

		// create new index
		cerr << "EcrFeature: Creating index file ..." << endl;
		readValues(data);


		// save index file
		WriteMocaData<EcrData> writer (pathFilename, ECR_FEATURE_FILENAME);
		writer.process(data);
		cerr << "EcrFeature: Creating index file ...done" << endl;
	} // if

	return;
}


// calculate ECR difference value between two color images
void
EcrFeature::processImages (const Image<uint8> &oldImg, const Image<uint8> &oldDilated, ieee64 oldEdgeCount,
						   const Image<uint8> &newImg, const Image<uint8> &newDilated, ieee64 newEdgeCount,
						   ieee64 &ecrValue, ieee64 &rhoOut, ieee64 &rhoIn) 
{
	ecrValue = 1.0;
	rhoOut = 1.0;
	rhoIn = 1.0;


	// calculate the outgoing edges
	rhoOut = 0.0;
	if (oldEdgeCount > 0.001) {
		ieee64 numberOutgoingEdges = getOutgoingPixel(oldImg, newDilated);
		rhoOut = numberOutgoingEdges / oldEdgeCount;
	} // if 
				
	// calculate the incoming edges
	rhoIn = 0.0;
	if (newEdgeCount > 0.001) {
		//ieee64 numberIncomingEdges = getIncomingPixel(newImg, oldDilated);
		ieee64 numberIncomingEdges = getOutgoingPixel(newImg, oldDilated);
		rhoIn = numberIncomingEdges / newEdgeCount;
	} // if 

	// set ecr
	ecrValue = MAX(rhoIn, rhoOut);


	return;
} // processImages



// calculate ECR difference values
void
EcrFeature::readValues (EcrData& data) 
{

	// open video
	VideoDecoder reader (pathFilename, CODEC_Y);
	size = reader.getSize();


    // set data
	data.size = size[DIM_Z];
    data.ecr.resize(data.size);
    data.pin.resize(data.size);
    data.pout.resize(data.size);

	data.avgEcrValue = 0;
	data.maxEcrValue = -1;
	data.minEcrValue = 10e10;


	// get filename without path, create new directory if necesarry
	data.version  = ECR_FEATURE_VERSION;
	CreateMocaPath pathReader (pathFilename, false);
	data.filename = pathReader.file;
	data.ecrFilename = "ecr_data.idx";
	data.pinFilename = "ecr_in_data.idx";
	data.poutFilename = "ecr_out_data.idx";



	for (uint32 i=0; i<data.size; i++) {
		data.ecr[i]  = 0;
		data.pin[i]  = 0;
		data.pout[i] = 0;
	} // for


	// get camera parameters
	CameraModelData cameraParameters;
	{
		CameraModel reader(pathFilename);
		reader.process(cameraParameters);
	}

	//GrayscaleFilter<uint8> gray;
	CropFilter<uint8> crop (size[DIM_X]/40+2,size[DIM_X]/40+2,size[DIM_Y]/40+2,size[DIM_Y]/40+2);
	int32 borderSize = min(size[DIM_X]/40+2,size[DIM_Y]/40+2);

	uint8 th_low = 3;
	uint8 th_high = 125;

	// replace canny with gauss and derivation filter
	//CannyFilter<uint8> canny (maskSize, th_low, th_high);
	int32 maskSize = 5;
	GaussFilter<uint8> gauss (maskSize);
	int32 edgeTh = 64;
	EdgeDetectionFilter<uint8> edge (edgeTh);
    DilationFilter<uint8> dilate(2*dilateRadius+1);

    // process images
    Image<uint8> /*img0,*/ img1, img2;
	cerr << endl;
    for (uint32 i= 0; i < size[DIM_Z]-1; i++) {
		cerr << "\rframe: " << i << "            " << ends;

		img1 = reader.getImage(i);
		img2 = reader.getImage(i+1);

		// convert to grayscale images
		//gray.process(img1);
		//gray.process(img2);


		// if the camera model is not correct, the ECR values will be very high
		// img0 calculates the ECR without camera compensation
		//img0 = img2;	

#if 0
		//WriteImage (img0, "D:/img0.png");
		WriteImage (img1, "D:/img1.png");
		WriteImage (img2, "D:/img2.png");
#endif

		// compensate camera motion, if model is valid
		if (cameraParameters.data[i].isValid()) {
			cameraParameters.data[i].applyTransformation (img2);
		} // if

#if 0
		WriteImage (img2, "D:/img2_trans.png");
#endif

		// crop images
		//crop.process(img0);
		crop.process(img1);
		crop.process(img2);

#if 0
		WriteImage (img1, FileHandling::getNumberedFilename ("D:/crop/crop_",i,"_a.png"));
		WriteImage (img2, FileHandling::getNumberedFilename ("D:/crop/crop_",i,"_b.png"));
		continue;

		//WriteImage (img0, "D:/img0_crop.png");
		WriteImage (img1, "D:/img1_crop.png");
		WriteImage (img2, "D:/img2_crop.png");
#endif

	
		// get edges (background is black, edges are white!)
		//canny.process (img0);
		//canny.process (img1);
		//canny.process (img2);
		gauss.process (img1);
		gauss.process (img2);
		edge.process (img1);
		edge.process (img2);


#if 0
		WriteImage (img1, FileHandling::getNumberedFilename ("D:/edge/edge_",i,"_a.png"));
		WriteImage (img2, FileHandling::getNumberedFilename ("D:/edge/edge_",i,"_b.png"));
		continue;

		//WriteImage (img0, "D:/img0_edges.png");
		WriteImage (img1, "D:/img1_edges.png");
		WriteImage (img2, "D:/img2_edges.png");
#endif

		// remove edge pixels at border
		ImageSize size = img1.getSize();
		for (int32 y=0; y<size[DIM_Y]; y++) {
			for (int32 x=0; x<borderSize; x++) {
				//img0(x,y) = 0;
				//img0(size[DIM_X]-x-1,y) = 0;
				img1(x,y) = 0;
				img1(size[DIM_X]-x-1,y) = 0;
				img2(x,y) = 0;
				img2(size[DIM_X]-x-1,y) = 0;
			} // for
		} // for
		for (int32 x=0; x<size[DIM_X]; x++) {
			for (int32 y=0; y<borderSize; y++) {
				//img0(x,y) = 0;
				//img0(x,size[DIM_Y]-y-1) = 0;
				img1(x,y) = 0;
				img1(x,size[DIM_Y]-y-1) = 0;
				img2(x,y) = 0;
				img2(x,size[DIM_Y]-y-1) = 0;
			} // for
		} // for

#if 0
		//WriteImage (img0, "D:/img0_edges_crop.png");
		WriteImage (img1, FileHandling::getNumberedFilename ("D:/edge/img_",i,"_a.png"));
		WriteImage (img2, FileHandling::getNumberedFilename ("D:/edge/img_",i,"_b.png"));
#endif


		// dilate frames
		Image<uint8> imgD0, imgD1, imgD2;
		//dilate.process   (img0, imgD0);
		dilate.process   (img1, imgD1);
		dilate.process   (img2, imgD2);


#if 0
	Image<uint8> tmp(size);
	tmp.clear(0);
    for (int y= 0; y < size[DIM_Y]; y++) {
		for (int x= 0; x < size[DIM_X]; x++) {
			if ((edgeImage(x,y) == 255) && (dilatedImage(x,y) == 0)) {
				tmp(x,y) = 255;
			} // if
		} // for
	} // for
	WriteImage (img2, FileHandling::getNumberedFilename ("D:/edge/incoming_",i,"_b.png"));
#endif


#if 0
		//WriteImage (imgD0, "D:/img0_dilated.png");
		WriteImage (imgD1, "D:/img1_dilated.png");
		WriteImage (imgD2, "D:/img2_dilated.png");
#endif

		// calculate number of edges
		//int32 edges0 = countEdgePixel(img0);
		int32 edges1 = countEdgePixel(img1);
		int32 edges2 = countEdgePixel(img2);


		// calculate and store ECR
		//ieee64 ecrV0, poutV0, pinV0;
		//processImages (img1, imgD1, edges1, img0, imgD0, edges0, ecrV0, poutV0, pinV0);

		ieee64 ecrV1, poutV1, pinV1;
		processImages (img1, imgD1, edges1, img2, imgD2, edges2, ecrV1, poutV1, pinV1);

		// store minimum of ECR (with camera compensation) and 
		// ECR (without camera compensation)
		data.ecr[i]  = ecrV1;
		data.pout[i] = poutV1;
		data.pin[i]  = pinV1;
		//data.ecr[i]  = min (ecrV0, ecrV1);
		//data.pout[i] = min (poutV0,poutV1);
		//data.pin[i]  = min(pinV0,pinV1);


		if (data.minEcrValue > data.ecr[i]) {
			data.minEcrValue = data.ecr[i];
		} // if
		if (data.maxEcrValue < data.ecr[i]) {
			data.maxEcrValue = data.ecr[i];
		} // if
		data.avgEcrValue += data.ecr[i];

    } // for
	data.avgEcrValue /= data.size;

	return;
} // process




ieee64
EcrFeature::getOutgoingPixel(const Image<uint8>& edgeImage, const Image<uint8>& dilatedImage)
{
    uint32 outgoingPixel=0;
	ImageSize size=edgeImage.getSize();

	// edge color = 255
    for (int y= 0; y < size[DIM_Y]; y++) {
		for (int x= 0; x < size[DIM_X]; x++) {
			if ((edgeImage(x,y) == 255) && (dilatedImage(x,y) == 0)) {
				outgoingPixel++;
			} // if
		} // for
	} // for



#if 0
	Image<uint8> tmp(size);
	tmp.clear(0);
    for (int y= 0; y < size[DIM_Y]; y++) {
		for (int x= 0; x < size[DIM_X]; x++) {
			if ((edgeImage(x,y) == 255) && (dilatedImage(x,y) == 0)) {
				tmp(x,y) = 255;
			} // if
		} // for
	} // for
	WriteImage(tmp, "D:/temp.png");
#endif

    return outgoingPixel;
} // getOutgoingPixel


uint32
EcrFeature::countEdgePixel (const Image<uint8>& img)
{

	// filter (background is black, edges are white!)
    uint32 edgePixelNo = 0;
    for (int y= 0; y < size[DIM_Y]; y++) {
	    for (int x= 0; x < size[DIM_X]; x++) {
            if (img (x,y) == 255) {
                edgePixelNo++;
            } // if
        } // for
    } // for

    return edgePixelNo;
} // countEdgePixel


