/*
    Face.cpp     store face data
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "Face.h"

bool 
operator== (const Face& f1, const Face& f2) 
{
    return (f1.frame == f2.frame && f1.x == f2.x && f1.y == f2.y);
} // operator==


bool 
operator!= (const Face& f1, const Face& f2) 
{
    return (f1.frame != f2.frame || f1.x != f2.x || f1.y != f2.y);
} // operator!=


bool 
operator< (const Face& f1, const Face& f2) {
    return (f1.frame < f2.frame ||
            f1.frame == f2.frame &&
            f1.neighbourCounter < f2.neighbourCounter ||
            f1.frame == f2.frame &&
            f1.neighbourCounter == f2.neighbourCounter &&
            f1.weight < f2.weight);
} // operator<


std::ostream& operator<< (ostream &os, const Face &data) 
{
	os << data.frame << "\t" 
	   << data.x << "\t" << data.y << "\t" 
	   << data.width << "\t" << data.angle << "\t" 
	   << data.weight << endl;

	return os;
} // operator <<

std::ostream& operator<< (ostream &os, const list<Face> &data) 
{
	os << "frame\tx\ty\twidth\tangle\tweight" << endl;
	list<Face>::const_iterator it;
	for (it = data.begin(); it != data.end(); it++) {
		Face f = (*it);
		os << f;
	} // for
	return os;
} // operator <<



bool 
Face::nextTo(const Face& f) 
{
    if (this == &f) {
        return false;
    } // if
    if (nextTo(*this, f.x + f.width/2, f.y + f.width/2) ||
        nextTo(*this, f.x + f.width/2, f.y - f.width/2) ||
        nextTo(*this, f.x - f.width/2, f.y + f.width/2) ||
        nextTo(*this, f.x - f.width/2, f.y - f.width/2) ||
        nextTo(f, x + width/2, y + width/2) ||
        nextTo(f, x + width/2, y - width/2) ||
        nextTo(f, x - width/2, y + width/2) ||
        nextTo(f, x - width/2, y - width/2)) {
        return true;
    }
    return false;
} // Face::nextTo()


bool 
Face::nextTo(const Face& f, const int x, const int y) 
{
    // tests if the corner is in the area of the face
    if ((f.x - f.width/2) <= x && (f.y - f.width/2) <= y &&
        (f.x + f.width/2) >= x && (f.y + f.width/2) >= y) {
        return true;
    } // if    
    return false;
} // Face::nextTo()







// FaceItem

std::ostream& operator<< (ostream& os, const FaceItem &item)
{
	os << "position: (" << item.p1[DIM_X] << "," << item.p1[DIM_Y] << ") (" << item.p2[DIM_X] << "," << item.p2[DIM_Y] << ")    ";
	os << "angle:";
	if (item.angle<0 && item.angle>-10) os << " ";
	if (item.angle>=0 && item.angle<10) os << "  ";
	if (item.angle>=10) os << " ";
	os << item.angle << "    ";
	os << "frame: ";
	if (item.frame<10) os << " ";
	if (item.frame<100) os << " ";
	if (item.frame<1000) os << " ";
	os << item.frame << endl;

	return os;
} // operator<<

OTagStream& operator<< (OTagStream& os, const FaceItem& item)
{
	TagFormat::writeStartTag (os, "face");
	os  << "\t" 
		<< setw(5) << item.p1[DIM_X] << "\t" 
		<< setw(5) << item.p1[DIM_Y] << "\t"  
		<< setw(5) << item.p2[DIM_X] << "\t"  
		<< setw(5) << item.p2[DIM_Y] << "\t" 
		<< setw(5) << item.angle     << "\t" 
		<< setw(5) << item.frame; 
	TagFormat::writeEndTag (os, "face");

    return os;
} // operator<<

ITagStream& operator>> (ITagStream& is, FaceItem& item)
{
    char c;
    bool failFlag = false;

	if (is >> c) {
		is.putback(c);

		TagFormat::readStartTag (is, "face");
		if (is.rdstate() & ios::failbit) {
			failFlag = true;
		} // if
		if (!failFlag) {
			is >> item.p1[DIM_X];
			is >> item.p1[DIM_Y];
			is >> item.p2[DIM_X];
			is >> item.p2[DIM_Y];
			is >> item.angle; 
			is >> item.frame; 
			TagFormat::readEndTag (is, "face");
		}
	} // if

	if (is.rdstate() & ios::failbit) {
		failFlag = true;
	} // if
    if (failFlag) {
		throw MocaException ("Face: Illegal file format!");
	} else {
        is.clear();
    } 

    return is;
} // operator>> 





// FaceData

std::ostream& operator<< (ostream& os, const FaceData &data)
{
	os << "version        : " << data.version << endl;
	os << "filename       : " << data.filename << endl;
	os << "number of faces: " << data.numberOfFaces << endl;
	os << endl;

	int32 cnt = 0;
	list<FaceItem>::const_iterator it;
	for (it=data.faces.begin(); it!=data.faces.end(); it++) {
		if ((*it).frame>-1 && (*it).p2[DIM_X]>0 && (*it).p2[DIM_Y]>0) {
			os << (*it);
			cnt++;
		} // if
	} // for
	cerr << "total number of faces: " << cnt << endl << endl;

	return os;
} // operator<<

OTagStream& operator<< (OTagStream& os, const FaceData& data)
{
	TagFormat::writeNodeValue (os, "version", data.version);
	TagFormat::writeNodeValue (os, "filename", data.filename);
	TagFormat::writeNodeValue (os, "numberOfFaces", data.numberOfFaces);
	os << endl;

	list<FaceItem>::const_iterator it;
	for (it=data.faces.begin(); it!=data.faces.end(); it++) {
		os << (*it);
	} // for

    return os;
} // operator<<

ITagStream& operator>> (ITagStream& is, FaceData& data)
{
    char c;
    bool failFlag = false;

	if (is >> c) {
		is.putback(c);

		TagFormat::readNodeValue (is, "version", data.version);
		TagFormat::readNodeValue (is, "filename", data.filename);
		TagFormat::readNodeValue (is, "numberOfFaces", data.numberOfFaces);
		if (is.rdstate() & ios::failbit) {
			failFlag = true;
		} // if

		if (!failFlag) {
			FaceItem item;
			for (int i=0; i< data.numberOfFaces; i++) {
				is >> item;
				if (is.rdstate() & ios::failbit) {
					failFlag = true;
				} else {
					data.faces.push_back(item);
				} // else
			} // for
		} // for
	} // if

	if (is.rdstate() & ios::failbit) {
		failFlag = true;
	} // if
    if (failFlag) {
		throw MocaException ("FaceData: Illegal file format!");
	} else {
        is.clear();
    } 

    return is;
} // operator>> 




