/*
   This file is part of the MoCA video library.
   Name: FaceDetection.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file



   \author    Stephan Kopf

              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim
              68131 Mannheim
              Germany
              email: kopf@informatik.uni-mannheim.de

   (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt

*/


#ifndef FACE_DETECTION_H
#define FACE_DETECTION_H

// C++ includes
#include <cmath>
#include <cstring>
#include <cstdlib>
#include <ctime>
#include <sstream>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <list>
#include <ctime>
#include <utility>
using namespace std;

// MoCA includes
#include <Face.h>
#include <VideoDecoder.h>
#include <RotationFilter.h>
#include <FaceMaskFilter.h>
#include <DrawRectangle.h>
#include <ScaleFilter.h>
#include <OImageStream.h>
#include <IImageStream.h>
#include <VideoStatistic.h>


// C includes
extern "C" {
#include "FaceData1.h"
#include "FaceData2.h"
#include "FaceData3.h"
#include "FaceData4.h"
#include "FaceData5.h"
}

// TODO: warning: unused structs ???
static int tmpFaceDetectionValue1 = nn1REC.NoOfInput;
static int tmpFaceDetectionValue2 = nn2REC.NoOfInput;
static int tmpFaceDetectionValue3 = nn3REC.NoOfInput;
static int tmpFaceDetectionValue4 = nn4REC.NoOfInput;
static int tmpFaceDetectionValue5 = nn5REC.NoOfInput;


static const string FACE_DETECTION_FILENAME ="faces.idx";
static const int    FACE_DETECTION_PARAMETER_SIZE = 20;
static const double FACE_DETECTION_PARAMETER_THRESHOLD= 0.5;

static const double FACE_DETECTION_PI = 3.1415927;
static const double FACE_DETECTION_SCALE_FACTOR = 0.83;
static const int    FACE_DETECTION_DEFAULT_NUMBER_OF_NETS = 5;

static const int    FACE_DETECTION_SIZE_DEFAULT = 0;
static const int    FACE_DETECTION_SIZE_MEDIUM  = 1;
static const int    FACE_DETECTION_SIZE_LARGE   = 2;
static const int    FACE_DETECTION_SIZE_HUGE    = 3;

static const int FACE_INPUT_UNKNOWN = 0;    
static const int FACE_INPUT_VIDEO = 1;   
static const int FACE_INPUT_IMAGE = 2;   



struct FaceDetectionParameter
{

	FaceDetectionParameter ()
	{
		inputData    = FACE_INPUT_UNKNOWN;
		useColorInfo = false;
		report       = false;
		outFile      = false;
		rotate       = false;
		verbose      = false;
		
		mpgFilename  = "";
		
		size         = FACE_DETECTION_PARAMETER_SIZE;
		fastSize     = size;
		defaultFaceSize = 0;
		srcFileFirst = -1;
		srcFileLast  = -1;
		stepWidth    = -1;
		numberOfNets = 1;
		degree       = 10;
		rotationStep = 0;
		threshold    = FACE_DETECTION_PARAMETER_THRESHOLD;
		analyseOne   = -1;
		analyseFrame.resize(SignalSize(1));
	} // FaceDetectionParameter

    FaceDetectionParameter& operator=(const FaceDetectionParameter &other) {
		if (this != &other) {
			inputData     = other.inputData;
			useColorInfo  = other.useColorInfo;
			report        = other.report;
			outFile       = other.outFile;
			rotate        = other.rotate;
			verbose       = other.verbose;
			
			mpgFilename   = other.mpgFilename;
			
			size          = other.size;
			fastSize      = other.fastSize;
			defaultFaceSize = other.defaultFaceSize;

			srcFileFirst  = other.srcFileFirst;
			srcFileLast   = other.srcFileLast;
			stepWidth     = other.stepWidth;
			numberOfNets  = other.numberOfNets;
			degree        = other.degree;
			rotationStep  = other.rotationStep;
			threshold     = other.threshold;
			analyseOne    = other.analyseOne;
			analyseFrame  = other.analyseFrame;
		} // if
        return *this;
    } // operator=


	int inputData;  // value must be defined!

	bool useColorInfo;
	bool report;
	bool outFile;
	bool rotate;
	bool verbose;

	string mpgFilename;

	int size;
	int fastSize;
	int defaultFaceSize; // DEFAULT=0, MEDIUM=1, LARGE=2, HUGE=3

	int srcFileFirst;
	int srcFileLast;
	int stepWidth;
	int numberOfNets;
	int degree;
	int rotationStep;
	double threshold;
	int32 analyseOne;
	Signal<bool> analyseFrame;
}; // FaceParameter


class FaceDetection {
public:
	static const string DEFAULT_FACE_FILE_NAME;


	// constructor
    FaceDetection(int argc, char** argv);
    FaceDetection(FaceDetectionParameter& param);

    ~FaceDetection();
  
	// analyze image or video
    void processVideo(FaceData&);
    void processImage(const Image<uint8> & src, list<Face> &faces);

	// draw all faces in the video 
	// filename: MPG or AVI video
	// data: faces of this video (result of process Video())
    void storeFaces (FaceData& data, string &filename);

	// draw a ractangle around the detected face
	void markFace (FaceItem &item, Image<uint8> &img);
    
private:

    ieee64 scalingFactor;
    int (*neuralNet[5])(float*, float*, int); // = {nn1, nn2, nn3, nn4, nn5};
    
    VideoDecoder *video;
    
    char *nameOfOutFile;

    double inverseFactor;
    float *nnInput;
    float **nnOutput;
    int colorCheckPos, colorCheckNeg;

    // current frame number
    int frameCounter;

    int **mask;
    int **pixel;

    int *columnSum;
    int *rowSum;

    ofstream videoFile;
    list<Face> *facesThroughNet[FACE_DETECTION_DEFAULT_NUMBER_OF_NETS];

    string facePath;
	string faceFilename;


    int lightingCorrectionAllocatedSize;
    int lightingCorrectionX;
    int lightingCorrectionX2;
    float networkUint8ToFloat[256];
	int32 totalNumberOfFaces;

	// correct shift of face based on rotation
    int32 shiftX;
    int32 shiftY;

private:


    // contains the orginal and unchanged image as it was requested from
    // the video. Should never be destroyed
    Image<uint8> orgColorImage;

    // grayscale version of orgColorImage; never changed
    Image<uint8> orgImage;

    // rotated color version of orgColorImage
    Image<uint8> colorImage;

    // scaled grayscale version of orgImage
    Image<uint8> image;

    Image<uint8> breakMask;
    
    void makeMask();

    void scaleImage(double scaleFactor);
    void rotateImage(int); // detection of rotated faces
    bool checkColor(const int, const int);
    void cropImage(int, int);
    void cropRotatedImage(int, int, double);

    // true = Konstrast ist ausreichend gross, als dass sich dort ein Gesicht
    //        befinden koennte
    // false= kein Gesicht an dieser Stelle
    bool lightingCorrection();

    void network(float*, float**);
    void reverseRotate(Face*, int32);
    void searchFaces(list<Face>**, int32);
    void thresholding(int, int, list<Face> &);
    void countNeighbours(Face&, const int, list<Face> &);
    void overlapDetection(list<Face> &);
    void doReport(list<Face> &, int no);

    // Zeichnet Rechteck um die gefundenen Gesichter und speichert es ab.
   // void drawRect(list<Face> &);

    void resetValues();
    //Image<uint8>& createBreakMask (const Image<uint8>& src, Image<uint8>& dest);
    //void searchInRotatedImage(int, int);

	FaceDetectionParameter param;

	void initializeParameters(int srcFileFirst=-1, int srcFileLast=-1, int stepWidth=-1, int32 analyseOne=-1);
	void detectFaces(FaceData& data);
	void analyseImage (list<Face> &frameList, int32 currentAngle);

};



#endif // FACE_DETECTION





