/*
   This file is part of the MoCA video library.
   Name: HistogramFeature.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#ifndef HISTOGRAM_FEATURE_H
#define HISTOGRAM_FEATURE_H

class VideoDecoder;


//C++ includes
//#include <vector>
#include <math.h>

// MoCA includes
#include <MocaSignal.h>
#include <Image.h>
//#include <GetType.h>
#include <VideoDecoder.h>


//! Class to calculate a Histogram of a given image
/*!
  This class implements a histogram feature. Use this class to get a histogram of the image. 
  It will be stored in a Signal<int32>.
 */
/*
const static string HISTOGRAM_VERSION  ="Version-0.35";
const static string HISTOGRAM_FILENAME ="histogramdata.idx";

struct HistogramData
{
	HistogramData () 
	{
		version   = HISTOGRAM_VERSION;
		msBits    = -1;
		grayscale = false;
		smoothing = false;
		areas     = 1;
		histSize      = -1;
		frames        = -1;
		data.resize(SignalSize(1));
	} 
    HistogramData& operator=(const HistogramData &other) {
		if (this != &other) {
			version   = other.version;
			msBits    = other.msBits;
			grayscale = other.grayscale;
			smoothing = other.smoothing;
			areas     = other.areas;
			histSize  = other.histSize;
			frames    = other.frames;
			data      = other.data;
		} // if
        return *this;
    } // operator=

	string version;
	int8 msBits;
	bool grayscale;
	bool smoothing;
	uint8 areas;
	int32 histSize;
	int32 frames;
	Signal<uint32> data;
};


inline std::ostream& operator<< (ostream& os, const HistogramData &data)
{
	os << "ms bits        : " << (int)data.msBits<< endl;
	if (data.grayscale) {
		os << "grayscale      : yes"  << endl;
	} else {
		os << "grayscale      : no"  << endl;
	} // else
	if (data.smoothing) {
		os << "smoothing      : yes"  << endl;
	} else {
		os << "smoothing      : no"  << endl;
	} // else
	os << "areas          : " << data.areas<< endl;
	os << "histogram size : " << data.histSize<< endl;
	os << "frames         : " << data.frames<< endl;
	os << "version        : " << data.version<< endl;
	os << endl;

	return os;
} // operator<<

inline OTagStream& operator<< (OTagStream& os, const HistogramData& data)
{
    TagFormat::writeNodeValue(os, "version",      data.version);
    TagFormat::writeNodeValue(os, "msBits",       data.msBits);
    TagFormat::writeNodeValue(os, "grayscale",    data.grayscale);
    TagFormat::writeNodeValue(os, "smoothing",    data.smoothing);
    TagFormat::writeNodeValue(os, "areas",        data.areas);
    TagFormat::writeNodeValue(os, "histSize",     data.histSize);
    TagFormat::writeNodeValue(os, "frames",       data.frames);

    TagFormat::writeNodeValue(os, data.data);

	return os;
} // operator<<



inline ITagStream& operator>> (ITagStream& is, HistogramData& data)
{
    char c;
    bool failFlag = true;

	if (is >> c) {
		failFlag = false;
		is.putback(c);

		TagFormat::readNodeValue(is, "version", data.version);
		TagFormat::readNodeValue(is, "msBits", data.msBits);
		TagFormat::readNodeValue(is, "grayscale", data.grayscale);
		TagFormat::readNodeValue(is, "smoothing",  data.smoothing);
		TagFormat::readNodeValue(is, "areas",  data.areas);
		TagFormat::readNodeValue(is, "histSize",  data.histSize);
		TagFormat::readNodeValue(is, "frames",  data.frames);
		if (is.rdstate() & ios::failbit) {
			failFlag = true;
		} // if

		SignalSize s(data.frames,data.histSize);
		TagFormat::readNodeValue(is, data.data,s);
		if (is.rdstate() & ios::failbit) {
			failFlag = true;
		} // if
	} // if

    if (failFlag) {
		throw MocaException ("HistogramFeature: Illegal histogram file format!");
	} // if
    is.clear();

    return is;
} // operator>> 

*/


//template <class T>
class HistogramFeature
{
public:

    //! Contructor
    /*! \param msbits of type uint8 representing the number of most significant bits of each colour value to be included in the calculation of the histogram
	    \grayscale of type bool defines, if a grayscale image is analysed
		\areas split image in areas, valid values: 1, 4, 9, 16, 25
	*/
    HistogramFeature (int8 msbits = -1, bool grayscale=false, bool smoothing=false, uint8 areas=1) 
	{
		factor=NULL;
		this->msbits = msbits;
		if (msbits==-1) {
			if (grayscale) {
				this->msbits = 4;
			} else {
				this->msbits = 3;
			} // else
		} // if
		if (msbits<1) msbits=1;
		if (msbits>8) msbits=8;

		this->grayscale = grayscale;
		this->smoothing = smoothing;

		this->areas     = areas;
		if (areas<=1) areas=1;
		if ((areas>1) && (areas<=4))  areas=4;
		if ((areas>4) && (areas<=9))  areas=9;
		if ((areas>9) && (areas<=16)) areas=16;
		if (areas>16) areas=25;

		if (!grayscale) {
			this->smoothing=false;
		} // if
	} // HistogramFeature 

    //! Default destructor
    ~HistogramFeature();

    //! Return the histrogram in the Signal given as parameter
    /*!
      \param img an argument of type Image<Type>& which references the source image
      \param histogram an argument of type Signal<int32>& referencing the vector which is going to contain the calculated histogram.\\
        The resulting vector is containing the number of pixels of the value corresponding to its index. E.g.: There are 20 pixels of value 235 in the image, so the targetVector value at index 235 will be 20.
    */
    void process (Image<uint8>& img, Signal<uint32>& histogram)
	{
		if ((grayscale) || (img.getSize()[DIM_BAND]==1)){
			processGrayscale (img, histogram);
		} else {
			processColor (img, histogram);
		} // else
	} // process

    //! Return the histrogram in the Signal given as parameter
    /*!
      \param filename of type string& is the name of the MPEG file
    */
    //void process (string &filename, Signal<int32>& histogram, bool saveData=false);


private:

    //! Return the histrogram in the Signal given as parameter
    /*!
      \param img an argument of type Image<Type>& which references the source image
      \param histogram an argument of type Signal<int32>& referencing the vector which is going to contain the calculated histogram.\\
        The resulting vector is containing the number of pixels of the value corresponding to its index. E.g.: There are 20 pixels of value 235 in the image, so the targetVector value at index 235 will be 20.
        In the case of pnm images, the histogram will be splitted into its three basic colours and stored in the vector in the order RED, GREEN, BLUE
      \param msbits of type uint8 representing the number of most significant bits of each colour value to be included in the calculation of the histogram
    */
    void processColor (Image<uint8>& img, Signal<uint32>& histogram);

    //! Return the histrogram in the Signal given as parameter
    /*!
      \param img an argument of type Image<Type>& which references the source image, only the first band is used
      \param targetArray an argument of type vector<int32>& referencing the vecotr which is going to contain the calculated histogram.\\
        The resulting vector is containing the number of pixels of the value corresponding to its index. E.g.: There are 20 pixels of value 235 in the image, so the targetVector value at index 235 will be 20.
        In the case of pnm images, the histogram will be splitted into its three basic colours and stored in the vector in the order RED, GREEN, BLUE
      \param msbits of type uint8 representing the number of most significant bits of each colour value to be included in the calculation of the histogram
      \param smoothing of type bool merges the histogram values between blocks
    */
    void processGrayscale (Image<uint8>& img, Signal<uint32>& histogram);

	int8 msbits;
	bool grayscale;
	bool smoothing;
	uint8 areas;
	uint32 *factor;

}; //class HistogramFeature

//#include <HistogramFeature.cpt>

#endif
