/*
    ImageStatistic.cpp     calculate statistical values in an image
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "ImageStatistic.h"

std::ostream& operator<< (ostream& os, const ImageStatisticData &item)
{
	os << "average  [BAND 0]: " << (int)item.avg[0] << endl;
	os << "average  [BAND 1]: " << (int)item.avg[1] << endl;
	os << "average  [BAND 2]: " << (int)item.avg[2] << endl;
	os << endl;
	os << "sigma [BAND 0]: " << item.sigma[0] << endl;
	os << "sigma [BAND 1]: " << item.sigma[1] << endl;
	os << "sigma [BAND 2]: " << item.sigma[2] << endl;	
	os << endl;
	os << "contrast based on absolute difference   : " << (int)item.contrastMaxMin << endl;
	os << "contrast based on histogram distribution: " << (int)item.contrastHistogram << endl;
	os << "contrast based on scaled images         : " << (int)item.contrastScale << endl;
	os << "average edge strength                   : " << (int)item.avgEdgeStrength << endl;
	os << endl;
	
	switch (item.color) {
	case TYPE_COLOR:
		os << "image type : color" << endl;	
		break;
	case TYPE_GRAYSCALE:
		os << "image type : grayscale" << endl;	
		break;
	case TYPE_NEAR_GRAYSCALE:
		os << "image type : near grayscale" << endl;	
		break;
	    default:
		os << "image type : UNKNOWN" << endl;	
		break;		
	} // switch

	return os;
} // operator<<


void 
ImageStatistic::processAvg(Image<uint8>& img, ImageStatisticData& data) 
{
	ImageSize size = img.getSize();
	int32 pixel = size[DIM_X] * size[DIM_Y];

	int32 avg0 = 0;
	int32 avg1 = 0;
	int32 avg2 = 0;

	if (size[DIM_BAND] == 1) {

		// average grayscale value
		for (int y=0; y<size[DIM_Y]; y++) {
			for (int x=0; x<size[DIM_X]; x++) {
				avg0 += img(x,y,0);
			} // for
		} // for
		avg0 /= pixel;
		data.color = TYPE_GRAYSCALE;

	} else {

		// average color value
		for (int y=0; y<size[DIM_Y]; y++) {
			for (int x=0; x<size[DIM_X]; x++) {
				avg0 += img(x,y,0);
				avg1 += img(x,y,1);
				avg2 += img(x,y,2);				
			} // for
		} // for		
		avg0 /= pixel;
		avg1 /= pixel;
		avg2 /= pixel;
	} // else


	// check values
	if (avg0<0) avg0=0; if (avg0>255) avg0=255;
	if (avg1<0) avg1=0; if (avg1>255) avg1=255;
	if (avg2<0) avg2=0; if (avg2>255) avg2=255;
	
	data.avg[0] = avg0;
	data.avg[1] = avg1;
	data.avg[2] = avg2;
	return;
}

void 
ImageStatistic::process(Image<uint8>& img, ImageStatisticData& data) 
{
	ImageSize size = img.getSize();
	int32 pixel = size[DIM_X] * size[DIM_Y];

	int32 avg0 = 0;
	int32 avg1 = 0;
	int32 avg2 = 0;

	ieee64 sigma0 = 0;
	ieee64 sigma1 = 0;
	ieee64 sigma2 = 0;

	data.color = TYPE_COLOR;
	if (size[DIM_BAND] == 1) {

		// average grayscale value
		for (int y=0; y<size[DIM_Y]; y++) {
			for (int x=0; x<size[DIM_X]; x++) {
				avg0 += img(x,y,0);
			} // for
		} // for
		avg0 /= pixel;

		for (int y=0; y<size[DIM_Y]; y++) {
			for (int x=0; x<size[DIM_X]; x++) {
				sigma0 += (avg0 - img(x,y,0)) * (avg0 - img(x,y,0));
			} // for
		} // for
		sigma0 /= pixel;
		sigma0 = sqrt (sigma0);
		data.color = TYPE_GRAYSCALE;

	} else {

		// average color value
		for (int y=0; y<size[DIM_Y]; y++) {
			for (int x=0; x<size[DIM_X]; x++) {
				avg0 += img(x,y,0);
				avg1 += img(x,y,1);
				avg2 += img(x,y,2);				
			} // for
		} // for		
		avg0 /= pixel;
		avg1 /= pixel;
		avg2 /= pixel;

		// variance
		for (int y=0; y<size[DIM_Y]; y++) {
			for (int x=0; x<size[DIM_X]; x++) {
				sigma0 += (avg0 - img(x,y,0)) * (avg0 - img(x,y,0));
				sigma1 += (avg1 - img(x,y,1)) * (avg1 - img(x,y,1));
				sigma2 += (avg2 - img(x,y,2)) * (avg2 - img(x,y,2));
			} // for
		} // for
		sigma0 /= pixel;
		sigma1 /= pixel;
		sigma2 /= pixel;

		// standard deviation
		sigma0 = sqrt (sigma0);
		sigma1 = sqrt (sigma1);
		sigma2 = sqrt (sigma2);


		// is color image?
		data.color = TYPE_COLOR;
		
		// thresholds
		const ieee64 AVG_GRAYSCALE_DIFF      =  4.0;
		const ieee64 AVG_NEAR_GRAYSCALE_DIFF =  7.0;
		const ieee64 SIGMA_GRAYSCALE         =  4.0;
		const ieee64 SIGMA_NEAR_GRAYSCALE    =  7.0;

		if ((abs(avg1 - 128.0) < AVG_NEAR_GRAYSCALE_DIFF) && 
			(abs(avg2 - 128.0) < AVG_NEAR_GRAYSCALE_DIFF) &&
			(sigma1 < SIGMA_NEAR_GRAYSCALE)                 && 
			(sigma2 < SIGMA_NEAR_GRAYSCALE) ) {
			data.color = TYPE_NEAR_GRAYSCALE;		

			if ((abs(avg1 - 128.0) < AVG_GRAYSCALE_DIFF) && 
				(abs(avg2 - 128.0) < AVG_GRAYSCALE_DIFF) &&
				(sigma1 < SIGMA_GRAYSCALE)                 && 
				(sigma2 < SIGMA_GRAYSCALE) ) {
				data.color = TYPE_GRAYSCALE;		
			} // if
		} // if
	} // else

	// check values
	if (avg0<0) avg0=0; if (avg0>255) avg0=255;
	if (avg1<0) avg1=0; if (avg1>255) avg1=255;
	if (avg2<0) avg2=0; if (avg2>255) avg2=255;
	
	if (sigma0<0) sigma0=0.0; if (sigma0>255) sigma0=255.0;
	if (sigma1<0) sigma1=0.0; if (sigma1>255) sigma1=255.0;
	if (sigma2<0) sigma2=0.0; if (sigma2>255) sigma2=255.0;

	data.avg[0] = avg0;
	data.avg[1] = avg1;
	data.avg[2] = avg2;
	data.sigma [0] = (int32)sigma0;
	data.sigma [1] = (int32)sigma1;
	data.sigma [2] = (int32)sigma2;


	// set contrast based on histograms (scaled to 0..1), contrast is based on lightness
	Signal<uint32> histogram;
	int8 msbits = 8;
	bool grayscale=true;
	bool smoothing=false;
	uint8 areas=1; 
    HistogramFeature hf(msbits, grayscale, smoothing, areas);
	int32 left;
	int32 right;
	int32 sum;
	int32 sizeX=img.getSize()[DIM_X];
	int32 sizeY=img.getSize()[DIM_Y];
	int32 searchPixel=(int32) (0.1*sizeX*sizeY);

	// calculate histogram
	hf.process (img, histogram);

	// calculate: contrastMaxMin
	left=0;
	right=255;
	while (histogram(left) == 0) {
		left++;
	} // while
	while (histogram(right) == 0) {
		right--;
	} // while
	left = right-left;
	if (left<0) left=0;
	if (left>255) left=255;
	data.contrastMaxMin = (uint8)(100.0*left/255.0);


	// calculate: contrastHistogram
	left=0;
	right=255;
	sum=0;
	while (sum<searchPixel) {
		sum+=histogram(left);
		left++;
	} // while
	
	sum=0;
	while (sum<searchPixel) {
		sum+=histogram(right);
		right--;
	} // while
	left = right-left+2;
	if (left<0) left=0;
	if (left>255) left=255;
	data.contrastHistogram = (uint8)(100.0*left/255.0);


	// calculate: contrastScale
	// copy image: copy only Y-plane
	Image<uint8> tmp;
	size[DIM_BAND]=1;
	tmp.realloc(size, ImageSize(0,0), BORDER_PERIODIC);
	for (int y=0; y<sizeY; y++) {
		for (int x=0; x<sizeX; x++) {
			tmp(x,y,0) = img(x,y,0);
		} // for
	} // for

	// scale image 
	ScaleFilter<uint8> sf;
	sf.reduce(tmp, 4, 4);
	size=tmp.getSize();

	// calculate histogram of scaled image
	hf.process (tmp, histogram);
	searchPixel=(int32) (0.1*size[DIM_X]*size[DIM_Y]);

	// calculate: contrastScale
	left=0;
	right=255;
	sum=0;
	while (sum<searchPixel) {
		sum+=histogram(left);
		left++;
	} // while
	
	sum=0;
	while (sum<searchPixel) {
		sum+=histogram(right);
		right--;
	} // while
	left = right-left+2;
	if (left<0) left=0;
	if (left>255) left=255;
	data.contrastScale= (uint8)(100.0*left/255.0);

	// calculate: avgEdgeStrength
	ieee64 value;
	{
		EdgeDetectionFilter<uint8> edf;
		edf.countEdgeValue (img, value);

		// scale value
		value = 50.0*value;
		if (value<0)     value=0;
		if (value>255.0) value=255;
	}
	data.avgEdgeStrength= (uint8)(value);

	return;
} // process


