/*
   This file is part of the MoCA video library.
   Name: MotionVector.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/


#ifndef MOTION_VECTOR_H
#define MOTION_VECTOR_H

// C++ includes
#include <iomanip>
#include <sstream>
#include <iostream>
using namespace std;


// MoCA includes
#include <MocaSignal.h>
#include <Image.h>
#include <MpegStatus.h>
#include <VideoDecoder.h>
#include <OImageStream.h>
#include <BlockMotionFeature.h>
#include <DrawDot.h>
#include <DrawLine.h>
#include <CreateMocaPath.h>
#include <WriteMocaData.h>
#include <GrayscaleFilter.h>
//#include <MedianFilter.h>
#include <GaussFilter.h>


static const string MOTION_VECTOR_VERSION  = "Version-0.06";
static const string MOTION_VECTOR_STATUS_FILENAME = "motionvectorstatus.idx";
static const string MOTION_VECTOR_DATA_FILENAME   = "motionvectordata.idx";


struct MotionVectorStatus
{
	MotionVectorStatus () 
	{
		version  = MOTION_VECTOR_VERSION;
		filename = "";
		size[DIM_X]=1;
		size[DIM_Y]=1;
		size[DIM_Z]=1;
		size[DIM_BAND]=3;
		blockSize = 16;
	} 
    MotionVectorStatus& operator=(const MotionVectorStatus &other) {
		if (this != &other) {
			version        = other.version;
			filename       = other.filename;
			size           = other.size;
			blockSize      = other.blockSize;
		} // if
        return *this;
    } // operator=
	
	string version;
	string filename;
	VideoSize size;
	int32 blockSize;
};

std::ostream& operator<< (ostream&,    const MotionVectorStatus&);
OTagStream&   operator<< (OTagStream&, const MotionVectorStatus&);
ITagStream&   operator>> (ITagStream&, MotionVectorStatus&);


const int32 DEFAULT_BUFFER_SIZE=500;
struct MotionVectorBuffer
{
	MotionVectorBuffer () 
	{
		size[DIM_X]=0;
		size[DIM_Y]=0;
		size[DIM_Z]=DEFAULT_BUFFER_SIZE;
		size[DIM_BAND]=3;
		offset=0;
	} 
    MotionVectorBuffer& operator=(const MotionVectorBuffer &other) {
		if (this != &other) {
			size           = other.size;
			data           = other.data;
			offset         = other.offset;
		} // if
        return *this;
    } // operator=
	
	VideoSize size;
	Video<int8> data; // band=0/1: motion in x-/y-direction, band=2: meassure for realibility of motion vector [0..100]
	int32 offset;
};
std::ostream& operator<< (ostream&,    const MotionVectorBuffer&);
OTagStream&   operator<< (OTagStream&, const MotionVectorBuffer&);
ITagStream&   operator>> (ITagStream&, MotionVectorBuffer&);



/*
Constructor: initializes variables

	process(): 
	if status file does not exist:
	calculate motion vectors of mpeg-file, save status file
	else:
	read motion vectors from status file
*/
const int32 DEFAULT_SEARCH_RANGE = 64;
class MotionVector
{
public:
	
	MotionVector(const string &pathFilename, int32 searchRange = DEFAULT_SEARCH_RANGE);
	~MotionVector() 
	{
		if (motionIts!=NULL) {
			delete motionIts;
		} // if
		motionIts=NULL;
	}
	
	// data(x,y,0): motion vector of one frame in block (x,y) in X-direction
	// data(x,y,1): motion vector of one frame in block (x,y) in Y-direction
	// data(x,y,2): meassure for realibility of motion vector [0..100], 0=not reliable, 100=very reliable
	//              based on difference of blocks and contrast of block
	void getMotionVector(int32 frame, Image<ieee64> &data);

	// summarize absolute difference of motion vectors
	void getSAD (Signal<ieee64> &data);


	void process (MotionVectorStatus&);
	void normalizeVectors (Image<ieee64> &);

private:
	void readValues (MotionVectorStatus& status);

	string pathFilename;
	int32 bufferOffset;
	MotionVectorBuffer buffer;
	MotionVectorStatus status;
	ITagStream *motionIts;
	int32 lastOffset;
	int32 searchRange;

};




// analyse motion vectors and 
// calculate the values of the 6 parameter motion estimation modell
//
//     x'      a b     x       dx 
//   (   ) = (     ) (   ) + (    )
//     y'      c d     y       dy
//  x,y: original position, x'y': new position
struct MotionParametrizationData
{
	MotionParametrizationData () 
	{
		a=d=1.0;
		b=c=0.0;
		dx=dy=0.0;
	} 
    MotionParametrizationData& operator=(const MotionParametrizationData &other) {
		if (this != &other) {
			a  = other.a;
			b  = other.b;
			c  = other.c;
			d  = other.d;
			dx = other.dx;
			dy = other.dy;
		} // if
        return *this;
    } // operator=
	
	ieee64 a,b,c,d;
	ieee64 dx,dy;
};
std::ostream& operator<< (ostream&, const MotionParametrizationData&);


class MotionParametrization
{
public:
	
	MotionParametrization(int32 macroBlockSize, const ImageSize &size);
	~MotionParametrization() {}
	
	void process (Image<ieee64> &motionvectors, MotionParametrizationData& data);

	//   x'     a b     x     dx
	//  (  ) = (   ) * ( ) + (  )
	//   y'     c d     y     dy
	void compensateForwardMotion (Image<uint8> &src, Image<uint8> &dest, MotionParametrizationData& data);
	void compensateBackwardMotion (Image<uint8> &src, Image<uint8> &dest, MotionParametrizationData& data);

private:
	void getMotionParameters (Image<ieee64> &motionVectors, MotionParametrizationData& data);


	int32 macroBlockSize;
	ImageSize imgSize;
};

#endif
