/*
   This file is part of the MoCA video library.
   Name: TextDetection.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/


#ifndef TEXT_DETECTION_H
#define TEXT_DETECTION_H



// C++ includes
#include <iomanip>
#include <sstream>
#include <iostream>
using namespace std;


// MoCA includes
#include <VideoStatistic.h>
#include <FileHandling.h>
#include <DrawRectangle.h>
#include <CannyFilter.h>
#include <CropFilter.h>
#include <TransposeFilter.h>
#include <RegionFeature.h>
#include <IImageStream.h>
#include <ErosionFilter.h>
#include <OpeningFilter.h>
#include <ClosingFilter.h>


const static string TEXT_DETECTION_VERSION = "Version-0.10";
const static string TEXT_DETECTION_FILENAME= "textarea.idx";



// store a bounding box for one character
struct TextDetectionChar
{
	TextDetectionChar ()
	{
		minX = -1;
		maxX = -1;
		minY = -1;
		maxY = -1;
		s    = "";
	}
	int32 minX, maxX, minY, maxY;
	string s;
};

struct TextDetectionBox
{
	TextDetectionBox ()
	{
		minX = -1;
		maxX = -1;
		minY = -1;
		maxY = -1;
		firstFrame = -1;
		lastFrame  = -1;
		rel  = 0.0; // reliability
	}
    TextDetectionBox& operator=(const TextDetectionBox &other) {
		if (this != &other) {
			minX = other.minX;
			maxX = other.maxX;
			minY = other.minY;
			maxY = other.maxY;
			firstFrame= other.firstFrame;
			lastFrame= other.lastFrame;
			rel  = other.rel;
		} // if
        return *this;
    } // operator=

	
	// transpose x and y values
	void transpose()
	{
		int32 tmp = minX;
		minX=minY;
		minY=tmp;

		tmp = maxX;
		maxX=maxY;
		maxY=tmp;
	}

	int32 minX, maxX, minY, maxY;
	int32 firstFrame, lastFrame;
	ieee64 rel;
};

struct TextDetectionData
{
	TextDetectionData () 
	{
		version = TEXT_DETECTION_VERSION;
		filename = "";
		numberOfBoxes = 0;
	} 
    TextDetectionData& operator=(const TextDetectionData &other) {
		if (this != &other) {
			version         = other.version;
			filename        = other.filename;

			box = other.box;
			numberOfBoxes = box.size();
		} // if
        return *this;
    } // operator=

	string version;
	string filename;
	int32 numberOfBoxes;

	// store (several) bounding boxes for each frame
	vector<TextDetectionBox> box;
};

std::ostream& operator<< (ostream &, const TextDetectionData&);
OTagStream& operator<< (OTagStream&, const TextDetectionData&);
ITagStream& operator>> (ITagStream&, TextDetectionData&);







struct TextArea
{
	TextArea ()
	{
		minX = -1;
		maxX = -1;
		minY = -1;
		maxY = -1;
		s="";
	}
    TextArea& operator=(const TextArea &other) {
		if (this != &other) {
			minX = other.minX;
			maxX = other.maxX;
			minY = other.minY;
			maxY = other.maxY;
			s    = other.s;
		} // if
        return *this;
    } // operator=

	int32 minX, maxX, minY, maxY;
	string s;
};
struct TextData
{
	TextData () {}
	TextData (TextDetectionData &ext) 
	{
		data.clear();
		for (int32 i=0; i<ext.box.size(); i++) {
			TextArea area;
			area.minX = ext.box[i].minX;
			area.maxX = ext.box[i].maxX;
			area.minY = ext.box[i].minY;
			area.maxY = ext.box[i].maxY;
			area.s="";

			data.push_back(area);
		} // for
	}

    TextData& operator=(const TextData &other) {
		if (this != &other) {
			data = other.data;
		} // if
        return *this;
    } // operator=

	vector<TextArea> data;
};
inline std::ostream& operator<< (ostream &os, const TextData &text) 
{
	os << text.data.size() << endl;
	for (int32 i=0; i<text.data.size(); i++) {
		os << text.data[i].minX << "   ";
		os << text.data[i].minY << "   ";
		os << text.data[i].maxX << "   ";
		os << text.data[i].maxY << "   ";
		os << text.data[i].s << endl;
	} // for
	os << endl;

	return os;
} // operator <<
inline ifstream& operator>> (ifstream& is, TextData &text)
{
    char c;
	if (is >> c) {
		char ch[128];
		is.putback(c);

		int32 count=0;
		is >> count;
		for (int32 i=0; i<count; i++) {
			TextArea area;
			is >> area.minX;
			is >> area.minY;
			is >> area.maxX;
			is >> area.maxY;
			is.getline(ch,128); 
			area.s = ch;
		} // for
	} // if

	if (is.rdstate() & ios::failbit) {
		cerr << "illegal text data file!" << endl;
	} else {
        is.clear();
    } // else
	
    return is;
} // operator>> 





/*
	Constructor: initializes variables
	
	process(): 
		if status file does not exist:
		   calculate the text areas in mpeg-file, save status file
		else:
		   read text areas from status file
*/
class TextDetection
{
public:

	TextDetection(const string &fullName, bool storeFrame=false)
	{
		pathFilename = fullName;
		store = storeFrame;
	}
	~TextDetection() {};

	void filterText (const Image<uint8> &img, Image<uint8> &region);
	void process (TextDetectionData&);
	void getTextRegion (const Image<uint8> &img, Image<uint8> &region, TextDetectionData& data);


private:
	void analyze (const Image<uint8> &src, const Image<uint8> &dx, const Image<uint8> &dy, Image<uint8> &region);
	void readValues (TextDetectionData& data);
	void storeFrame(TextDetectionData& data);

	string pathFilename;
	bool store;
};

#endif
