/*
    TextFont.cpp     load a bitmap image and generate text fonts
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


// MoCA includes
#include <TextFont.h>



// struct handles one character
CharacterData::CharacterData (Image<uint8> image, int32 x0, int32 y0, int32 x1, int32 y1, char c) 
{
	this->c = c;
	width = x1-x0+1;
	height= y1-y0+1;
	nRelPixel=0;
	
	ImageSize size(width, height,image.getSize()[DIM_BAND]);
	img.resize(size);
	SampleBlockPosition srcPos(x0,y0,0,0);
	SampleBlockPosition destPos(0,0,0,0);
	img.copy (image, srcPos, destPos);
	
#if 0
	// debug
	DrawLine line (Color::RED);
	line.process(ImagePosition(0,0),ImagePosition(width-1,0),img);
	line.process(ImagePosition(0,0),ImagePosition(0,height-1),img);
	line.process(ImagePosition(0,height-1),ImagePosition(width-1,height-1),img);
	line.process(ImagePosition(width-1,0),ImagePosition(width-1,height-1),img);
#endif
	
	// initialize relPixel with 1
	size[DIM_BAND]=1;
	relPixel.resize(size);
	relPixel.clear(1);
	
} // TextInfo

ieee64 
CharacterData::getDist(CharacterData &other)
{
	
	int32 width = MIN(this->width,other.width);
	int32 height = MIN(this->height,other.height);
	
	int32 offWidth=MAX(0,other.width-width);
	int32 offHeight=MAX(0,other.height-height);
	
	ieee64 minDist=10e30;
	for (int32 offW=0; offW<=offWidth; offW++) {
		for (int32 offH=0; offH<=offHeight; offH++) {
			ieee64 dist=0;
			for (int32 y=0; y<height; y++) {
				for (int32 x=0; x<width; x++) {
					dist += abs(this->img(x,y)-other.img(offW+x,offH+y));
				} // for
			} // for
			if (minDist > dist) minDist=dist;
		} // for
	} // for
	
	return minDist;
} // dist


ieee64 
CharacterData::getDist(CharacterData &other, int32 &px, int32 &py)
{
	
	px = 0;
	py = 0;
	
	int32 width = MIN(this->width,other.width);
	int32 height = MIN(this->height,other.height);
	
	int32 offWidth=MAX(0,other.width-width);
	int32 offHeight=MAX(0,other.height-height);
	
	ieee64 minDist=10e30;
	for (int32 offW=0; offW<=offWidth; offW++) {
		for (int32 offH=0; offH<=offHeight; offH++) {
			ieee64 dist=0;
			for (int32 y=0; y<height; y++) {
				for (int32 x=0; x<width; x++) {
					dist += abs(this->img(x,y)-other.img(offW+x,offH+y));
				} // for
			} // for
			if (minDist > dist) { 
				minDist=dist;
				px = offW;
				py = offH;
			} // if
			
		} // for
	} // for
	
	return minDist;
} // dist

void 
CharacterData::setRelevantPixels(CharacterData &other)
{
	
	int32 px,py;
	ieee64 d = getDist(other, px,py);
	
	
	nRelPixel=0;
	for (int32 y=0; y<height; y++) {
		for (int32 x=0; x<width; x++) {
			if (this->img(x,y) != other.img(px+x,py+y)) {
				nRelPixel++;
				relPixel(x,y)++;
				if (relPixel(x,y)>250) relPixel(x,y)=250;
			} // if
		} // for
	} // for
	
	return;
} // setRelevantPixels

// ----------------------------------------------------------------------------

void 
TextFontData::analyseText ()	
{
	line[0] = 0; // top line (e.g. top at characters "A" or "b")
	line[1] = 0; // main top line 
	line[2] = 0; // main bottom line 
	line[3] = 0; // bottom line (e.g. bottom at characters "q" or "g")
	
	SignalSize dySize(text[0].height + 10);
	Signal<int32> dy (dySize);
	dy.clear(0);
	
	// background color
	uint8 bc = text[0].img(0,0);
	
	for (int32 i=0; i<text.size(); i++) {
		Image<uint8> tmp = text[i].img;
		
		for (int32 y=0; y<text[i].height; y++) {
			for (int32 x=0; x<text[i].width; x++) {
				if (tmp(x,y) != bc) dy(y)++;
			} // for
		} // for
	} // for
	
#if 0	
	// debug
	for (int32 y=0; y<dySize[DIM_X]; y++) {
		cout << dy(y) << endl;
	} // for
	exit(1);
#endif
	for (int32 y=0; y<dySize[DIM_X]; y++) {
		if (dy(y) > 0) {
			line[0] = y;
			break;
		} // if
	} // for
	for (int32 y=dySize[DIM_X]-1; y>=0; y--) {
		if (dy(y) > 0) {
			line[3] = y;
			break;
		} // if
	} // for
	for (int32 y=0; y<line[3]/2; y++) {
		if (dy(y) > dy(line[1])) {
			line[1] = y;
		} // if
	} // for
	for (int32 y=line[3]/2+1; y<dySize[DIM_X]; y++) {
		if (dy(y) > dy(line[2])) {
			line[2] = y;
		} // if
	} // for
	
	// set isLargeFont
	const ieee64 thLargeFont = 0.15;
	ieee64 topLine = ((ieee64)line[1]-line[0])/line[3];
	if (topLine<thLargeFont) {
		isLargeFont = true;
	} else {
		isLargeFont = false;
	} // else
	
#if 0	
	cerr << line[0] << endl;
	cerr << line[1] << endl;
	cerr << line[2] << endl;
	cerr << line[3] << endl;
	cerr << topLine << endl;
	if (isLargeFont) {
		cerr << "isLargeFont:true"<< endl;
	} else {
		cerr << "isLargeFont:false"<< endl;
	} // else
#endif
	
	return;
} // analyseText

void 
TextFontData::draw (Image<uint8> &img, int32 x, int32 y, int32 offset)
{
	for (int32 i=0; i<text.size(); i++) {
		Image<uint8> tmp = text[i].img;
		ImagePosition p(x,y);
		img.copy(tmp, p);
		x+=text[i].width+offset;
	} // for
	
#if 0
	// debug
	DrawLine dl (Color::RED);
	ImagePosition p0 (0,line[1]+y);
	ImagePosition p1 (img.getSize()[DIM_X]-1,line[1]+y);
	ImagePosition p2 (0,line[2]+y);
	ImagePosition p3 (img.getSize()[DIM_X]-1,line[2]+y);
	dl.process(p0,p1,img);
	dl.process(p2,p3,img);
#endif
} // draw


void 
TextFontData::drawRelevance (Image<uint8> &img, int32 x, int32 y, int32 offset)
{
	for (int32 i=0; i<text.size(); i++) {
		
		ImageSize size = text[i].relPixel.getSize();
		if (size[DIM_X]>0) {
			
			Image<uint8> tmp = text[i].img;
			ImagePosition p(x,y);
			img.copy(tmp, p);
			
			Image<uint8> tmp1 = text[i].relPixel;
			for (int32 py=0; py<size[DIM_Y]; py++) {
				for (int32 px=0; px<size[DIM_X]; px++) {
					if (tmp1(px,py)>1) {
						int v = 24*tmp1(px,py);
						if (v>255) v=255;
						img (px+x,py+y,0)=v;
						img (px+x,py+y,1)=255-v;
						img (px+x,py+y,2)=v;
					} // if
				} // for
			} // for
		} // if
		x+=text[i].width+offset;
	} // for
	
#if 0
	// debug
	DrawLine dl (Color::RED);
	ImagePosition p0 (0,line[1]+y);
	ImagePosition p1 (img.getSize()[DIM_X]-1,line[1]+y);
	ImagePosition p2 (0,line[2]+y);
	ImagePosition p3 (img.getSize()[DIM_X]-1,line[2]+y);
	dl.process(p0,p1,img);
	dl.process(p2,p3,img);
#endif
} // draw

// ----------------------------------------------------------------------------




void 
TextFont::process (TextFontData& data, FontType font) 
{

	//storeFont (arial_A, arial_a, times_A, times_a);
	switch (font) {
	case ARIAL_SMALL:
		data = arial_a;
		break;
	case ARIAL_LARGE:
		data = arial_A;
		break;
	case TIMES_SMALL:
		data = times_a;
		break;
	case TIMES_LARGE:
		data = times_A;
		break;
	} // switch



	return;	
} // readValues


void 
TextFont::initializeCharacters (Image<uint8> &img, TextFontData &arial_A, TextFontData &arial_a,
		  				        TextFontData &times_A, TextFontData &times_a)
{

	BinarizeFilter<uint8> bf (128);
	bf.process(img);


	// read first text line
	int32 posY=0;
	string n1 = "arial_large";
	readCharacterLine (img, arial_A, posY, 'A', n1);
	arial_A.analyseText();

	// read second text line
	posY+=5;
	string n2 = "arial_small";
	readCharacterLine (img, arial_a, posY, 'a', n2);
	arial_a.analyseText();

	// read third text line
	posY+=5;
	string n3 = "times_large";
	readCharacterLine (img, times_A, posY, 'A', n3);
	times_A.analyseText();

	// read second text line
	posY+=5;
	string n4 = "times_small";
	readCharacterLine (img, times_a, posY, 'a', n4);
	times_a.analyseText();


	return;
} // initializeCharacters 




void 
TextFont::initializeFont(TextFontData &arial_A, TextFontData &arial_a, TextFontData &times_A, TextFontData &times_a, 
					     Image<uint8> &img, ieee64 thPercent)
{
	initializeCharacters (img, arial_A, arial_a, times_A, times_a);

	// print results
	int32 x=16;
	int32 y=95;
	arial_A.draw (img, x, y, 9);
	x=16;
	y=200;
	arial_a.draw (img, x, y, 9);



	// compare each character with each other and get distance
	const int32 fontSize = 26*4;
	ieee64 dist[fontSize][fontSize];
	ieee64 avg=0;
	for (int32 i=0; i<fontSize; i++) {

		TextFontData *first  = &arial_A;
		if (i>=26) first=&arial_a;
		if (i>=2*26) first=&times_A;
		if (i>=3*26) first=&times_a;

		for (int32 j=0; j<fontSize; j++) {
			dist[i][j]=0;

			TextFontData *second  = &arial_A;
			if (j>=26) second=&arial_a;
			if (j>=2*26) second=&times_A;
			if (j>=3*26) second=&times_a;

			dist[i][j] = first->text[i%26].getDist(second->text[j%26]);
			avg+=dist[i][j];
#if 0
			cerr << "\r" << i << " "  << j << ends;
#endif
		} // for
	} // for
	avg/=fontSize*fontSize;

	// find the most similar characters
	ieee64 percentSimilar = thPercent*fontSize*fontSize;
	ieee64 thAvg = avg*0.6;
	int32 n=fontSize*fontSize;
	while (n> percentSimilar) {
		thAvg*=0.9;
		n=0;
		for (int32 i=0; i<fontSize; i++) {
			for (int32 j=0; j<fontSize; j++) {
				if (dist[i][j]<thAvg) n++;
			} // for
		} // for
	} // while
	n=0;
	while (n< percentSimilar) {
		thAvg*=1.05;
		n=0;
		for (int32 i=0; i<fontSize; i++) {
			for (int32 j=0; j<fontSize; j++) {
				if (dist[i][j]<thAvg) n++;
			} // for
		} // for
	} // while


	// mark characters with small differences (similar characters)
	for (int32 i=0; i<fontSize; i++) {
		for (int32 j=0; j<fontSize; j++) {
			if (dist[i][j] < thAvg) {
				dist[i][j] = 1;
			} else {
				dist[i][j] = 0;
			} // else
			if (i==j) dist[i][j] = 0;
		} // for
	} // for


#if 0
    // debug
	for (int32 i=0; i<fontSize; i++) {
		for (int32 j=0; j<fontSize; j++) {
			if (dist[i][j] == 1) {
				if (i<26) {
					cout << (char)('A'+i) << "  ";
				} else {
					cout << (char)('a'+i-26) << "  ";
				} // else
				if (j<26) {
					cout << (char)('A'+j) << endl;
				} else {
					cout << (char)('a'+j-26) << endl;
				} // else
            } // if
		} // for
	} // for
#endif

	// get most important pixels for each character
	for (int32 i=0; i<fontSize; i++) {

		TextFontData *first  = &arial_A;
		if (i>=26) first=&arial_a;
		if (i>=2*26) first=&times_A;
		if (i>=3*26) first=&times_a;

		for (int32 j=0; j<fontSize; j++) {
			if (dist[i][j] == 1) {

				TextFontData *second  = &arial_A;
				if (j>=26) second=&arial_a;
				if (j>=2*26) second=&times_A;
				if (j>=3*26) second=&times_a;

			 	first->text[i%26].setRelevantPixels(second->text[j%26]);
			} // if
		} // for
	} // for

	return;
} // initializeFont


void 
TextFont::readCharacterLine (Image<uint8> &img, TextFontData &alphabet, int32 &posY, char c, string &fontname)
{
	alphabet.fontname = fontname;

	// find exact position of top and bottom line
	ImageSize size = img.getSize();
	uint8 bc = img(0,0);
	uint8 fc=0;
	int32 px = -1;
	int32 py = -1;
	for (int32 y=posY; y<size[DIM_Y]; y+=4) {
		for (int32 x=0; x<size[DIM_X]; x++) {
			if (img(x,y) != bc) {
				px = x;
				py = y;
				fc = img(x,y);
				x+=size[DIM_X];
				y+=size[DIM_Y];
			} // if
		} // for
	} // for


	// locate top and bottom position of text line
	int32 top=py-1;
	int32 bottom=py+1;
	bool fcFound = true;
	while (fcFound) {
		fcFound=false;
		for (int32 x=0; x<size[DIM_X]; x++) {
			if (img(x,top) == fc) fcFound=true;
		} // for
		top--;
	} // for
	fcFound=true;
	while (fcFound) {
		fcFound=false;
		for (int32 x=0; x<size[DIM_X]; x++) {
			if (img(x,bottom) == fc) fcFound=true;
		} // for
		bottom++;
	} // for


	// find characters
	Signal<bool> isText(size[DIM_X]);
	isText.clear(false);
	for (int32 x=0; x<size[DIM_X]; x++) {
		for (int32 y=top+1; y<bottom; y++) {
			if (img(x,y) == fc) {
				isText(x)=true;
				break;
			} // if
		} // for
	} // for

	int32 left = -1;
	int32 right = -1;
	for (int32 x=0; x<size[DIM_X]; x++) {
		if (isText(x)) {
			left = x-2;
			while((x<size[DIM_X]) && isText(x)) x++;
			right=x+1;

			CharacterData t (img, left, top, right, bottom, c);
			alphabet.add (t);
			c++;
		} // if
	} // for

	posY=bottom;

	return;
} // readCharacterLine 



