/*
   This file is part of the MoCA video library.
   Name: TextFont.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: kopf@informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/*! \file

 
  
   \author    Stephan Kopf
   
	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de
	
	 (c) Lehrstuhl fuer Praktische Informatik IV, details in COPYRIGHT.txt
	 
*/


#ifndef TEXT_FONT_H
#define TEXT_FONT_H


// C++ includes
#include <iomanip>
#include <sstream>
#include <iostream>
using namespace std;


// MoCA includes
#include <Image.h>
#include <IImageStream.h>
#include <BinarizeFilter.h>


enum FontType {
    ARIAL_SMALL   = 1,
    ARIAL_LARGE   = 2,
    TIMES_SMALL   = 3,
    TIMES_LARGE   = 4,
};


// struct handles one character
struct CharacterData
{
public:
	CharacterData () 
	{
		c = ' ';
		width = 0;
		height = 0;
		nRelPixel = 0;
	}
	CharacterData (Image<uint8> image, int32 x0, int32 y0, int32 x1, int32 y1, char c);

	
    CharacterData& operator=(const CharacterData &other) {
		if (this != &other) {
			c        = other.c;
			width    = other.width;
			height   = other.height;
			img      = other.img;
			relPixel = other.relPixel;
			nRelPixel= other.nRelPixel;
		} // if
        return *this;
    } // operator=


	// additional functions for characters
	ieee64 getDist(CharacterData &other);
	ieee64 getDist(CharacterData &other, int32 &px, int32 &py);
	void setRelevantPixels(CharacterData &other);

	char c;
	int32 width;
	int32 height;
	Image<uint8> img;
	Image<uint8> relPixel;
	int32 nRelPixel;
};




// struct handles a list of characters (one alphabet)
struct TextFontData
{
public:
	TextFontData () 
	{
		line[0] = 0;
		line[1] = 0;
		line[2] = 0;
		line[3] = 0;
		text.clear();
		fontname = "undefined";
		isLargeFont = true;
	}

	void add (CharacterData &t) 
	{
		text.push_back(t);
	} // add

	// functions
	void analyseText ();
	void draw (Image<uint8> &img, int32 x, int32 y, int32 offset);
	void drawRelevance (Image<uint8> &img, int32 x, int32 y, int32 offset);


	int16 line[4];
	vector<CharacterData> text;
	string fontname;
	bool isLargeFont;
};






/*
	Constructor: initializes variables (filename og image with fonts)
	
	process(): 
		read the specified font
*/
class TextFont
{
public:

	// filename with font data
	TextFont(const string &fullName)
	{
		pathFilename = fullName;

		// read image
		Image<uint8> img;
		ReadImage (pathFilename, img);

		// read fonts
		ieee64 thSimilar = 0.03;
		initializeFont (arial_A, arial_a, times_A, times_a, img, thSimilar);

	}
	~TextFont() {};

	void process (TextFontData& data, FontType type = ARIAL_LARGE);

private:
	void initializeFont(TextFontData &arial_A, TextFontData &arial_a, TextFontData &times_A, TextFontData &times_a, 
					    Image<uint8> &img, ieee64 thPercent);
	void initializeCharacters (Image<uint8> &img, TextFontData &arial_A, TextFontData &arial_a,
		  				       TextFontData &times_A, TextFontData &times_a);
	void readCharacterLine (Image<uint8> &img, TextFontData &alphabet, int32 &posY, char c, string &fontname);


	string pathFilename;

	// data with fonts
	TextFontData arial_A, arial_a, times_A, times_a;

};

#endif
