/*
   This file is part of the MoCA video library.
   Name: ConvertFilter.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: {kuehne,kopf}@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


/* -*- C++ -*-

   ConvertFilter.h  -- convert an image, e. g. from uint8 to ieee64.
       
   
   Author:    Stephan Kopf
              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim, Germany
              email: {kuehne,kopf}@pi4.informatik.uni-mannheim.de
*/



#ifndef CONVERT_FILTER_H
#define CONVERT_FILTER_H

#include <SampleBlock.h>
#include <GetType.h>


// CROP   : Keep all values if possible; the value is cropped if it is too low/high 
// STRETCH: Stretching the values to the range of the type. For float types the range must be specified, e.g. [0;1].
//
// Example:
//
//     ConvertFilter<uint8,ieee64> cf(CONVERT_STRETCH, 0, 1);
//     converts image src(0..255) to dest (0..1)
//
//     ConvertFilter<ieee64,int16> cf(CONVERT_STRETCH);
//     converts image src(0..1) to dest (0..+32K)
//
//     ConvertFilter<int16,double> cf(CONVERT_CROP);
//     converts image src(0..16K) to dest (0..1), nearly all values will be cropped!!!

enum ConvertFilterMode {
    CONVERT_CROP,
    CONVERT_SCALE
};

template<class S, class D>
class ConvertFilter 
{
public:
	// constructor can be used for int type only
    ConvertFilter(const ConvertFilterMode &convertMode)
    {
        mode = convertMode;
		srcMin=0;
		srcMax=1;
	    destMin=0;
	    destMax=1;
		bool srcIsFloat = !GetType::isInteger(srcMin);
		bool destIsFloat= !GetType::isInteger(destMin);

		if ((srcIsFloat) || (destIsFloat)) {
			cerr << "ConvertFilter::error, please specify the range of float type in constructor!" << endl;
			exit(1);
		} // if

		srcMin  = GetType::getMin(srcMin);
		srcMax  = GetType::getMax(srcMin);
		destMin = GetType::getMin(destMin);
		destMax = GetType::getMax(destMin);
    };

	// constructor can be used for one float type only, the range of the float is specified
    ConvertFilter(const ConvertFilterMode &convertMode, ieee64 minV, ieee64 maxV)
    {

        mode = convertMode;
	    srcMin=0;
	    srcMax=1;
	    destMin=0;
	    destMax=1;
		bool srcIsFloat = !GetType::isInteger(srcMin);
		bool destIsFloat= !GetType::isInteger(destMin);

		if ((srcIsFloat) && (destIsFloat)) {
			cerr << "ConvertFilter::error, please specify the range of both(!) float-types in constructor!" << endl;
			exit(1);
		} // if
		if ((!srcIsFloat) && (!destIsFloat)) {
			cerr << "ConvertFilter::error, please specify the range of both(!) int-types in constructor!" << endl;
			exit(1);
		} // if
		if (srcIsFloat) {
			srcMin  = (S)minV;
			srcMax  = (S)maxV;
			destMin  = GetType::getMin(destMin);
			destMax  = GetType::getMax(destMin);
		} else {
			srcMin  = GetType::getMin(srcMin);
			srcMax  = GetType::getMax(srcMin);
			destMin  = (D)minV;
			destMax  = (D)maxV;
		} // else
    };

	// constructor can be used for float and int types, the minimum and maximum values are specified
    ConvertFilter(const ConvertFilterMode &convertMode, ieee64 srcMinV, ieee64 srcMaxV, ieee64 destMinV, ieee64 destMaxV)
    {

        mode = convertMode;
		srcMin  = (S)srcMinV;
		srcMax  = (S)srcMaxV;
		destMin  = (D)destMinV;
		destMax  = (D)destMaxV;
    };


    ~ConvertFilter () {};

    //! convert image in different type
    void process (const SampleBlock<S>& src, SampleBlock<D>& dest);

private:
    ConvertFilterMode mode;    
	S srcMin;
	S srcMax;
	D destMin;
	D destMax;
}; // ConvertFilter

#include <ConvertFilter.cpt>

#endif

