/*
    FaceMaskFilter.cpp     get face regions in images
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <FaceMaskFilter.h>

const int FaceMaskFilter::FACE_MASK_COLOR = 1;
const int FaceMaskFilter::FACE_MASK_EDGES = 2;
const int FaceMaskFilter::FACE_MASK_COLOREYES = 4;


void
FaceMaskFilter::
process (Image<uint8>& srcDest)
{
    Image<uint8> dest;
    process (srcDest, dest);
    srcDest = dest;
    return;
} // process


void
FaceMaskFilter::
process (Image<uint8>& src, Image<uint8>& dest)
{
    ImageSize size = src.getSize();
    if (size[DIM_BAND] != 3) {
        cerr << "Error in FaceMaskFilter::process(src [, dest])." << endl;
        cerr << "Src image requires 3 bands (RBG image)." << endl;
        exit(1);
    } // if

    // adjust dest image
    size[DIM_BAND] = 1;
	dest.resize(size);
	dest.clear(255);

	// create face mask
    if (mode & FACE_MASK_COLOR) {
        analyseColor (src, dest);  
	} // if
    if (mode & FACE_MASK_EDGES) {
        analyseEdges (src, dest);
	} // if
    if (mode & FACE_MASK_COLOREYES) {
        analyseColorEyes (src, dest);
	} // if

    return;
} // process



// Input image must be RGB with uint8
void
FaceMaskFilter::
analyseColor (Image<uint8>& src, Image<uint8>& dest) 
{
    ImageSize size = src.getSize();    
    sizeType x,y;
    ieee64 sum, red, green;

    for (y=0; y<size[DIM_Y]; y++) {
        for (x=0; x<size[DIM_X]; x++) {
            sum   = 1.0 / (src(x,y,0) + src(x,y,1) + src(x,y,2) + 10e-10);
            red   = src(x,y,0) * sum; 
            green = src(x,y,1) * sum; 

            // check for face-like color
            if ((red >= 0.37) && (red <= 0.58) && (green >= 0.26) && (green <= 0.36)) {
				// everything okay (is face pixel)
			} else {
				// remove face pixel
                dest(x,y) = 0;
			} // else
        } // for
    } // for

    int thRadius = 3;
    DilationFilter<uint8> df(thRadius);
    df.process (dest);

	return;
} // analyseColor



// Input image may be grayscale image with uint8
void
FaceMaskFilter::
analyseEdges (Image<uint8>& src, Image<uint8>& dest) 
{

    Image<uint8> img;

    GrayscaleFilter<uint8> gf;
    gf.process (src, img);

    CannyFilter<uint8> cf;   
    cf.process (img);

    uint8 thBinarize = 30;  // thBinarize = 20
    BinarizeFilter<uint8> bf(thBinarize);
    bf.process (img);

    int thRadius = 5;
    DilationFilter<uint8> df(thRadius);
    df.process (img);

    AndFilter<uint8> af;
    af.process (img, dest);
    
    return;
} // analyseEdges




// Input image may be grayscale image with uint8
void
FaceMaskFilter::
analyseColorEyes (Image<uint8>& src, Image<uint8>& dest) 
{

    Image<uint8> img;
	ImageSize size = src.getSize();

    GrayscaleFilter<uint8> gf;
    gf.process (src, img);

    OpeningFilter<uint8> of;
    of.process (img);

	// enhance contrast
	EdgeDetectionFilter<uint8> edf;
    Image<uint8> sobel;
	edf.filterSobel (img, sobel);
	ieee64 v;
	ieee64 alpha = 0.5;
	//ieee64 mult  = 255; //(11-8);

	for (int y=0; y<size[DIM_Y]; y++) {
		for (int x=0; x<size[DIM_X]; x++) {
			v = alpha * sobel(x,y) + (1-alpha) * img(x,y);
			//cout << v << endl;
			if (v>255) {
				v = 255;
				cout << "+" << flush;
			} // if
			if (v<0) {
				v = 0;
				cout << "-" << flush;
			} // if
			img(x,y) = (uint8) v;
		} // for
	} // for

	dest = img;

    return;
} // analyseEdges



