/*
   This file is part of the MoCA video library.
   Name: ScaleFilter.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: {kuehne,kopf}@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#ifndef SCALE_FILTER_H
#define SCALE_FILTER_H

/* -*- C++ -*-

   ScaleFilter.h  -- implementation of an image scaling routine

   Author:    Stephan Kopf
              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim, Germany
              email: {kuehne,kopf}@pi4.informatik.uni-mannheim.de

*/

// TODO convert to standard MoCA style, e. g. filter()


// C++ includes
#include <iostream>
#include <cmath>

// MoCA includes
#include <ImageFilterBase.h>
#include <Image.h>

/*
  Scales the row- and column-dimensions of an image by separate
  real-valued factors or to a destination size using bilinear
  interpolation.  If necessary, the image to be scaled will be reduced
  to an appropriate size before bilinear interpolation.
*/

enum {SCALE_REAL_FACTOR, SCALE_INT_FACTOR, SCALE_SIZE};

template<class T>
class ScaleFilter : public ImageFilterBase<T>
{
public:

    //! constructor, image is scaled by seperate real-valued factors
    // Scales the row- and column-dimensions of an image by separate
    // real-valued factors using bilinear interpolation.
    // If necessary, the image to be scaled will be reduced to an 
    // appropriate size before bilinear interpolation.
    ScaleFilter(double scaleX, double scaleY):
            mode          (SCALE_REAL_FACTOR),
            row_scale     (scaleY),
            column_scale  (scaleX),
            row_factor    (0),
            column_factor (0),
            sizeDest      (ImageSize(0,0)) {};

	//! constructor, image is scaled fast by seperate integer-valued factor; no interpolation
	//! use functions reduce or expand with this constructor
    ScaleFilter():
            mode          (SCALE_INT_FACTOR),
            row_scale     (0),
            column_scale  (0),
            row_factor    (-1),
            column_factor (-1),
            sizeDest      (ImageSize(0,0)) {};

	//! constructor, image is scaled to the defined size 
    // Scales the row- and column-dimensions of an image by separate
    // real-valued factors using bilinear interpolation.
    // If necessary, the image to be scaled will be reduced to an 
    // appropriate size before bilinear interpolation.
    ScaleFilter(ImageSize size):
            mode          (SCALE_SIZE),
            row_scale     (0),
            column_scale  (0),
            row_factor    (0),
            column_factor (0),
            sizeDest      (size) {};


    //! the filter
    void process (Image<T>& srcDest)
	{
		Image<T> tmp = srcDest;
		process (tmp, srcDest);
		return;
	} // process

    //! the filter
    void process (const Image<T>& src, Image<T>& dest);

    // scale src down fast to dest by integer-valued factor
    void reduce (Image<T>& srcDest, int row_scale, int column_scale);
    void reduce (const Image<T>& src, Image<T>& dest, int row_scale, int column_scale);

    // scale src up fast to dest by integer-valued factor
    void expand (Image<T>& srcDest, int row_scale, int column_scale);
    void expand (const Image<T>& src, Image<T>& dest, int row_scale, int column_scale);
	
private:
    // Perform scaling in the row- and column-dimensions of an image using bilinear interpolation.
    Image<T>& BiLinearScale (const Image<T>& src, Image<T>& dest, int dest_nrows, int dest_ncolumns);
  
    // rounds double value to nearest value of type T
    inline T rint (double d) { return (T) (d + 0.5); };

	
	// Returns the value of a specified source pixel. If the specified
	// pixel does not exist, then return zero.
    inline
    T GetSrcPixel(const Image<T>& src, int band, int row, int column) {
        ImageSize ts = src.getSize(); 
        return (row >= ts[DIM_Y] || column >= ts[DIM_X]) ? (T) 0 : src (column,row,band);
    } // GetSrcPixel
    
    int mode;
    ieee64 row_scale;
    ieee64 column_scale;
    int row_factor;
    int column_factor;
    ImageSize sizeDest;       
}; // ScaleFilter

#include <ScaleFilter.cpt>


#endif // SCALE_FILTER_H
