/*
    AudioDecoder.cpp     decode the audio
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <AudioDecoder.h>


std::ostream& operator<< (ostream& os, const AudioDecoderData &data)
{
	os << "version        : " << data.version << endl;
	os << "filename       : " << data.filename<< endl;
	os << "audio filename : " << data.audioFilename<< endl;
	if (data.hasAudio) {
		os << "audio track    : found" << endl;
	} else {
		os << "audio track    : not available" << endl;
	} // else
	os << "audio parameters:" << endl;
	os << data.audioData << endl;
	return os;
} // operator<<


OTagStream& operator<< (OTagStream& os, const AudioDecoderData& data)
{
    TagFormat::writeNodeValue(os, "version", data.version);
    TagFormat::writeNodeValue(os, "filename", data.filename);
    TagFormat::writeNodeValue(os, "audioFilename", data.audioFilename);
    TagFormat::writeNodeValue(os, "hasAudio", data.hasAudio);
    TagFormat::writeNodeValue(os, "hasAudio", data.hasAudio);
	os << data.audioData;
	os << endl;
    return os;
} // operator<<

ITagStream& operator>> (ITagStream& is, AudioDecoderData& data)
{
    char c;
    bool failFlag = false;

	if (is >> c) {
		is.putback(c);

		TagFormat::readNodeValue(is, "version", data.version);
		TagFormat::readNodeValue(is, "filename", data.filename);
		TagFormat::readNodeValue(is, "audioFilename", data.audioFilename);
		TagFormat::readNodeValue(is, "hasAudio", data.hasAudio);
		is >> data.audioData;

		if (is.rdstate() & ios::failbit) {
			failFlag = true;
		} // if

	} // if

    if (failFlag) {
		throw MocaException ("AudioDecoder: Illegal format of audio index file!");
	} else {
        is.clear();
    } 
    return is;
} // operator>> 



// constuctor specifying name of MPEG/AVI file
AudioDecoder::AudioDecoder (string &filename)
{	
	// modify filename, if spaces in filename/path exist
	pathFilename = filename;
	CreateMocaPath pathCheck (pathFilename, false);
	pathCheck.validatePath (pathFilename);
	audioPathFilename = "";

	// get mpeg index information
	VideoIndex reader (pathFilename);
	reader.process(index);
	videodecodingExecutable = reader.getVideoDecoderCmd();

	fh_in = NULL;

#ifdef COMPILER_MSVC
	//  Switch Win32 standard input to binary mode, so that piping the
	//  raw image data works. 
	int result = _setmode( _fileno(stdin), _O_BINARY );
	if (result == -1) {
		cerr << "Cannot set standard input to binary mode.\n";
		exit(1);
	}
#endif
} // AudioDecoder



void AudioDecoder::close()
{
	try {
		if (fh_in != NULL) {
			int error = pclose(fh_in);
			if (error == -1) {
				cerr << "error while closing process: "<< videodecodingExecutable << endl;
			} // if
		} // if
	} catch (...) {
		cerr << "error while closing process: "<< videodecodingExecutable << endl;
	} // catch
	fh_in = NULL;
	return;
}

void AudioDecoder::open(string &cmd)
{
	// close fh_in
	if (fh_in != NULL) {
		close();
	} // if
	fh_in = NULL;

	// open process
	OpenProcess::open(&fh_in, cmd, P_READ, true);
	return;	
}

void AudioDecoder::process(AudioDecoderData &data)
{
	// set audioPathFilename
	CreateMocaPath mocaPath (pathFilename, false);
	string fn      = mocaPath.file;
	string audioFn = fn.substr(0, fn.length()-4) + ".wav";
	audioPathFilename = mocaPath.destinationPath + audioFn;

	
	// read data
	ReadMocaData<AudioDecoderData> reader (pathFilename, AUDIO_DECODER_FILENAME);
	bool dataValid = reader.process (AUDIO_DECODER_VERSION, data); 
	
	if (dataValid) {
		if (data.hasAudio) {
			if (!FileHandling::fileExists (audioPathFilename)) {
				dataValid=false;
			} // if
		} // if

		// check size of WAV file
		int32 fileSize = 0;
		FileHandling::getFileSize (audioPathFilename, fileSize);
		if (data.audioData.riffLength +8 != fileSize) {
			dataValid=false;
		} // if

	} // if

	if (!dataValid) {
		
		// create new index
		cerr << "\rAudioDecoder: Creating index file ...          " << ends;
		readValues(data);
		
		// save index file
		WriteMocaData<AudioDecoderData> writer (pathFilename, AUDIO_DECODER_FILENAME);
		writer.process(data);
		cerr << "\rAudioDecoder: Creating index file ... done          " << endl;
	} // if

	return;
}


void AudioDecoder::readValues(AudioDecoderData &data)
{
	// initialize values
	data.version = AUDIO_DECODER_VERSION;
	data.hasAudio = false;

	CreateMocaPath mocaPath1 (pathFilename, false);
	data.filename = mocaPath1.file;
	CreateMocaPath mocaPath2 (audioPathFilename, false);
	data.audioFilename = mocaPath2.file;

	WaveHeader audio = index.audioData; 
	if ((audio.channels      >0) && 
		(audio.samplesPerSec >0) && 
		(audio.bitsPerSample >0) && 
		(audio.dataLength    >0)) {
		data.hasAudio = true;
	} // if

	if (data.hasAudio) {

		// open WAV writer
		OWaveAudio writer (
			audioPathFilename, 
			audio.channels, 
			audio.samplesPerSec, 
			audio.bitsPerSample, 
			audio.getNumberOfSamples() );
		
		// extract audio track
		bool error=false;
		try {
			// open file handler
			string cmd = videodecodingExecutable + " -vc dummy -vo null -ao pcm -aofile - -really-quiet " + pathFilename;
			open(cmd);
			
			// skip status information and header 
			bool dataFound = false;
			try {
				char buf[2];
				bool error = true;
				if (fh_in != NULL) {
					error=false;	
			
					// read next character
					while (!feof(fh_in)) {
				
						// search "WAVEfmt " ... "data"
						fread (buf, 1, 1, fh_in); if (buf[0]=='W') {

							fread (buf, 1, 1, fh_in); if (buf[0]=='A') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='V') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='E') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='f') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='m') {
							fread (buf, 1, 1, fh_in); if (buf[0]=='t') {
							fread (buf, 1, 1, fh_in); if (buf[0]==' ') {

								// skip 20 bytes
								for (int i=0; i<20; i++) fread (buf, 1, 1, fh_in); 
								fread (buf, 1, 1, fh_in); if (buf[0]=='d') {
								fread (buf, 1, 1, fh_in); if (buf[0]=='a') {
								fread (buf, 1, 1, fh_in); if (buf[0]=='t') {
								fread (buf, 1, 1, fh_in); if (buf[0]=='a') {
									for (int i=0; i<4; i++) fread (buf, 1, 1, fh_in); 
									dataFound = true;
									break;
							}}}}}}}}}}}
						}
					} // while
			
				}
			}  catch (...) {
				error=true;
			} // catch

			// write data
			if ((dataFound) && (!error)) {
				writer.write(fh_in);
			} // if

		}  catch (...) {
			error=true;
		} // catch

		if (error) {
			cerr << "AudioDecoder: could not extract audio track!" << endl << endl;
		} // if
	} // if

	return;
} // readValues
