/*
    VideoEncoder.cpp     encode a video
    Copyright (C) 2004 Stephan Kopf

	Lehrstuhl fuer Praktische Informatik IV
	Universitaet Mannheim
	68131 Mannheim
	Germany
	email: kopf@informatik.uni-mannheim.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <VideoEncoder.h>


// constuctor setting filename
VideoEncoder::VideoEncoder (const VideoEncoderOptions &options)
{
#ifdef COMPILER_MSVC
	//  Switch Win32 standard output to binary mode, so that piping the
	//  raw image data works. 
	{
		int result = _setmode( _fileno(stdout), _O_BINARY );
		if (result == -1)
		{
			cerr << "Cannot set standard output to binary mode.\n";
			exit(1);
		}
	}
#endif

	// initialize parameters
	this->options = options;

	// initialize variables
	fh_out = NULL;

	error = 0;
#ifdef _WIN32
	cmd = WINDOWS_VIDEO_ENCODER_EXECUTABLE;
#else
	cmd = LINUX_VIDEO_ENCODER_EXECUTABLE;
#endif

	open();
} // VideoEncoder::VideoEncoder 


// destructor
VideoEncoder::~VideoEncoder ()
{
	close();
} // VideoEncoder::~VideoEncoder 


// copy buffer to stdout
void
VideoEncoder::writeData(const unsigned char *buf, int size)
{
	// stream data to videoencoder
	int bytesOut = 0;
	int n;
	int out;
	error = 0;
	while (bytesOut < size) {
		
		n = 128;
		if (size - bytesOut < 128) {
			n = size - bytesOut;
		} // if
		
		out = 0;
		out = fwrite(&buf[bytesOut],1,n, fh_out);
		if (out == n) {
			error=0;
		} else {

			if (out == 0) {
				error++;
				if (error == 5) {
					cerr << "VideoEncoder: warning, could not all bytes in buffer, trying again..." << endl;
				} // if
				if (error > 50) {
					cerr << endl << endl;
					cerr << "VideoEncoder::writeData: could not write data." << endl << endl;
					cerr << "executable " << cmd << " probably not available." << endl << endl;
					exit(1);
				} // else
			} // else
		} // if
		bytesOut += out;	
	} // while

	// flush buffer
	out = fflush (fh_out);

	return;
} // VideoEncoder::writeData

	
// write frame (the frame will be croped)
void 
VideoEncoder::write(const Image<uint8> &img)
{
	Image<uint8> dest;
	ConvertColorSpace::process(options.imgCodec, CODEC_YUV420, img, dest);

	// write Y plane
	ImageSize size = img.getSize();
	int32 dx = size[DIM_X];
	int32 dy = size[DIM_Y];
	const uint8* buf = dest.getSamplePtr(0,0,0);
	writeData (buf, dx*dy);

	// write U plane
	dx /= 2;
	dy /= 2;
	for (int32 y=0; y<dy; y++) {
		const uint8* buf = dest.getSamplePtr(0,y,1);
		writeData (buf, dx);
	} // for

	// write V plane
	for (int32 y=0; y<dy; y++) {
		const uint8* buf = dest.getSamplePtr(0,y,2);
		writeData (buf, dx);
	} // for

	return;
} // VideoEncoder::write


void VideoEncoder::open()
{
	close();

	int sizeX = options.size[DIM_X];
	int sizeY = options.size[DIM_Y];
	if ((sizeX%2 == 1) || (sizeY%2 == 1)) {
		cerr << endl 
			 << "VideoEncoder: the resolution of the video should be even (coding errors may occur)!" 
			 << endl;
	} // if
	if ((sizeX == 0) || (sizeY == 0)) {
		cerr << endl 
			 << "VideoEncoder: the resolution of the video is undefined!" 
			 << endl;
		exit(1);
	} // if

	// modify filename, if spaces in filename/path exist
	string tmpFilename=options.pathFilename;
	CreateMocaPath pathCheck (tmpFilename, false);
	pathCheck.validatePath (tmpFilename);


	// set image size
	stringstream strTxt;
	strTxt << " -rawvideo on:w=" << sizeX << ":h=" << sizeY;
	string sTxt = strTxt.str();

	// set codec
	string codecNameTxt = "";
	switch (options.encCodec) {
	case CODEC_WMV_1:
		codecNameTxt = "wmv1";
		break;
	case CODEC_WMV_2:
		codecNameTxt = "wmv2";
		break;
	case CODEC_MPEG1:
		codecNameTxt = "mpeg1video";
		break;
	case CODEC_MPEG2:
		codecNameTxt = "mpeg2video";
		break;
	case CODEC_MPEG4:
		codecNameTxt = "mpeg4";
		break;
	default:
		codecNameTxt = "mpeg4";
		break;
	}
	stringstream strCodec;
	strCodec << " -ovc lavc -lavcopts vcodec=" << codecNameTxt;	  // set codec

	// set codec options (every 132th frame is a key frame)
	strCodec << ":keyint=132"; 

	// set codec options (grayscale)
	if (options.createGrayscale) strCodec << ":gray"; 
	
	// set codec options (bitrate)
	if (options.bitrate > 0) {
		int32 nBitrate = (int32)options.bitrate;
		if (nBitrate < 4) {
			nBitrate = 4;
			cerr << "Warning in VideoEncoder::open: bitrate should be larger than 4!" << endl;
		} // if
		if (nBitrate > 16000) {
			nBitrate = 16000;
			cerr << "Warning in VideoEncoder::open: bitrate should be smaller than 16000!" << endl;
		} // if
		string highQ = "";
		if (options.quality == QUALITY_HIGH) {
			highQ = ":vhq";
		} // if
		strCodec << ":vbitrate=" << nBitrate << highQ;
	} // if
	string sCodec = strCodec.str();


	// set additional options (framerate)
	stringstream strAddOptions;
	if (options.framerate > 0) {
		strAddOptions << " -fps " << options.framerate << " ";
	} // if
	string sAddOptions = strAddOptions.str();


	// disable output (quiet)
	string quiet = " -quiet ";


	stringstream ossCmd;
	ossCmd 
		<< cmd						// videodecoder
		<< quiet					// disable output
		<< sTxt						// image resolution
		<< sCodec					// set codec 
		<< sAddOptions				// set additional options
		<< " -o " << tmpFilename    // set output filename
		<< " - ";					// stream to stdout (decoder reads stdin)
	string sCmd = ossCmd.str();

	if (!options.silent) {
		cerr << "executing command: " << endl << sCmd << endl;
	} // if
	fh_out = NULL;

	// open encoder
#if 1
	OpenProcess::open(&fh_out, sCmd, P_WRITE);
#else
	// store result as file
	cerr << "output to file" << endl;
	string ffn = "D:/test.yuv";
	OpenFile::process(&fh_out, ffn, F_WRITE);
#endif

	return;	
} // VideoEncoder::open


void 
VideoEncoder::close()
{
	int error = fflush (fh_out);
	if (error != 0) {
		cerr << endl << "VideoEncoder: could not flush buffer" << endl << endl;
	} // if

	if (fh_out != NULL) {
#if 1
		pclose(fh_out);
#else
		fclose(fh_out);
#endif
	} // if
	fh_out = NULL;

	return;
} // VideoEncoder::close


