/*
   This file is part of the MoCA video library.
   Name: Video.h
   Copyright (C) 2004 Stephan Kopf

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: {kuehne,kopf}@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

/* -*- C++ -*-

   file.h  -- short description

   Author:    Gerald Kuehne
              Stephan Kopf
              Lehrstuhl fuer Praktische Informatik IV
              Universitaet Mannheim, Germany
              email: {kuehne,kopf}@pi4.informatik.uni-mannheim.de

   $Id: Video.h,v 1.2 2004/03/05 15:21:24 kopf Exp $
*/

#ifndef VIDEO_H
#define VIDEO_H

// C++ includes
#include <list>

// Moca includes
#include <SampleBlock.h>
#include <VideoSize.h>
#include <VideoPosition.h>
#include <Image.h>

template<class S> class SubVideo;


template<class T> class Video : public SampleBlock<T>
{
#ifdef COMPILER_MSVC
	friend class SubVideo;
#else
    template<class S> friend class SubVideo;
#endif

public:
    // Construction/Destruction
    Video (const VideoSize& = VideoSize (0, 0, 0, 0),
           const VideoSize& = VideoSize (0, 0, 0, 0),
           BorderMode = BORDER_REPLICATE);
    Video (const Video&);

    virtual ~Video ();

    Video& operator= (const Video&);

    T& operator() (indexType, indexType, indexType, indexType);
    T& operator() (indexType, indexType, indexType);
    const T& operator() (indexType, indexType, indexType, indexType) const;
    const T& operator() (indexType, indexType, indexType) const;

    // Interface
    void resize  (const VideoSize&);
    void realloc (const VideoSize&, const VideoSize&, BorderMode);

    VideoSize getSize () const;
    ImageSize getImageSize () const;
    VideoSize getTileSize () const;

    T* getSamplePtr (indexType, indexType, indexType, indexType);
    T* getSamplePtr (indexType, indexType, indexType);
    const T* getSamplePtr (indexType, indexType, indexType, indexType) const;
    const T* getSamplePtr (indexType, indexType, indexType) const;

    T* getTilePtr   (indexType, indexType, indexType, indexType) const;
    T* getTilePtr   (indexType, indexType, indexType) const;

    //! get all neighbours
    list<VideoPosition>
    getNeighbours(const VideoPosition &pos, const int ng) const
        throw (MocaException);

    //! get a frame of the video
    Image<T> getImage(indexType) const;

    //! set a frame of the video
    void setImage(indexType, const Image<T> &) throw (MocaException);

}; // Video

template<class T>
inline
Video<T>::
Video (const VideoSize& size, const VideoSize& tileSize, BorderMode borderMode)
  : SampleBlock<T> (size, tileSize, borderMode)
{}

template<class T>
inline
Video<T>::
Video (const Video& other)
  : SampleBlock<T> (other)
{}

template<class T>
Video<T>::
~Video ()
{}

template<class T>
inline
Video<T>&
Video<T>::operator= (const Video<T>& other)
{
  SampleBlock<T>::operator= (other);
  return *this;
}

template<class T>
inline
T&
Video<T>::
operator() (indexType x, indexType y, indexType t, indexType band)
{
  return (T&) *(this->sampleIndexX[x] + this->sampleIndexY[y] + this->sampleIndexZ[t] +
                this->sampleIndexBand[band]);
}

template<class T>
inline
T&
Video<T>::
operator() (indexType x, indexType y, indexType t)
{
  return (T&) *(this->sampleIndexX[x] + this->sampleIndexY[y] + this->sampleIndexZ[t]);
}

template<class T>
inline
const T&
Video<T>::
operator() (indexType x, indexType y, indexType t, indexType band) const
{
  return (T&) *(this->sampleIndexX[x] + this->sampleIndexY[y] + this->sampleIndexZ[t] +
                this->sampleIndexBand[band]);
}

template<class T>
inline
const T&
Video<T>::
operator() (indexType x, indexType y, indexType t) const
{
  return (T&) *(this->sampleIndexX[x] + this->sampleIndexY[y] + this->sampleIndexZ[t]);
}


template<class T>
inline
void
Video<T>::
resize (const VideoSize& size)
{
  SampleBlock<T>::resize (size);
}


template<class T>
inline
void
Video<T>::
realloc (const VideoSize& size, const VideoSize& tileSize,
         BorderMode borderMode)
{
  SampleBlock<T>::realloc (size, tileSize, borderMode);
}

template<class T>
inline
VideoSize
Video<T>::
getSize () const
{
  return VideoSize (this->size[DIM_X], this->size[DIM_Y], this->size[DIM_Z], this->size[DIM_BAND]);
}

template<class T>
inline
ImageSize 
Video<T>::
getImageSize () const
{
  return ImageSize (this->size[DIM_X], this->size[DIM_Y], this->size[DIM_BAND]);
}

template<class T>
inline
VideoSize
Video<T>::
getTileSize () const
{
  return VideoSize (this->tileSize[DIM_X], this->tileSize[DIM_Y], this->tileSize[DIM_Z],
                    this->tileSize[DIM_BAND]);
}

template<class T>
inline
T*
Video<T>::
getSamplePtr (indexType x, indexType y, indexType t, indexType band)
{
  return (T*) (this->sampleIndexX[x] + this->sampleIndexY[y] + this->sampleIndexZ[t] + this->sampleIndexBand[band]);
}

template<class T>
inline
T*
Video<T>::
getSamplePtr (indexType x, indexType y, indexType t)
{
  return (T*) (this->sampleIndexX[x] + this->sampleIndexY[y] + this->sampleIndexZ[t]);
}

template<class T>
inline
const T*
Video<T>::
getSamplePtr (indexType x, indexType y, indexType t, indexType band) const
{
  return (T*) (this->sampleIndexX[x] + this->sampleIndexY[y] + this->sampleIndexZ[t] + this->sampleIndexBand[band]);
}

template<class T>
inline
const T*
Video<T>::
getSamplePtr (indexType x, indexType y, indexType t) const
{
  return (T*) (this->sampleIndexX[x] + this->sampleIndexY[y] + this->sampleIndexZ[t]);
}

template<class T>
inline
T*
Video<T>::
getTilePtr (indexType x, indexType y, indexType t, indexType band) const
{
  return (T*) (this->tileIndexX[x] + this->tileIndexY[y] + this->tileIndexZ[t] + this->tileIndexBand[band]);
}

template<class T>
inline
T*
Video<T>::
getTilePtr (indexType x, indexType y, indexType t) const
{
  return (T*) (this->tileIndexX[x] + this->tileIndexY[y] + this->tileIndexZ[t]);
}

/*!

  Get the neighbours of a voxel. Weak (6) and Strong (26) neighbourship will be
  supported.

  \warning not implemented yet...

 */
template<class T>
inline
list<VideoPosition>
Video<T>::getNeighbours(const VideoPosition &pos, const int ng) const throw (MocaException)
{
    throw MocaException("Not implemented yet...");
} // Video<T>::getNeighbours

/*!

  \warning totally uneffective

 */
template<class T>
inline
Image<T>
Video<T>::getImage(indexType num) const
{
    Image<T> temp((ImageSize(this->size[DIM_X], this->size[DIM_Y], this->size[DIM_BAND])));

    for (sizeType x = 0; x < this->size[DIM_X]; x++) {
        for (sizeType y = 0; y < this->size[DIM_Y]; y++) {
            for (sizeType band = 0; band < this->size[DIM_BAND]; band++) {
                temp(x, y, band) = operator()(x, y, num, band);
            } // for
        } // for
    } // for

    return temp;
} // Video<T>::getImage


template<class T>
inline
void
Video<T>::setImage(indexType num, const Image<T> &img) throw (MocaException)
{

    // check sizes
    if (img.getSize()[DIM_X] != this->size[DIM_X]
        || img.getSize()[DIM_Y] != this->size[DIM_Y]
        || img.getSize()[DIM_BAND] != this->size[DIM_BAND]) {
        throw MocaException("Size of image does not fit.");
    } // if

    for (sizeType x = 0; x < this->size[DIM_X]; x++) {
        for (sizeType y = 0; y < this->size[DIM_Y]; y++) {
            for (sizeType band = 0; band < this->size[DIM_BAND]; band++) {
                operator()(x, y, num, band) = img(x, y, band);
            } // for
        } // for
    } // for

} // Video<T>::setImage




// handles a vector of videos
template<class T>
class VideoVector
{
public: 
	VideoVector (int32 n) {
		size = n;
		data.resize(size);
	}

    void resize (uint32 size) {
        this->size = size;
        data.resize(size);
        return;
 	}

    // get pointer to video
 	Video<uint8>* getVideo (uint32 number) {
        if (number>=size) {
            number=size-1;
        } // if
        return &(data[number]);
 	}

 	void setVideo (const Video<T> &vid, uint32 number) {
        if (number>=size) {
            number=size-1;
        } // if

        // copy image
        data[number]=vid;
        return;
    }
    
    
private:
	int32 size;  // number of images
public:
	vector<Video<T> > data;
};


#endif
